// ############################################################################################
//  VirtualWalk.cc
//  A demo application for VortexGE
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This program is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

//#define __HIGH_RESOLUTION__

// Common headers for demo application
#include "DemoCommon.h"


// Just for shortcut
typedef CBasicApplication<> MyApplicationBase;

// Specialized basic application class
class MyApplication : public MyApplicationBase {
  public:
    MyApplication();
    ~MyApplication() throw();
    virtual void EventKeyPress(EKeySymbol eKeySym);
    virtual void EventKeyRelease(EKeySymbol eKeySym);
    virtual void EventUpdateFrame(bool bWindowIsActive);
    virtual void EventCloseWindow();

    inline bool detectCollision(const float afCurPosXYZ[3], float afTransFactorXYZ[3]) {
      if(csTreeSpr.Collide(afCurPosXYZ, afTransFactorXYZ)) return(false);
      if(csGround.PlaneCollide(afCurPosXYZ[0] + afTransFactorXYZ[0],
                               afCurPosXYZ[1] + afTransFactorXYZ[1],
                               afCurPosXYZ[2] + afTransFactorXYZ[2],
                               10.0f) != ECT_None ) return(false);
      return(true);
    }

  private:
    static const int32_t NormalViewHeight = 65;

    CPanel3DBuffer        csGround;
    CWalkingCameraControl csCamera;
    CImageBuilder         csImgBuild;
    CImage                csImgTree;
    CTexture              csTGround;
    CSprite2D             csTreeSpr;
    CSoundBuffer         *pcsSBWalk;
    CSoundBuffer         *pcsSBJump;
    bool                  bForward, bBackward;     // For animation
    bool                  bLeft, bRight, bSL, bSR; // ---
    bool                  bRun, bLU, bLD;          // ---
    int32_t               n32Jump;                 // ---
};



// Callback functions
static bool WCCCallback(EWCCContext eContext, CWalkingCameraControl *pcsWCC, const float afCurPosXYZ[3], float afTransFactorXYZ[3], void *pvUserData)
{
  UNUSED_PARAMS_2(eContext, pcsWCC);
  MyApplication *pcsMyApp = reinterpret_cast<MyApplication*>(pvUserData);
  return(pcsMyApp->detectCollision(afCurPosXYZ, afTransFactorXYZ));
}


// Main program
int main()
{
  MyApplication csMyApp;
  return(csMyApp.DoMainLoop());
}


// MyApplication implementation
MyApplication::MyApplication()
#ifdef __HIGH_RESOLUTION__
: MyApplicationBase("Demo : Virtual Walk", NULL, 640, 480, 640, 480),
#else
: MyApplicationBase("Demo : Virtual Walk", NULL, 640, 480, 320, 240),
#endif
  pcsSBWalk(NULL), pcsSBJump(NULL)
{
  // Tell the user to be patient
  ShowInitSystemPleaseWaitMessage(iVideoEnv());

  // Set wanted FPS
  SetWantedFPS(50.0f);

  // Load audios
  if(!AudioIsDisabled()) {
    pcsSBWalk = iAudioEnv().NewSoundBuffer();
    pcsSBWalk->LoadWAV("../media/wav/walk_s.wav");
    pcsSBWalk->AdjustAmplitude(0.2f, 0.2f);
    pcsSBJump = iAudioEnv().NewSoundBuffer();
    pcsSBJump->LoadWAV("../media/wav/jump.wav");
    pcsSBJump->AdjustAmplitude(0.5f, 0.5f);
  }

  // Load ground texture
  try {
    csImgBuild.LoadJPG("../media/image/tex1.jpg", true);
    //csImgBuild.LoadPNG("/home/aloy/Images/konsole-vge.png", true);
    csImgBuild.Resize(256, 256);
    csImgBuild.SmearEffect(2);
  }
  catch(...) {
    csImgBuild.CreateNoiseImage(256, 256, 4, 20, 4, 2, 6, 2);
  }
  csImgBuild.TransferToTexture(csTGround);

  // Load tree sprite image
  try {
    csImgBuild.LoadPNG("../media/image/tree2.png", true);
  }
  catch(...) {
    csImgBuild.CreateEmptyImage(600, 600);

    FractalBamboo csMF1;
    csMF1.Draw(&csImgBuild, 320-10, 400   , 100);
    csMF1.Draw(&csImgBuild, 320+10, 400   , 100);
    csMF1.Draw(&csImgBuild, 320   , 400-50, 100);
    csMF1.Draw(&csImgBuild, 320   , 400-30, 100);
    csMF1.Draw(&csImgBuild, 320   , 400-10, 100);

    csImgBuild.AutoCrop(0, 0, 0);
  }
  csImgBuild.RoughResize(400, 600);
  csImgBuild.TransferToImage(csImgTree);

  // Free image
  csImgBuild.FreeImage();

  // Create scene
  i3DEnv().SetFogState(12, 12, 12, 12);
  i3DEnv().SetViewport(0, 0, i3DEnv().RenderBufferWidth(), i3DEnv().RenderBufferHeight());
  i3DEnv().SetZClipping(50, 1500);
  i3DEnv().SetDefaultOmniLightPosition(300, 600, 1100);
  i3DEnv().SetDefaultOmniLightColor(31, 31, 4);

  float groundVertices[41 * 3] = {
    -400,  0,1200,   -200,  0,1200,   0,  0,1200,   200,  0,1200,   400,  0,1200,
    -400,  0,1000,   -200,  0,1000,   0,  0,1000,   200,  0,1000,   400,  0,1000,
    -400,  0, 800,   -200,  0, 800,   0,  0, 800,   200,  0, 800,   400,  0, 800,
    -400,  0, 400,   -200,  0, 400,   0,  0, 400,   200,  0, 400,   400,  0, 400,
    -400,  0, 200,   -200,  0, 200,   0,  0, 200,   200,  0, 200,   400,  0, 200,
    -400,200,1200,   -200,200,1200,   0,200,1200,   200,200,1200,   400,200,1200,
    -400,200,1000,                                                  400,200,1000,
    -400,200, 800,                                                  400,200, 800,
    -400,200, 400,                                                  400,200, 400,
    -400,200, 200,   -200,200, 200,   0,200, 200,   200,200, 200,   400,200, 200 };
  uint32_t groundPanels[32 * 4] = {
     0, 1, 6, 5,    1, 2, 7, 6,    2, 3, 8, 7,    3, 4, 9, 8,
     5, 6,11,10,    6, 7,12,11,    7, 8,13,12,    8, 9,14,13,
    10,11,16,15,   11,12,17,16,   12,13,18,17,   13,14,19,18,
    15,16,21,20,   16,17,22,21,   17,18,23,22,   18,19,24,23,
    25,26, 1, 0,   26,27, 2, 1,   27,28, 3, 2,   28,29, 4, 3,
    29,31, 9, 4,   31,33,14, 9,   33,35,19,14,   35,40,24,19,
    40,39,23,24,   39,38,22,23,   38,37,21,22,   37,36,20,21,
    36,34,15,20,   34,32,10,15,   32,30, 5,10,   30,25, 0, 5 };
  csGround.InitFromMemory(groundVertices, 41, groundPanels, 32);
  csGround.SetTexture(&csTGround);
  csGround.CalcDefTexCoords();

  float afTPos[] = { -300, 0, 320,   -100, 0, 900,   100, 0, 900,   300, 0, 320,
                        0, 0, 320,   -300, 0, 610,   300, 0, 610,     0, 0, 610 };
  csTreeSpr.Initialize(afTPos, 8, &csImgTree);

  // Initialize camera
  csCamera.SetCallback(WCCCallback, this);
  csCamera.SetPosition(-200, NormalViewHeight, 400);
  csCamera.SetYRotation(0);
  csCamera.UpdateCameraMatrix();
  bForward = bBackward = bLeft = bRight = bRun = bLU = bLD = bSL = bSR = false;
  n32Jump = 0;
}

MyApplication::~MyApplication() throw()
{
  delete pcsSBWalk;
  delete pcsSBJump;
}

void MyApplication::EventKeyPress(EKeySymbol eKeySym)
{
       if(eKeySym == '\\')       iVideoEnv().SetFullscreen(!iVideoEnv().IsFullscreen());
  else if(eKeySym == VKS_ESC)    EventCloseWindow();
  else if(eKeySym == VKS_LEFT)   bLeft = (bRight) ? false : true;
  else if(eKeySym == VKS_RIGHT)  bRight = (bLeft) ? false : true;
  else if(eKeySym == VKS_UP)     bForward = (bBackward) ? false : true;
  else if(eKeySym == VKS_DOWN)   bBackward = (bForward) ? false : true;
  else if(eKeySym == VKS_LSHIFT) bRun = true;
  else if(eKeySym == 'a')        bLU = (bLD) ? false : true;
  else if(eKeySym == 'z')        bLD = (bLU) ? false : true;
  else if(eKeySym == 'q')        bSL = (bSR) ? false : true;
  else if(eKeySym == 'w')        bSR = (bSL) ? false : true;
  else if(eKeySym == ' ') {
    if(AudioIsDisabled() && (n32Jump <= 0))
      n32Jump = 4;
    else if(!pcsSBJump->IsPlaying() && (n32Jump <= 0)) {
        pcsSBJump->Play();
        n32Jump = 4;
    }
  }
}

void MyApplication::EventKeyRelease(EKeySymbol eKeySym)
{
       if(eKeySym == VKS_LEFT)   bLeft = false;
  else if(eKeySym == VKS_RIGHT)  bRight = false;
  else if(eKeySym == VKS_UP)     bForward = false;
  else if(eKeySym == VKS_DOWN)   bBackward = false;
  else if(eKeySym == VKS_LSHIFT) bRun = false;
  else if(eKeySym == 'a')        bLU = false;
  else if(eKeySym == 'z')        bLD = false;
  else if(eKeySym == 'q')        bSL = false;
  else if(eKeySym == 'w')        bSR = false;
}

void MyApplication::EventUpdateFrame(bool bWindowIsActive)
{
  // Update frame (if the application is not active)
  if(!bWindowIsActive) {
    iVideoEnv().PutRenderImageOnDisplay();
    iVideoEnv().SetForegroundColor(iVideoEnv().CreateColor(255, 255, 255));
    iVideoEnv().UseDefaultBigFont();
    iVideoEnv().DrawText(0, 150, "Inactive State", ETA_CenterScreen);
    iVideoEnv().UseDefaultSmallFont();
    iVideoEnv().UpdateDisplay();
    return;
  }

  // Update Y rotation and XY position
  if(!bRun) {
    if(bLeft)     csCamera.RotateLeft(10);
    if(bRight)    csCamera.RotateRight(10);
    if(bForward)  csCamera.StepForward(4);
    if(bBackward) csCamera.StepBackward(4);
    if(bSL)       csCamera.StrafeLeft(4);
    if(bSR)       csCamera.StrafeRight(4);
  }
  else {
    if(bLeft)     csCamera.RotateLeft(20);
    if(bRight)    csCamera.RotateRight(20);
    if(bForward)  csCamera.StepForward(8);
    if(bBackward) csCamera.StepBackward(8);
    if(bSL)       csCamera.StrafeLeft(8);
    if(bSR)       csCamera.StrafeRight(8);
  }
  // Update Y position
  if(n32Jump) {
    n32Jump--;
    csCamera.ClimbUp(10);
  }
  else if(csCamera.GetCurYPos() > NormalViewHeight) {
    csCamera.ClimbDown(5);
  }
  // Update X rotation
  if(bLU)
    csCamera.LookUp(150);
  else if(bLD)
    csCamera.LookDown(150);
  else
    csCamera.LookStraight();

  // Update matrix
  csCamera.UpdateCameraMatrix();

  // Clear scene
  i3DEnv().ClearRenderBuffer(VColor::DimGray);
  i3DEnv().ClearZBuffer();

  // Render the scene
  csGround.FogTGRender(csCamera.GetCameraMatrix());
  csTreeSpr.FogSFRender(csCamera.GetCameraMatrix());
  //csGround.TGRender(csCamera.GetCameraMatrix());
  //csTreeSpr.SFRender(csCamera.GetCameraMatrix());

  // Update display
  iVideoEnv().PutRenderImageOnDisplay();
  iVideoEnv().SetForegroundColor(iVideoEnv().CreateColor(255, 255, 255));
  iVideoEnv().DrawText(10, 20, GetFPSInfoString());
  iVideoEnv().DrawText(10,  50, "Press 'ARROW' to move around");
  iVideoEnv().DrawText(10,  65, "Press 'SHIFT' to to run");
  iVideoEnv().DrawText(10,  80, "Press 'SPACE' to jump");
  iVideoEnv().DrawText(10,  95, "Press 'A' to look up");
  iVideoEnv().DrawText(10, 110, "Press 'Z' to look down");
  iVideoEnv().DrawText(10, 125, "Press 'Q' to strafe left");
  iVideoEnv().DrawText(10, 140, "Press 'W' to strafe right");
  iVideoEnv().DrawText(10, 170, "Press escape to exit ...");
  iVideoEnv().UpdateDisplay();

  // Update audio
  if(!AudioIsDisabled()) {
    if((bLeft || bRight || bForward || bBackward || bSL || bSR) &&
       (!pcsSBWalk->IsPlaying())) pcsSBWalk->Play();
    iAudioEnv().UpdateBuffer();
  }
}

void MyApplication::EventCloseWindow()
{
  iVideoEnv().MessageBox("End Program", "Click \"OK\" to exit ...");
  MyApplicationBase::EventCloseWindow();
}
