// ############################################################################################
//  AnimatedImageController.cc
//  Implementation of CTexture, CAnimatedTexture, CImage, CAnimatedImage, and 
//  CImageBuilder classes
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "../Base/MessageDisplayer.h"
#include "../System/SystemUtilities.h"
#include "AnimatedImageController.h"

// C++ headers
#include <fstream>


namespace vge {

//
// VAC header
//
#define _VAC_ID1_ 'V'
#define _VAC_ID2_ 'A'
#define _VAC_ID3_ 'C'
#define _VAC_VER_ 100

struct _TVACHeader {
  char   acID[3];         // File ID, must be "VAC"
  uint8_t  un8Version;    // File version
  uint32_t un32TotalAS;   // Total animation sets
  uint32_t un32Reserved1; // Reserved, must be zero
  uint32_t un32Reserved2; // Reserved, must be zero
} _attr_packed;


//
// The advanced frame control for animated image class
//
CAImgControl::CAImgControl()
: _pcsImgSrc    (NULL), _ppun8ImgBuff(NULL), _ptASList    (NULL),
  _un32TotalAS  (0),    _un32SelAS   (0),    _un32SelFrame(0),
  _un32StepDelay(0)
{
}

CAImgControl::~CAImgControl() throw()
{
  FreeAll();
}

void CAImgControl::SetImageSource(CAnimatedImage *pcsAnimatedImage)
{
  // Check
  vassert(pcsAnimatedImage->_un32TotFrame,
          "Can't use an uninitialized CAnimatedImage");

  // Select new image source
  _pcsImgSrc = pcsAnimatedImage;

  // Destroy old reference buffer, if any
  if(_ppun8ImgBuff) free(_ppun8ImgBuff);

  // Allocate memory
  _ppun8ImgBuff = reinterpret_cast<uint8_t**>(malloc(_pcsImgSrc->_un32TotFrame * sizeof(uint8_t *)));
  if(_ppun8ImgBuff == NULL)
    VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for a new reference buffer");

  // Set references
  for(size_t c = 0; c < _pcsImgSrc->_un32TotFrame; ++c)
    _ppun8ImgBuff[c] = _pcsImgSrc->_pun8BufferArray + _pcsImgSrc->_un32FrameSize * c;

  // Refersh data
  _un32SelAS     = 0;
  _un32SelFrame  = 0;
  _un32StepDelay = 0;
  _pcsImgSrc->_pun8Buffer = _pcsImgSrc->_pun8BufferArray;
}

void CAImgControl::FreeAll()
{
  if(_ptASList) {
    for(size_t c = 0; c < _un32TotalAS; ++c)
      if(_ptASList[c].pun32FrameList) free(_ptASList[c].pun32FrameList);
    free(_ptASList);
  }
  if(_ppun8ImgBuff) free(_ppun8ImgBuff);
  _ptASList     = NULL;
  _ppun8ImgBuff = NULL;
  _un32TotalAS  = 0;
  _un32SelAS    = 0;
  _un32SelFrame = 0;
}

uint32_t CAImgControl::AddAnimationSet(uint32_t un32TotalASFrames)
{
  // Temporary variables
  _TAnimSet *ptAS;

  // Allocate memory
  ptAS = reinterpret_cast<_TAnimSet*>(malloc((_un32TotalAS + 1) * sizeof(_TAnimSet)));
  if(ptAS == NULL)
    VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for a new animation set");

  // Copy and free old data
  memcpy(ptAS, _ptASList, _un32TotalAS * sizeof(_TAnimSet));
  memset(&ptAS[_un32TotalAS], 0, sizeof(_TAnimSet));
  free(_ptASList);
  _ptASList = ptAS;

  // Allocate memory for frames
  _AllocASFrame_(_un32TotalAS, un32TotalASFrames);
  _un32TotalAS++;

  // Return the index to the new animation set
  return(_un32TotalAS - 1);
}

uint32_t CAImgControl::AddAnimationSets(uint32_t un32ASCount, uint32_t aun32TotalASFrames[])
{
  // Allocate memory
  _TAnimSet *ptAS = reinterpret_cast<_TAnimSet*>(malloc((_un32TotalAS + un32ASCount) * sizeof(_TAnimSet)));
  if(ptAS == NULL)
    VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for new animation sets");

  // Copy and free old data
  memcpy(ptAS, _ptASList, _un32TotalAS * sizeof(_TAnimSet));
  memset(&ptAS[_un32TotalAS], 0, un32ASCount * sizeof(_TAnimSet));
  free(_ptASList);
  _ptASList = ptAS;

  // Allocate memory for frames
  if(aun32TotalASFrames) {
    for(size_t c = 0; c < un32ASCount; ++c)
      _AllocASFrame_(_un32TotalAS + c, aun32TotalASFrames[c]);
  }
  _un32TotalAS += un32ASCount;

  // Return the start index to the new animation sets
  return(_un32TotalAS - un32ASCount);
}

uint32_t CAImgControl::GetAnimationSetCount() const
{
  return(_un32TotalAS);
}

void CAImgControl::AddASFrame(uint32_t un32AnimationSetIndex, uint32_t un32FrameReference)
{
  // Temporary variables
  uint32_t *pun32Tmp, un32Tot;

  // Save old datas
  pun32Tmp = _ptASList[un32AnimationSetIndex].pun32FrameList;
  un32Tot  = _ptASList[un32AnimationSetIndex].un32TotalFrames;

  // Allocate memory
  _AllocASFrame_(un32AnimationSetIndex, _ptASList[un32AnimationSetIndex].un32TotalFrames + 1);

  // Copy and free old data
  memcpy(_ptASList[un32AnimationSetIndex].pun32FrameList, pun32Tmp,  un32Tot * sizeof(uint32_t));
  free(pun32Tmp);

  // Set new data
  _ptASList[un32AnimationSetIndex].pun32FrameList[un32Tot] = un32FrameReference;
}

void CAImgControl::AddASFrames(uint32_t un32AnimationSetIndex, uint32_t un32FrameCount, uint32_t aun32FrameReferences[])
{
  // Save old datas
        uint32_t *oldFrames = _ptASList[un32AnimationSetIndex].pun32FrameList;
  const size_t    totFrames = _ptASList[un32AnimationSetIndex].un32TotalFrames;

  // Allocate new memory block
  _AllocASFrame_(un32AnimationSetIndex, _ptASList[un32AnimationSetIndex].un32TotalFrames + un32FrameCount);

  // Copy and free old data
  memcpy(_ptASList[un32AnimationSetIndex].pun32FrameList, oldFrames, totFrames * sizeof(uint32_t));
  free(oldFrames);

  // Set new datas
  for(size_t c = 0; c < un32FrameCount; c++)
    _ptASList[un32AnimationSetIndex].pun32FrameList[totFrames + c] = aun32FrameReferences[c];
}

uint32_t CAImgControl::GetASFrameCount(uint32_t un32AnimationSetIndex) const
{
  return(_ptASList[un32AnimationSetIndex].un32TotalFrames);
}

void CAImgControl::SetASFrame(uint32_t un32AnimationSetIndex, uint32_t un32ASFrameIndex, uint32_t un32FrameReference)
{
  _ptASList[un32AnimationSetIndex].pun32FrameList[un32ASFrameIndex] = un32FrameReference;
}

void CAImgControl::SetASFrames(uint32_t un32AnimationSetIndex, uint32_t aun32FrameReferences[])
{
  for(size_t c = 0; c < _ptASList[un32AnimationSetIndex].un32TotalFrames; ++c) {
    _ptASList[un32AnimationSetIndex].pun32FrameList[c] = aun32FrameReferences[c];
  }
}

void CAImgControl::_AllocASFrame_(uint32_t un32AnimationSetIndex, uint32_t un32TotalASFrames)
{
  // Allocate memory for frames
  if(un32TotalASFrames == 0) return;
  _ptASList[un32AnimationSetIndex].pun32FrameList = reinterpret_cast<uint32_t*>(malloc(un32TotalASFrames * sizeof(uint32_t)));
  if(_ptASList[un32AnimationSetIndex].pun32FrameList == NULL)
    VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for new frames");
  _ptASList[un32AnimationSetIndex].un32TotalFrames = un32TotalASFrames;
}

void CAImgControl::SaveVAC(const CDataSink &sink) const
{
  // Check
  if(_un32TotalAS == 0)
    VError(VERR_NotInitialized, _NFL_, "No data to be saved (this CAImgControl is not initialized)");

  // Get the stream
  const char *strmName = sink.Name().c_str();
  ostream    &os       = sink.Stream();
  os.exceptions(ios::badbit | ios::failbit | ios::eofbit);

  // Make header
  _TVACHeader tHeader;
  memset(&tHeader, 0, sizeof(tHeader));
  tHeader.acID[0]     = _VAC_ID1_;
  tHeader.acID[1]     = _VAC_ID2_;
  tHeader.acID[2]     = _VAC_ID3_;
  tHeader.un8Version  = _VAC_VER_;
  tHeader.un32TotalAS = _un32TotalAS;

  // Write the data
  try {
    // Write header
    os.write(reinterpret_cast<char*>(&tHeader), sizeof(tHeader));
    // Write animation datas
    for(size_t i = 0; i < _un32TotalAS; ++i) {
      // Write total frames
      os.write(reinterpret_cast<char*>(&_ptASList[i].un32TotalFrames), sizeof(_ptASList[i].un32TotalFrames));
      // Write frame references
      os.write(reinterpret_cast<char*>(_ptASList[i].pun32FrameList), sizeof(uint32_t) * _ptASList[i].un32TotalFrames);
    }
  }
  catch(ios_base::failure&) {
    VError(VERR_WriteStream, _NFL_, "Couldn't write stream \"%s\"", strmName);
  }
  catch(...) {
    throw;
  }
}

void CAImgControl::LoadVAC(const CDataSource &source)
{
  // Get the stream
  const char *strmName = source.Name().c_str();
  pistream    &is      = source.PStream();

  // Read the header
  _TVACHeader tHeader;
  is.read(reinterpret_cast<char*>(&tHeader), sizeof(tHeader));
  if(is.fail())
    VError(VERR_ReadStream, _NFL_, "Couldn't read header from stream \"%s\"", strmName);

  // Check header
  if((tHeader.acID[0] != _VAC_ID1_) ||
     (tHeader.acID[1] != _VAC_ID2_) || (tHeader.acID[2] != _VAC_ID3_))
    VError(VERR_InvalidStream, _NFL_, "\"%s\" is an invalid VAC stream", strmName);
  if(tHeader.un8Version != _VAC_VER_)
    VError(VERR_IncorrectVersion, _NFL_, "\"%s\" has an incorrect version", strmName);
  if(tHeader.un32Reserved1 || tHeader.un32Reserved2)
    VError(VERR_InvalidHeaderData, _NFL_, "Invalid header data found in stream \"%s\"", strmName);

   // Allocate memory
  _TAnimSet *ptASList = reinterpret_cast<_TAnimSet*>(malloc(sizeof(_TAnimSet) * tHeader.un32TotalAS));
  if(!ptASList)
    VError(VERR_NotEnoughMemory, _NFL_, "Not enough memory to create (a) new animation set(s)");
  memset(ptASList, 0, sizeof(_TAnimSet) * tHeader.un32TotalAS);

  // Read animation datas
  is.exceptions(ios::badbit | ios::failbit | ios::eofbit);
  try {
    for(uint32_t i = 0; i < tHeader.un32TotalAS; i++) {
      // Read total frames
      is.read(reinterpret_cast<char*>(&ptASList[i].un32TotalFrames), sizeof(ptASList[i].un32TotalFrames));
      // Allocate memory
      ptASList[i].pun32FrameList = reinterpret_cast<uint32_t*>(malloc(sizeof(uint32_t) * ptASList[i].un32TotalFrames));
      if(!ptASList[i].pun32FrameList) {
        // Free the previously allocated memory
        for(uint32_t d = 0; d < tHeader.un32TotalAS; d++)
          if(ptASList[d].pun32FrameList) {
            free(ptASList[d].pun32FrameList);
            ptASList[d].pun32FrameList = NULL;
          }
        free(ptASList);
        // Raise error
        VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for loading frame references data");
      }
      // Read frame references
      is.read(reinterpret_cast<char*>(ptASList[i].pun32FrameList), sizeof(uint32_t) * ptASList[i].un32TotalFrames);
    }
  }
  catch(ios_base::failure&) {
    VError(VERR_ReadStream, _NFL_, "Couldn't read animation data from stream \"%s\"", strmName);
  }
  catch(...) {
    throw;
  }

  // Free previous datas
  if(_ptASList) {
    for(uint32_t i = 0; i < _un32TotalAS; i++)
      if(_ptASList[i].pun32FrameList) free(_ptASList[i].pun32FrameList);
    free(_ptASList);
  }

  // Set variables
  _ptASList      = ptASList;
  _un32TotalAS   = tHeader.un32TotalAS;
  _un32SelAS     = 0;
  _un32SelFrame  = 0;
  _un32StepDelay = 0;
}

} // namespace vge
