// ############################################################################################
//  BitmapFont.cc
//  Implementation of CBitmapFont class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "../3DEnvironment/3DEnvironment.h"
#include "../Base/MessageDisplayer.h"
#include "../System/SystemUtilities.h"
#include "BitmapFont.h"
#include "ImageBuilder.h"

// C++ headers
#include <fstream>


namespace vge {

//
// VBF header
//
#define _VBF_ID1_ 'V'
#define _VBF_ID2_ 'B'
#define _VBF_ID3_ 'F'
#define _VBF_VER_ 100

struct _TVBFHeader {
  char   acID[3];         // File ID, must be "VBF"
  uint8_t  un8Version;    // File version
  uint32_t un32FWidth;    // Font width in pixels
  uint32_t un32FHeight;   // Font height in pixels
  uint32_t un32FRow;      // Total rows in font matrix
  uint32_t un32FCol;      // Total columns in font matrix
  uint32_t un32PBytes;    // Size of packed data in bytes
  uint32_t un32Reserved1; // Reserved, must be zero
  uint32_t un32Reserved2; // Reserved, must be zero
} _attr_packed;


//
// Font helper functions
//
static void _PackFontData_(u_char *pucDstBuff, uint32_t un32DstSize, const u_char*pucSrcBuff, uint32_t un32SrcSize);
static void _UnpackFontData_(u_char *pucDstBuff, uint32_t un32DstSize, const u_char*pucSrcBuff, uint32_t un32SrcSize);


//
// The bitmap font class
//
CBitmapFont::CBitmapFont()
: _un32FW(0), _un32FH(0)
{
  memset(_apcFontArray, 0, sizeof(char *) * _FONT_ROW_ * _FONT_COL_);
}

CBitmapFont::~CBitmapFont() throw()
{
  FreeFont();
}

void CBitmapFont::FreeFont()
{
  for(uint32_t un32Cnt = 0; un32Cnt < (_FONT_ROW_ * _FONT_COL_); un32Cnt++) {
    if(_apcFontArray[un32Cnt]) free(_apcFontArray[un32Cnt]);
    _apcFontArray[un32Cnt] = NULL;
  }
  _un32FW = 0;
  _un32FH = 0;
}
void CBitmapFont::PutText(int32_t n32X, int32_t n32Y, const string &csText, XColor15 xColorVRGB, EFontStyle eStyle) const
{
  // Temporary variables
  uint32_t     un32Cnt, un32X, un32Y;
  uint32_t     *pun32DPos, *pun32DTmp, *pun32DTmp2;
  char         *pcFont;
  uint32_t     un32Len  = csText.length();
  const u_char *pucChar = reinterpret_cast<const u_char*>(csText.c_str());

  // Do simple Y clipping
  if(n32Y < int32_t(C3DEnv::Active3DEnv()->_n32MinY)) return;
  if(n32Y > int32_t(C3DEnv::Active3DEnv()->_n32MaxY - _un32FH1)) return;

  // Draw text
  pun32DPos = reinterpret_cast<uint32_t*>(&C3DEnv::Active3DEnv()->_tRB->pun8Buffer[(n32Y * C3DEnv::Active3DEnv()->_tRB->un32Width) << 2]);
  for(un32Cnt = 0; un32Cnt < un32Len; un32Cnt++) {
    // Do simple X clipping
    if((n32X < int32_t(C3DEnv::Active3DEnv()->_n32MinX)) || (n32X > int32_t(C3DEnv::Active3DEnv()->_n32MaxX - _un32FW1))) {
      n32X += _un32FW1;
      pucChar++;
      continue;
    }
    // Put the character if not clipped;
    else {
      pcFont = _apcFontArray[uint32_t(*pucChar++) + uint32_t(eStyle)];
      if(pcFont) {
        pun32DTmp = pun32DPos + n32X;
        for(un32Y = 0; un32Y < _un32FH; un32Y++) {
          pun32DTmp2 = pun32DTmp;
          for(un32X = 0; un32X < _un32FW; un32X++) {
            if(*pcFont++) *pun32DTmp2 = xColorVRGB;
            pun32DTmp2++;
          }
          pun32DTmp += C3DEnv::Active3DEnv()->_tRB->un32Width;
        }
      }
      n32X += _un32FW1;
    }
  }
}

void CBitmapFont::PutTextSolid(int32_t n32X, int32_t n32Y, const string &csText, XColor15 xTextColorVRGB, XColor15 xBackColorVRGB, EFontStyle eStyle) const
{
  // Temporary variables
  uint32_t     un32Cnt, un32X, un32Y;
  uint32_t     *pun32DPos, *pun32DTmp, *pun32DTmp2;
  char         *pcFont;
  uint32_t     un32Len  = csText.length();
  const u_char *pucChar = reinterpret_cast<const u_char*>(csText.c_str());

  // Do simple Y clipping
  if(n32Y < int32_t(C3DEnv::Active3DEnv()->_n32MinY)) return;
  if(n32Y > int32_t(C3DEnv::Active3DEnv()->_n32MaxY - _un32FH1)) return;

  // Draw text
  pun32DPos = reinterpret_cast<uint32_t*>(&C3DEnv::Active3DEnv()->_tRB->pun8Buffer[(n32Y * C3DEnv::Active3DEnv()->_tRB->un32Width) << 2]);
  for(un32Cnt = 0; un32Cnt < un32Len; un32Cnt++) {
    // Do simple X clipping
    if((n32X < int32_t(C3DEnv::Active3DEnv()->_n32MinX)) || (n32X > int32_t(C3DEnv::Active3DEnv()->_n32MaxX - _un32FW1))) {
      n32X += _un32FW1;
      pucChar++;
      continue;
    }
    // Put the character if not clipped;
    else {
      pcFont = _apcFontArray[uint32_t(*pucChar++) + uint32_t(eStyle)];
      if(pcFont) {
        pun32DTmp = pun32DPos + n32X;
        for(un32Y = 0; un32Y < _un32FH; un32Y++) {
          pun32DTmp2 = pun32DTmp;
          for(un32X = 0; un32X < _un32FW; un32X++) {
            if(*pcFont++)
              *pun32DTmp2++ = xTextColorVRGB;
            else
              *pun32DTmp2++ = xBackColorVRGB;
          }
          pun32DTmp += C3DEnv::Active3DEnv()->_tRB->un32Width;
        }
      }
      else {
        pun32DTmp = pun32DPos + n32X;
        for(un32Y = 0; un32Y < _un32FH; un32Y++) {
          pun32DTmp2 = pun32DTmp;
          for(un32X = 0; un32X < _un32FW; un32X++) *pun32DTmp2++ = xBackColorVRGB;
          pun32DTmp += C3DEnv::Active3DEnv()->_tRB->un32Width;
        }
      }
      n32X += _un32FW1;
    }
  }
}

void CBitmapFont::LoadBMP(const CDataSource &source)
{
  // Temporary variables
  CImageBuilder  csIB;
  CAnimatedImage csAI;
  uint32_t       un32Cnt, un32CntH, un32CntW, un32CW, un32CH, un32FW, un32FH;
  uint32_t       *pun32Src, *pun32Get;
  char           *apcTmp[_FONT_ROW_ * _FONT_COL_], *pucPut;
  bool           bEmpty;

  // Load image
  csIB.LoadBMP(source, false);
  csIB.Grayscale(true);

  // Check image
  if(csIB.GetWidth() % _FONT_COL_) {
    csIB.FreeImage();
    VError(VERR_InvalidStream, _NFL_, "Image width must be divisible by %d", _FONT_COL_);
  }
  if(csIB.GetHeight() % _FONT_ROW_) {
    csIB.FreeImage();
    VError(VERR_InvalidStream, _NFL_, "Image height must be divisible by %d", _FONT_ROW_);
  }
  un32CW  = csIB.GetWidth() / _FONT_COL_;
  un32CH  = csIB.GetHeight() / _FONT_ROW_;
  un32FW  = un32CW - _FONT_DX_;
  un32FH  = un32CH - _FONT_DY_;

  // Transfer to animated image
  try {
    csIB.TransferToAnimatedImage(csAI, _FONT_ROW_, _FONT_COL_);
  }
  catch(...) {
    csIB.FreeImage();
    throw;
  }
  csIB.FreeImage();

  // Transfer to font array
  memset(apcTmp, 0, sizeof(char *) * _FONT_ROW_ * _FONT_COL_);
  pun32Src = reinterpret_cast<uint32_t*>(csAI._pun8BufferArray);
  for(un32Cnt = 0; un32Cnt < (_FONT_ROW_ * _FONT_COL_); un32Cnt++) {
    // Check if all zeroes
    bEmpty   = true;
    pun32Get = pun32Src;
    for(un32CntH = 0; un32CntH < un32FH; un32CntH++) {
      for(un32CntW = 0; un32CntW < un32FW; un32CntW++) {
        if(*pun32Get++) {
          bEmpty = false;
          break;
        }
      }
      pun32Get += _FONT_DX_;
    }
    // Create font cell only if not empty
    if(!bEmpty) {
      // Allocate memory
      apcTmp[un32Cnt] = reinterpret_cast<char*>(malloc(un32FW * un32FH));
      if(!apcTmp[un32Cnt]) {
        // Free the previously allocated memory
        for(un32Cnt = 0; un32Cnt < (_FONT_ROW_ * _FONT_COL_); un32Cnt++)
          if(apcTmp[un32Cnt]) free(apcTmp[un32Cnt]);
        // Raise error
        VError(VERR_NotEnoughMemory, _NFL_, "Not enough memory to load font");
      }
      // Extract font image
      pun32Get = pun32Src;
      pucPut   = apcTmp[un32Cnt];
      for(un32CntH = 0; un32CntH < un32FH; un32CntH++) {
        for(un32CntW = 0; un32CntW < un32FW; un32CntW++) {
          if(*pun32Get)
            *pucPut = 1;
          else
            *pucPut = 0;
          pucPut++;
          pun32Get++;
        }
        pun32Get += _FONT_DX_;
      }
    }
    pun32Src += (csAI._un32FrameSize >> 2);
  }
  csAI.FreeImage();

  // Copy references
  FreeFont();
  for(un32Cnt = 0; un32Cnt < (_FONT_ROW_ * _FONT_COL_); un32Cnt++) _apcFontArray[un32Cnt] = apcTmp[un32Cnt];
  _un32FW  = un32FW;
  _un32FH  = un32FH;
  _un32FW1 = _un32FW - 1;
  _un32FH1 = _un32FH - 1;
}

void CBitmapFont::LoadVBF(const CDataSource &source)
{
  // Get the stream
  const char *strmName = source.Name().c_str();
  pistream    &is      = source.PStream();

  // Read the header
  _TVBFHeader tHeader;
  is.read(reinterpret_cast<char*>(&tHeader), sizeof(tHeader));
  if(is.fail())
    VError(VERR_ReadStream, _NFL_, "Couldn't read VBF header from stream \"%s\"", strmName);

  // Check header
  if((tHeader.acID[0] != _VBF_ID1_) ||
     (tHeader.acID[1] != _VBF_ID2_) || (tHeader.acID[2] != _VBF_ID3_))
    VError(VERR_InvalidStream, _NFL_, "\"%s\" is an invalid VBF stream", strmName);
  if(tHeader.un8Version != _VBF_VER_)
    VError(VERR_IncorrectVersion, _NFL_, "\"%s\" has an incorrect version", strmName);
  if((tHeader.un32FRow != _FONT_ROW_) && (tHeader.un32FCol != _FONT_COL_))
    VError(VERR_InvalidHeaderData, _NFL_, "Invalid row and/or column size found in stream \"%s\"", strmName);
  if(tHeader.un32Reserved1 || tHeader.un32Reserved2)
    VError(VERR_InvalidHeaderData, _NFL_, "Invalid header data found in stream \"%s\"", strmName);
  
  // Read font map
  u_char aucMap[_FONT_MAP_];
  is.read(reinterpret_cast<char*>(aucMap), sizeof(aucMap[0]) * _FONT_MAP_); 
  if(is.fail())
    VError(VERR_ReadStream, _NFL_, "Couldn't read font map data from stream \"%s\"", strmName);

  // Array for font cells
  char *apcFontArray[_FONT_ROW_ * _FONT_COL_];
  memset(apcFontArray, 0, sizeof(apcFontArray[0]) * _FONT_ROW_ * _FONT_COL_);

  // Allocate memory for temporary buffer
  vector<u_char> pucTmp(tHeader.un32PBytes);

  // Allocate memory for font cells
  uint32_t un32UBytes = tHeader.un32FWidth * tHeader.un32FHeight;
  for(uint32_t i = 0; i < _FONT_MAP_; i++) {
    for(uint32_t j = 0; j < 8; j++)
      if(aucMap[i] & (0x01 << j)) {
        apcFontArray[(i << 3) + j] = reinterpret_cast<char*>(malloc(un32UBytes));
        if(!apcFontArray[(i << 3) + j]) {
          // Free the previously allocated memory
          for(uint32_t k = 0; k < (_FONT_ROW_ * _FONT_COL_); k++) 
            if(apcFontArray[k]) free(apcFontArray[k]);
          // Raise error
          VError(VERR_NotEnoughMemory, _NFL_, "Not enough memory to allocate font cells");
        }
      }
  }
  
  // Read font data
  for(uint32_t i = 0; i < (_FONT_MAP_ << 3); i++) {
    if(apcFontArray[i]) {
      // Read the data
      is.read(reinterpret_cast<char*>(&pucTmp[0]), tHeader.un32PBytes); 
      if(is.fail()) {
        // Free the previously allocated memory
        for(uint32_t k = 0; k < (_FONT_ROW_ * _FONT_COL_); k++) 
          if(apcFontArray[k]) free(apcFontArray[k]);
        // Raise error
        VError(VERR_ReadStream, _NFL_, "Couldn't read font cells data from stream \"%s\"", strmName);
      }
      // Unpack the data
      _UnpackFontData_(reinterpret_cast<u_char*>(apcFontArray[i]), un32UBytes, &pucTmp[0], tHeader.un32PBytes);
    }
  }

  // Copy variables
  FreeFont();
  memcpy(_apcFontArray, apcFontArray, sizeof(apcFontArray[0]) * _FONT_ROW_ * _FONT_COL_);
  _un32FW  = tHeader.un32FWidth;
  _un32FH  = tHeader.un32FHeight;
  _un32FW1 = _un32FW - 1;
  _un32FH1 = _un32FH - 1;
}

void CBitmapFont::SaveVBF(const CDataSink &sink) const
{
  // Check
  if((_un32FW == 0) || (_un32FH == 0))
    VError(VERR_NotInitialized, _NFL_, "No data to be saved (this CBitmapFont is not initialized)");

  // Allocate temporary memory
  uint32_t un32PSize = (_un32FW * _un32FH) >> 3;
  if((un32PSize << 3) < (_un32FW * _un32FH)) un32PSize++;
    
  vector<u_char> pucTmp(un32PSize);

  // Make header
  _TVBFHeader tHeader;
  memset(&tHeader, 0, sizeof(tHeader));
  tHeader.acID[0]     = _VBF_ID1_;
  tHeader.acID[1]     = _VBF_ID2_;
  tHeader.acID[2]     = _VBF_ID3_;
  tHeader.un8Version  = _VBF_VER_;
  tHeader.un32FWidth  = _un32FW;
  tHeader.un32FHeight = _un32FH;
  tHeader.un32FRow    = _FONT_ROW_;
  tHeader.un32FCol    = _FONT_COL_;
  tHeader.un32PBytes  = un32PSize;

  // Create font map
  char aucMap[_FONT_MAP_];
  for(uint32_t i = 0; i < _FONT_MAP_; i++) {
    aucMap[i] = 0x00;
    for(uint32_t j = 0; j < 8; j++)
      if(_apcFontArray[(i << 3) + j]) aucMap[i] |= (0x01 << j);
  }

  // Get the stream
  const char *strmName = sink.Name().c_str();
  ostream    &os       = sink.Stream();
  os.exceptions(ios::badbit | ios::failbit | ios::eofbit);

  // Write the file 
  try {
    // Write the header
    os.write(reinterpret_cast<char*>(&tHeader), sizeof(tHeader));
    // Write font map
    os.write(reinterpret_cast<char*>(aucMap), sizeof(aucMap[0]) * _FONT_MAP_);
    // Write font cells data
    for(uint32_t i = 0; i < (_FONT_MAP_ << 3); i++) {
      if(_apcFontArray[i]) {
        _PackFontData_(&pucTmp[0], un32PSize, reinterpret_cast<u_char*>(_apcFontArray[i]), _un32FW * _un32FH);
        os.write(reinterpret_cast<char*>(&pucTmp[0]), un32PSize);
      }
    }
  }
  catch(ios_base::failure&) {
    VError(VERR_WriteStream, _NFL_, "Couldn't write stream \"%s\"", strmName);
  }
  catch(...) {
    throw;
  }
}


//
// Local functions implementation
//
static void _PackFontData_(u_char *pucDstBuff, uint32_t un32DstSize, const u_char*pucSrcBuff, uint32_t un32SrcSize)
{
  // Temporary variables
  uint32_t un32Cnt1, un32Cnt2;

  // Pack font data
  memset(pucDstBuff, 0, un32DstSize);
  un32Cnt2 = 0;
  for(un32Cnt1 = 0; un32Cnt1 < un32DstSize; un32Cnt1++) {
    if(pucSrcBuff[un32Cnt2]) pucDstBuff[un32Cnt1] |= 0x01;
    un32Cnt2++; if(un32Cnt2 >= un32SrcSize) return;
    if(pucSrcBuff[un32Cnt2]) pucDstBuff[un32Cnt1] |= 0x02;
    un32Cnt2++; if(un32Cnt2 >= un32SrcSize) return;
    if(pucSrcBuff[un32Cnt2]) pucDstBuff[un32Cnt1] |= 0x04;
    un32Cnt2++; if(un32Cnt2 >= un32SrcSize) return;
    if(pucSrcBuff[un32Cnt2]) pucDstBuff[un32Cnt1] |= 0x08;
    un32Cnt2++; if(un32Cnt2 >= un32SrcSize) return;
    if(pucSrcBuff[un32Cnt2]) pucDstBuff[un32Cnt1] |= 0x10;
    un32Cnt2++; if(un32Cnt2 >= un32SrcSize) return;
    if(pucSrcBuff[un32Cnt2]) pucDstBuff[un32Cnt1] |= 0x20;
    un32Cnt2++; if(un32Cnt2 >= un32SrcSize) return;
    if(pucSrcBuff[un32Cnt2]) pucDstBuff[un32Cnt1] |= 0x40;
    un32Cnt2++; if(un32Cnt2 >= un32SrcSize) return;
    if(pucSrcBuff[un32Cnt2]) pucDstBuff[un32Cnt1] |= 0x80;
    un32Cnt2++; if(un32Cnt2 >= un32SrcSize) return;
  }
}

static void _UnpackFontData_(u_char *pucDstBuff, uint32_t un32DstSize, const u_char*pucSrcBuff, uint32_t un32SrcSize)
{
  // Temporary variables
  uint32_t un32Cnt1, un32Cnt2;

  // Unpack font data
  memset(pucDstBuff, 0, un32DstSize);
  un32Cnt2 = 0;
  for(un32Cnt1 = 0; un32Cnt1 < un32SrcSize; un32Cnt1++) {
    if(pucSrcBuff[un32Cnt1] & 0x01) pucDstBuff[un32Cnt2] = 1;
    un32Cnt2++; if(un32Cnt2 >= un32DstSize) return;
    if(pucSrcBuff[un32Cnt1] & 0x02) pucDstBuff[un32Cnt2] = 1;
    un32Cnt2++; if(un32Cnt2 >= un32DstSize) return;
    if(pucSrcBuff[un32Cnt1] & 0x04) pucDstBuff[un32Cnt2] = 1;
    un32Cnt2++; if(un32Cnt2 >= un32DstSize) return;
    if(pucSrcBuff[un32Cnt1] & 0x08) pucDstBuff[un32Cnt2] = 1;
    un32Cnt2++; if(un32Cnt2 >= un32DstSize) return;
    if(pucSrcBuff[un32Cnt1] & 0x10) pucDstBuff[un32Cnt2] = 1;
    un32Cnt2++; if(un32Cnt2 >= un32DstSize) return;
    if(pucSrcBuff[un32Cnt1] & 0x20) pucDstBuff[un32Cnt2] = 1;
    un32Cnt2++; if(un32Cnt2 >= un32DstSize) return;
    if(pucSrcBuff[un32Cnt1] & 0x40) pucDstBuff[un32Cnt2] = 1;
    un32Cnt2++; if(un32Cnt2 >= un32DstSize) return;
    if(pucSrcBuff[un32Cnt1] & 0x80) pucDstBuff[un32Cnt2] = 1;
    un32Cnt2++; if(un32Cnt2 >= un32DstSize) return;
  }
}

} // namespace vge
