/** \file DefiniteAreaFractals.h */
// ############################################################################################
//  DefiniteAreaFractals.h
//  Definition of CDefiniteAreaFractal derivative classes
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  Some of the algorithms used in this file were written based on the book :
//    Pemrograman Fraktal Resolusi Tinggi untuk Kartu Trident SVGA
//    Ir. P. Insap Santosa, M.Sc. & Ir. R. Tory K.
//    Penerbit ANDI Yogyakarta
//    Indonesia, 1997
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _2D_MANIPULATION___FRACTAL___DEFINITE_AREA_FRACTALS_H_
#define _2D_MANIPULATION___FRACTAL___DEFINITE_AREA_FRACTALS_H_

// VortexGE headers
#include "BasicFractals.h"


// Foward reference from <complex>
namespace std { template<typename T> class complex; }

namespace vge {

  // Definite area fractal default paramaters
  const uint32_t INDEFINITE_AREA_FRACTAL_DEFAULT_ITERATION = uint32_t(64); //!< Default maximum iteration


  //!
  //! \brief The base of complex-number-based fractal classes
  //!
  //! \api experiment
  class VGE_EXPORT CCNBFractal : public CDefiniteAreaFractal {
    public:
      CCNBFractal();
      virtual ~CCNBFractal() throw();

      // Window selection funtions
      //!
      //! Use this to select the coordinates of the cartesian window to be used for iterating (drawing) the equation.
      void SelectWindow(float fX1, float fY1, float fX2, float fY2);
      //!
      //! Use this to select the default coordinates of the cartesian window.
      virtual void SelectDefaultWindow() = 0;

      // Seed selection funtions
      //!
      //! Use this to select the seed for the equation.
      void SelectSeed(float fPv, float fQv);
      //!
      //! Use this to select the default seed for the equation.
      virtual void SelectDefaultSeed() = 0;

      // Drawing functions
      //!
      //! Derivable drawing function for seed-replicator fractals.
      //! \param[in] pcsDstIBuild                Destination CImageBuilder instance.
      //! \param[in] (n32X1,n32Y1)-(n32X2,n32Y2) Where shall the fractal be mapped (drawn) on the screen?
      //!
      //! \note All coordinates are in standard screen coordinates.
      virtual void Draw(CImageBuilder *pcsDstIBuild, int32_t n32X1, int32_t n32Y1, int32_t n32X2, int32_t n32Y2);

      // Drawing callback functions
      //!
      //! Override this callback to follow the progress of the drawing.
      //! \param[in] n32MaxX,n32MaxY Maximum coordinates to be drawn.
      //! \param[in] n32CurX,n32CurY The coordinates currently drawn.
      //!
      //! \note All coordinates are in standard screen coordinates.
      virtual bool DrawingProgressCallback(int32_t n32MaxX, int32_t n32CurX, int32_t n32MaxY, int32_t n32CurY);

  protected:
      // Equation iteration functions
      //!
      //! Override this to iterate the specific equation at the given point.
      //! Will be called for each pixel to be drawn in order to iterate the equation for that pixel.\n
      //! \n
      //! \note All coordinates are in cartesian space.
      virtual uint32_t IterateEqu(float fX, float fY, uint32_t un32MaxIter) = 0;

      // Parameter-control functions
      //!
      //! Override this to control the maximum allowed iteration.
      //! Will be called only once to determine the maximum iteration.
      virtual uint32_t EquMaxIteration();
      //!
      //! Override this callback to control the color of the pixel to be drawn.
      //! Will be called for each pixel to be drawn in order to update the color.
      virtual void UpdateColor(uint32_t un32MaxIter, uint32_t un32Iter, float fColorMulFactor);

      // Drawing functions
      //! Drawing function implementation for complex-number-based fractals.\n
      //! \n
      //! \note All coordinates are in standard screen coordinates.
      void DrawImpl(int32_t n32X1, int32_t n32Y1, int32_t n32X2, int32_t n32Y2, uint32_t un32MaxIter);

      // Data
      struct {
        float fX1, fY1, fX2, fY2; // The cartesian window coordinates
        float fPv, fQv;           // The complex number seed
      } _tW;
      bool _bCWInitialized;     // Flag to indicate that the cartesian window coordinates have ever been initialized
      bool _bCSInitialized;     // Flag to indicate that the complex number seed has ever been initialized
  };


  //!
  //! \brief The base of transcendental-function-based fractal classes
  //!
  //! \api experiment
  class VGE_EXPORT CTFBFractal : public CCNBFractal {
    public:
      // Seed selection funtions
      //!
      //! Use this to select the default seed for the equation.
      virtual void SelectDefaultSeed() {};

    protected:
      // Equation iteration functions (all coordinates are in cartesian space)
      //!
      //! Override this to calculate the equation with the given point.\n
      //! \n
      //! \note All coordinates are in cartesian space.\n
      //! \note This function will be called for each pixel to be drawn in order to iterate the equation on that pixel.
      virtual void CalculateEqu(complex<float> &tZ, const complex<float> &tC) = 0;
      //!
      //! Override this to iterate the specific equation at the given point.\n
      //! \n
      //! \note All coordinates are in cartesian space.\n
      //! \note This function will be called for each pixel to be drawn in order to iterate the equation on that pixel.
      virtual uint32_t IterateEqu(float fX, float fY, uint32_t un32MaxIter);
  };


  //!
  //! \brief The base of pixel-color-interpolator fractal classes
  //!
  //! \api experiment
  class VGE_EXPORT CPCIFractal : public CDefiniteAreaFractal {
    public:
      // Drawing functions
      //!
      //! Derivable drawing function for seed-replicator fractals.
      //! \param[in] pcsDstIBuild                Destination CImageBuilder instance.
      //! \param[in] (n32X1,n32Y1)-(n32X2,n32Y2) Where shall the fractal be mapped (drawn) on the screen?
      //!
      //! \note All coordinates are in standard screen coordinates.
      virtual void Draw(CImageBuilder *pcsDstIBuild, int32_t n32X1, int32_t n32Y1, int32_t n32X2, int32_t n32Y2);

      // Drawing callback functions
      //!
      //! Override this callback to follow the progress of the drawing.
      //! \param[in] (n32X1,n32Y1)-(n32X2,n32Y2) The rectangle area currently drawn.
      //!
      //! \note All coordinates are in standard screen coordinates.
      virtual bool DrawingProgressCallback(int32_t n32X1, int32_t n32Y1, int32_t n32X2, int32_t n32Y2);

    protected:
      //! Drawing function implementation for pixel-color-interpolator fractals.\n
      //! \n
      //! \note All coordinates are in standard screen coordinates.
      bool DrawImpl(int32_t n32X1, int32_t n32Y1, int32_t n32X2, int32_t n32Y2, bool bDrawCenterPixel, bool bFirstCall = false);

      // Parameter-control functions (all coordinates are in standard screen coordinates)
      //!
      //! Override this to enable/disable the drawing of center pixels.
      //! Will be called once to determine wether the center pixels should be drawn or not.
      virtual bool DrawCenterPixel()
      { return(true); }
      //!
      //! Override this to draw the custom seed.
      //! Will be called only once to draw the seed pixels.
      //! \param[out] (n32X1,n32Y1)-(n32X2,n32Y2) Rectangle specifying the are for drawing the seed.
      virtual void DrawSeed(int32_t n32X1, int32_t n32Y1, int32_t n32X2, int32_t n32Y2) = 0;

      //!
      //! Override this callback to set the random factor when drawing pixels.
      //! Will be called for each pixel to be drawn in order to get the random factor for that pixel.
      //! \param[in] n32Distance A factor, can be used for calculating the random factor.
      virtual uint8_t RandomFactor(int32_t n32Distance);
      //!
      //! Override this callback to calculate the color for the pixels.
      //! Will be called for each pixel to be drawn in order to calculate the new pixel color.
      //! \param[out] un8Rx,un8Gx,un8Bx Return the calculated color from here.
      //! \param[in]  un8R1,un8G1,un8B1 Reference color #1.
      //! \param[in]  un8R2,un8G2,un8B2 Reference color #2.
      virtual void  CalculateColor(uint8_t &un8Rx, uint8_t &un8Gx, uint8_t &un8Bx, uint8_t un8R1, uint8_t un8G1, uint8_t un8B1, uint8_t un8R2, uint8_t un8G2, uint8_t un8B2);
      //!
      //! Override this callback to randomize the color for the pixels.
      //! will be called for each pixel to be drawn in order to randomize the new pixel color.
      //! \param[in,out] un8Rx,un8Gx,un8Bx Get the initial color and return the randomized color from here.
      //! \param[in]     un8Factor         A factor, can be used when randomizing.
      virtual void  RandomizeColor(uint8_t &un8R, uint8_t &un8G, uint8_t &un8B, uint8_t un8Factor);
  };

} // namespace vge
#endif
