// ############################################################################################
//  ImageBuilder6.cc
//  Implementation of CImageBuilder class (6th part)
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "../VideoEnvironment/VideoEnvironment.h"
#include "ImageBuilder.h"

// C++ headers
#include <fstream>

// 3rd party headers
#include <zlib.h>


namespace vge {

//
// Compression type
//
#define _VIB_NONE_         uint8_t(EICM_None)
#define _VIB_ZLIB_DEFLATE_ uint8_t(EICM_ZlibDeflate)

//
// VIB header
//
#define _VIB_ID1_ 'V'
#define _VIB_ID2_ 'I'
#define _VIB_ID3_ 'B'
#define _VIB_VER_ 111
#define _VIB_OV1_ 100
#define _VIB_OV2_ 110

struct _TVIBHeader100 {        // VIB 1.00 header
  char     acID[3];            // File ID, must be "VIB"
  uint8_t  un8Version;         // File version, must be 100
  uint32_t un32ImageWidth;     // Image width in pixels
  uint32_t un32ImageHeight;    // Image height in pixels
  uint8_t  un8Compression;     // Compression type
  uint32_t un32SizeCompressed; // Compressed image size in bytes
  uint32_t un32SizePacked;     // Packed image size in bytes
  uint8_t  un8Reserved1;       // Reserved for VIB 1.00, must be zero
  uint32_t un32Reserved2;      // Reserved for VIB 1.00, must be zero
} _attr_packed;

struct _TVIBHeader110 {         // VIB 1.10 and 1.11 header
  char      acID[3];            // File ID, must be "VIB"
  uint8_t   un8Version;         // File version, must be 110 or 111
  uint32_t  un32ImageWidth;     // Image width in pixels
  uint32_t  un32ImageHeight;    // Image height in pixels
  uint8_t   un8Compression;     // Compression type
  uint32_t  un32SizeCompressed; // Compressed image size in bytes
  uint32_t  un32SizePacked;     // Packed image size in bytes
  _TImgInfo tImgInfo;           // Image information structure
} _attr_packed;


//
// External image helper functions (defined inside "lib/2DManioulation/ImageBuilder1.cc"
//
extern void _OrderedDitherRGB24ToVRGB_(TUN8RGB *buffer, uint32_t width, uint32_t height);
extern void _ClipRGB24ToVRGB_(TUN8RGB *buffer, uint32_t width, uint32_t height);

//
// Local image helper functions (defined inside "lib/2DManioulation/ImageBuilder1.cc"
//
static uint32_t _PackImage_(u_char *pucDstBuff, const u_char*pucSrcBuff, uint32_t un32SrcSize);
static bool     _UnpackImage_(u_char *pucDstBuff, const u_char *pucSrcBuff, uint32_t un32DstSize, uint32_t un32SrcSize);
static void     _PackBytes_(u_char aucDst[5], const u_char aucSrc[8]);
static void     _UnpackBytes_(u_char aucDst[8], const u_char aucSrc[5]);
static uint32_t _CompressImageBuffer_(u_char *pucDstBuff, const u_char *pucSrcBuff, uint32_t un32SrcBuffSize, EICMode eCompression = EICM_None);
static uint32_t _UncompressImageBuffer_(u_char *pucDstBuff, uint32_t un32DstBuffSize, const u_char *pucSrcBuff, uint32_t un32SrcBuffSize, EICMode eCompression = EICM_None);


//
// I/O functions
//
void CImageBuilder::LoadRawRGB(const CDataSource &source, uint32_t un32Width, uint32_t un32Height, bool bUseDithering)
{
  // Get the stream
  const char *strmName = source.Name().c_str();
  pistream    &is      = source.PStream();

  // Allocate memory for loading image
  TUN8RGB *ptTmp = reinterpret_cast<TUN8RGB*>(malloc(sizeof(TUN8RGB) * un32Width * un32Height));
  if(!ptTmp) VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for loading bitmap");

  // Load the image
  TUN8RGB *ptPut = ptTmp;
  for(uint32_t un32CntY = 0; un32CntY < un32Height; un32CntY++) {
    for(uint32_t un32CntX = 0; un32CntX < un32Width; un32CntX++) {
      // Get the pixel data
      uint8_t px[3];
      is.read(reinterpret_cast<char*>(px), 3);
      if(is.fail()) {
        free(ptTmp);
        VError(VERR_ReadStream, _NFL_, "Couldn't read stream \"%s\"", strmName);
      }
      // Put the pixel
      ptPut->un8R = px[0];
      ptPut->un8G = px[1];
      ptPut->un8B = px[2];
      ptPut++;
    }
  }

  // Convert the color format
  if(bUseDithering)
    _OrderedDitherRGB24ToVRGB_(ptTmp, un32Width, un32Height);
  else
    _ClipRGB24ToVRGB_(ptTmp, un32Width, un32Height);

  // Destroy old image, if any
  FreeImage();

  // Copy variables
  _ptBuffer   = ptTmp;
  _un32Width  = un32Width;
  _un32Height = un32Height;
}

void CImageBuilder::SaveRawRGB(const CDataSink &sink) const
{
  // Check
  if(_ptBuffer == NULL)
    VError(VERR_NotInitialized, _NFL_, "No data to be saved (this CImageBuilder is not initialized)");
  if(_tImgInfo.un8Flag != _IMG_TYPE_RGB5_)
    VError(VERR_InvalidCall, _NFL_, "Can't save a non _IMG_TYPE_RGB5_ image");

  // Get the stream
  const char *strmName = sink.Name().c_str();
  ostream    &os       = sink.Stream();

  // Save the image
  for(uint32_t un32Cnt = 0; un32Cnt < (_un32Width * _un32Height); un32Cnt++) {
    uint8_t px[3] = { _ptBuffer[un32Cnt].un8R << 3,
                      _ptBuffer[un32Cnt].un8G << 3,
                      _ptBuffer[un32Cnt].un8B << 3 };
    os.write(reinterpret_cast<const char*>(px), 3);
    if(os.fail()) VError(VERR_WriteStream, _NFL_, "Couldn't write stream \"%s\"", strmName);
  }
}

void CImageBuilder::LoadVIB(const CDataSource &source)
{
  // Get the stream
  const char *strmName = source.Name().c_str();
  pistream    &is      = source.PStream();

  // Peek header
  _TVIBHeader110 tHeader;
  is.peek(reinterpret_cast<char*>(&tHeader), sizeof(tHeader));
  if(is.fail())
    VError(VERR_ReadStream, _NFL_, "Couldn't read VIB file header data from stream \"%s\"", strmName);  

  // Check header
  if((tHeader.acID[0] != _VIB_ID1_) || 
     (tHeader.acID[1] != _VIB_ID2_) || 
     (tHeader.acID[2] != _VIB_ID3_))
    VError(VERR_InvalidStream, _NFL_, "\"%s\" is an invalid VIB stream", strmName);
  if(tHeader.un8Version == _VIB_OV1_) {
    __LoadVIB100(source);
    return;
  }
  if(tHeader.un8Version == _VIB_OV2_) {
    __LoadVIB110(source);
    return;
  }
  if(tHeader.un8Version != _VIB_VER_)
    VError(VERR_IncorrectVersion, _NFL_, "\"%s\" has an incorrect version", strmName);
  if((tHeader.un8Compression != _VIB_NONE_) && 
     (tHeader.un8Compression != _VIB_ZLIB_DEFLATE_))
    VError(VERR_InvalidHeaderData, _NFL_, "Invalid compression type found in stream \"%s\"", strmName);

  // Discard the already processed header data from the stream
  is.discard(sizeof(tHeader));

  // Calculate the image size
  uint32_t un32Size = tHeader.un32ImageWidth * tHeader.un32ImageHeight * 3;

  // Allocate memory for temporary buffer
  u_char *pucLoad = reinterpret_cast<u_char*>(malloc(tHeader.un32SizeCompressed));
  if(!pucLoad)
    VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for temporary image buffer");

  u_char *pucPacked = NULL;
  if(tHeader.un8Compression != _VIB_NONE_) {
    pucPacked = reinterpret_cast<u_char*>(malloc(tHeader.un32SizePacked));
    if(!pucPacked) {
      free(pucLoad);
      VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for temporary image buffer");
    }
  }

  u_char *pucUnpacked = reinterpret_cast<u_char*>(malloc(un32Size));
  if(!pucUnpacked) {
    free(pucLoad);
    if(pucPacked) free(pucPacked);
    VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for temporary image buffer");
  }

  // Read image data
  is.read(reinterpret_cast<char*>(pucLoad), tHeader.un32SizeCompressed);
  if(is.fail()) {
    free(pucLoad);
    if(pucPacked) free(pucPacked);
    free(pucUnpacked);
    VError(VERR_ReadStream, _NFL_, "Couldn't read image data from stream \"%s\"", strmName);  
  }

  // Uncompress image
  if(tHeader.un8Compression == _VIB_NONE_)
    pucPacked = pucLoad;
  else {
    if(_UncompressImageBuffer_(pucPacked, tHeader.un32SizePacked,
                               pucLoad, tHeader.un32SizeCompressed,
                               EICMode(tHeader.un8Compression)) != tHeader.un32SizePacked) {
      free(pucLoad);
      if(pucPacked) free(pucPacked);
      free(pucUnpacked);
      VError(VERR_UncompressData, _NFL_, "Couldn't uncompress image data");
    }
  }

  // Unpack
  if((tHeader.tImgInfo.un8Flag == _IMG_TYPE_UDCT_RGB5_) ||
     (tHeader.tImgInfo.un8Flag == _IMG_TYPE_CDCT_RGB5_))
    memcpy(pucUnpacked, pucPacked, tHeader.un32SizePacked);
  else if(tHeader.tImgInfo.un8Flag == _IMG_TYPE_YV12_) {
    if(!_UnpackImage_(pucUnpacked, pucPacked, tHeader.tImgInfo.un32OW, tHeader.un32SizePacked)) {
      free(pucLoad);
      if(pucPacked) free(pucPacked);
      free(pucUnpacked);
      VError(VERR_UnpackData, _NFL_, "Couldn't unpack image data");
    }
  }
  else {
    if(!_UnpackImage_(pucUnpacked, pucPacked, un32Size, tHeader.un32SizePacked)) {
      free(pucLoad);
      if(pucPacked) free(pucPacked);
      free(pucUnpacked);
      VError(VERR_UnpackData, _NFL_, "Couldn't unpack image data");
    }
  }

  // Free memory
  if(pucPacked) free(pucPacked);
  if(pucLoad != pucPacked) free(pucLoad);

  // Destroy previous image, if any
  FreeImage();

  // Copy variables
  _un32Width       = tHeader.un32ImageWidth;
  _un32Height      = tHeader.un32ImageHeight;
  _ptBuffer        = reinterpret_cast<TUN8RGB*>(pucUnpacked);
  memcpy(&_tImgInfo, &tHeader.tImgInfo, sizeof(tHeader.tImgInfo));
}

void CImageBuilder::SaveVIB(const CDataSink &sink, EICMode eCompression) const
{
  // Check
  if(_ptBuffer == NULL)
    VError(VERR_NotInitialized, _NFL_, "No data to be saved (this CImageBuilder is not initialized)");

  // Allocate memory for temporary buffers
  vector<u_char> pucPacked(_un32Width * _un32Height * 3);
  vector<u_char> pucCompressed(_un32Width * _un32Height * 3);

  // Pack image
  int32_t un32PackedSize = 0;
  if((_tImgInfo.un8Flag == _IMG_TYPE_UDCT_RGB5_) ||
     (_tImgInfo.un8Flag == _IMG_TYPE_CDCT_RGB5_)) {
    un32PackedSize = _un32Width * _un32Height * 3;
    copy_b(_ptBuffer, _un32Width * _un32Height * 3, pucPacked);
  }
  else if(_tImgInfo.un8Flag == _IMG_TYPE_YV12_)
    un32PackedSize = _PackImage_(&pucPacked[0], reinterpret_cast<u_char*>(_ptBuffer), _tImgInfo.un32OW);
  else
    un32PackedSize = _PackImage_(&pucPacked[0], reinterpret_cast<u_char*>(_ptBuffer), _un32Width * _un32Height * 3);

  // Compress image
  uint32_t un32CompressedSize = _CompressImageBuffer_(&pucCompressed[0], &pucPacked[0], un32PackedSize, eCompression);

  // Make header
  _TVIBHeader110 tHeader;
  memset(&tHeader, 0, sizeof(tHeader));
  tHeader.acID[0]         = _VIB_ID1_;
  tHeader.acID[1]         = _VIB_ID2_;
  tHeader.acID[2]         = _VIB_ID3_;
  tHeader.un8Version      = _VIB_VER_;
  tHeader.un32ImageWidth  = _un32Width;
  tHeader.un32ImageHeight = _un32Height;
  tHeader.un32SizePacked  = un32PackedSize;
  if(un32CompressedSize) {
    tHeader.un8Compression     = uint8_t(eCompression);
    tHeader.un32SizeCompressed = un32CompressedSize;
  }
  else {
    tHeader.un8Compression     = _VIB_NONE_;
    tHeader.un32SizeCompressed = un32PackedSize;
  }
  memcpy(&tHeader.tImgInfo, &_tImgInfo, sizeof(_tImgInfo));

  // Get the stream
  const char *strmName = sink.Name().c_str();
  ostream    &os       = sink.Stream();

  // Write header
  os.write(reinterpret_cast<char*>(&tHeader), sizeof(tHeader));
  if(os.fail())
    VError(VERR_WriteStream, _NFL_, "Couldn't write VIB file header data to stream \"%s\"", strmName);  

  // Write image data
  if(un32CompressedSize)
    os.write(reinterpret_cast<char*>(&pucCompressed[0]), un32CompressedSize);
  else
    os.write(reinterpret_cast<char*>(&pucPacked[0]), un32PackedSize);
  if(os.fail())
    VError(VERR_WriteStream, _NFL_, "Couldn't write image data to stream \"%s\"", strmName);  
}

void CImageBuilder::__LoadVIB100(const CDataSource &source)
{
  // Get the stream
  const char *strmName = source.Name().c_str();
  pistream    &is      = source.PStream();

  // Read header
  _TVIBHeader100 tHeader;
  is.read(reinterpret_cast<char*>(&tHeader), sizeof(tHeader));
  if(is.fail())
    VError(VERR_ReadStream, _NFL_, "Couldn't read VIB file header data from stream \"%s\"", strmName);  

  // Check header
  if((tHeader.acID[0] != _VIB_ID1_) || 
     (tHeader.acID[1] != _VIB_ID2_) || 
     (tHeader.acID[2] != _VIB_ID3_))
    VError(VERR_InvalidStream, _NFL_, "\"%s\" is an invalid VIB stream", strmName);
  if(tHeader.un8Version != _VIB_OV1_)
    VError(VERR_IncorrectVersion, _NFL_, "\"%s\" has an incorrect version", strmName);
  if(tHeader.un8Compression != _VIB_NONE_)
    VError(VERR_InvalidHeaderData, _NFL_, "Invalid compression type found in stream \"%s\"", strmName);
  if(tHeader.un8Reserved1 || tHeader.un32Reserved2)
    VError(VERR_InvalidHeaderData, _NFL_, "Invalid header data found in stream \"%s\"", strmName);
  
  // Calculate the image size
  uint32_t un32Size = tHeader.un32ImageWidth * tHeader.un32ImageHeight * 3;

  // Allocate memory for temporary buffer
  u_char *pucLoad = reinterpret_cast<u_char*>(malloc(tHeader.un32SizeCompressed));
  if(!pucLoad)
    VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for temporary image buffer");

  u_char *pucUnpacked = reinterpret_cast<u_char*>(malloc(un32Size));
  if(!pucUnpacked) {
    free(pucLoad);
    VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for temporary image buffer");
  }

  // Read image data
  is.read(reinterpret_cast<char*>(pucLoad), tHeader.un32SizeCompressed);
  if(is.fail()) {
    free(pucLoad);
    free(pucUnpacked);
    VError(VERR_ReadStream, _NFL_, "Couldn't read image data from stream \"%s\"", strmName);  
  }

  // Unpack
  if(!_UnpackImage_(pucUnpacked, pucLoad, un32Size, tHeader.un32SizePacked)) {
    free(pucLoad);
    free(pucUnpacked);
    VError(VERR_UnpackData, _NFL_, "Couldn't unpack image data");
  }

  // Free memory
  free(pucLoad);

  // Destroy previous image, if any
  FreeImage();

  // Copy variables
  _un32Width  = tHeader.un32ImageWidth;
  _un32Height = tHeader.un32ImageHeight;
  _ptBuffer   = reinterpret_cast<TUN8RGB*>(pucUnpacked);
}

void CImageBuilder::__SaveVIB100(const CDataSink &sink) const
{
  // Check
  if(_ptBuffer == NULL)
    VError(VERR_NotInitialized, _NFL_, "No data to be saved (this CImageBuilder is not initialized)");
  if(_tImgInfo.un8Flag != _IMG_TYPE_RGB5_)
    VError(VERR_NotSupported, _NFL_, "Current image format can't be saved with VIB version 1.00");

  // Allocate memory for temporary buffers
  vector<u_char> pucPacked(_un32Width * _un32Height * 3);

  // Pack image
  uint32_t un32PackedSize = _PackImage_(&pucPacked[0], reinterpret_cast<u_char*>(_ptBuffer), _un32Width * _un32Height * 3);

  // Make header
  _TVIBHeader100 tHeader;
  memset(&tHeader, 0, sizeof(tHeader));
  tHeader.acID[0]            = _VIB_ID1_;
  tHeader.acID[1]            = _VIB_ID2_;
  tHeader.acID[2]            = _VIB_ID3_;
  tHeader.un8Version         = _VIB_OV1_;
  tHeader.un32ImageWidth     = _un32Width;
  tHeader.un32ImageHeight    = _un32Height;
  tHeader.un32SizePacked     = un32PackedSize;
  tHeader.un8Compression     = _VIB_NONE_;
  tHeader.un32SizeCompressed = un32PackedSize;

  // Get the stream
  const char *strmName = sink.Name().c_str();
  ostream    &os       = sink.Stream();

  // Write header
  os.write(reinterpret_cast<char*>(&tHeader), sizeof(tHeader));
  if(os.fail())
    VError(VERR_WriteStream, _NFL_, "Couldn't write VIB file header data to stream \"%s\"", strmName);  

  // Write image data
  os.write(reinterpret_cast<char*>(&pucPacked[0]), un32PackedSize);
  if(os.fail())
    VError(VERR_WriteStream, _NFL_, "Couldn't write image data to stream \"%s\"", strmName);  
}

void CImageBuilder::__LoadVIB110(const CDataSource &source)
{
  // Get the stream
  const char *strmName = source.Name().c_str();
  pistream    &is      = source.PStream();

  // Peek header
  _TVIBHeader110 tHeader;
  is.peek(reinterpret_cast<char*>(&tHeader), sizeof(tHeader));
  if(is.fail())
    VError(VERR_ReadStream, _NFL_, "Couldn't read VIB file header data from stream \"%s\"", strmName);  

  // Check header
  if((tHeader.acID[0] != _VIB_ID1_) ||
     (tHeader.acID[1] != _VIB_ID2_) ||
     (tHeader.acID[2] != _VIB_ID3_))
    VError(VERR_InvalidStream, _NFL_, "\"%s\" is an invalid VIB stream", strmName);
  if(tHeader.un8Version == _VIB_OV1_) {
    __LoadVIB100(source);
    return;
  }
  if(tHeader.un8Version != _VIB_OV2_)
    VError(VERR_IncorrectVersion, _NFL_, "\"%s\" has an incorrect version", strmName);
  if((tHeader.un8Compression != _VIB_NONE_) && 
     (tHeader.un8Compression != _VIB_ZLIB_DEFLATE_))
    VError(VERR_InvalidHeaderData, _NFL_, "Invalid compression type found in stream \"%s\"", strmName);

  // Discard the already processed header data from the stream
  is.discard(sizeof(tHeader));

  // Calculate the image size
  uint32_t un32Size = tHeader.un32ImageWidth * tHeader.un32ImageHeight * 3;

  // Allocate memory for temporary buffer
  u_char *pucLoad = reinterpret_cast<u_char*>(malloc(tHeader.un32SizeCompressed));
  if(!pucLoad)
    VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for temporary image buffer");

  u_char *pucPacked = NULL;
  if(tHeader.un8Compression != _VIB_NONE_) {
    pucPacked = reinterpret_cast<u_char*>(malloc(tHeader.un32SizePacked));
    if(!pucPacked) {
      free(pucLoad);
      VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for temporary image buffer");
    }
  }

  u_char *pucUnpacked = reinterpret_cast<u_char*>(malloc(un32Size));
  if(!pucUnpacked) {
    free(pucLoad);
    if(pucPacked) free(pucPacked);
    VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for temporary image buffer");
  }

  // Read image data
  is.read(reinterpret_cast<char*>(pucLoad), tHeader.un32SizeCompressed);
  if(is.fail()) {
    free(pucLoad);
    if(pucPacked) free(pucPacked);
    free(pucUnpacked);
    VError(VERR_ReadStream, _NFL_, "Couldn't read image data from stream \"%s\"", strmName);  
  }

  // Uncompress image
  if(tHeader.un8Compression == _VIB_NONE_)
    pucPacked = pucLoad;
  else {
    if(_UncompressImageBuffer_(pucPacked, tHeader.un32SizePacked,
                               pucLoad, tHeader.un32SizeCompressed,
                               EICMode(tHeader.un8Compression)) != tHeader.un32SizePacked) {
      free(pucLoad);
      if(pucPacked) free(pucPacked);
      free(pucUnpacked);
      VError(VERR_UncompressData, _NFL_, "Couldn't uncompress image data");
    }
  }

  // Unpack
  if(!tHeader.tImgInfo.un8Flag) {
    if(!_UnpackImage_(pucUnpacked, pucPacked, un32Size, tHeader.un32SizePacked)) {
      if(pucPacked) free(pucPacked);
      free(pucUnpacked);
      free(pucLoad);
      VError(VERR_UnpackData, _NFL_, "Couldn't unpack image data");
    }
  }
  else
    memcpy(pucUnpacked, pucPacked, tHeader.un32SizePacked);

  // Free memory
  if(pucPacked) free(pucPacked);
  if(pucLoad != pucPacked) free(pucLoad);

  // Destroy previous image, if any
  FreeImage();

  // Copy variables
  _un32Width       = tHeader.un32ImageWidth;
  _un32Height      = tHeader.un32ImageHeight;
  _ptBuffer        = reinterpret_cast<TUN8RGB*>(pucUnpacked);
  memcpy(&_tImgInfo, &tHeader.tImgInfo, sizeof(tHeader.tImgInfo));
}

void CImageBuilder::__SaveVIB110(const CDataSink &sink, EICMode eCompression) const
{
  // Check
  if(_ptBuffer == NULL)
    VError(VERR_NotInitialized, _NFL_, "No data to be saved (this CImageBuilder is not initialized)");
  if((_tImgInfo.un8Flag != _IMG_TYPE_RGB5_) &&
     (_tImgInfo.un8Flag != _IMG_TYPE_UDCT_RGB5_) &&
     (_tImgInfo.un8Flag != _IMG_TYPE_CDCT_RGB5_) &&
     (_tImgInfo.un8Flag != _IMG_TYPE_FYUV_))
    VError(VERR_NotSupported, _NFL_, "Current image format can't be saved with VIB version 1.10");

  // Allocate memory for temporary buffers
  vector<u_char> pucPacked(_un32Width * _un32Height * 3);
  vector<u_char> pucCompressed(_un32Width * _un32Height * 3);

  // Pack image
  uint32_t un32PackedSize = 0;
  if(!_tImgInfo.un8Flag)
    un32PackedSize = _PackImage_(&pucPacked[0], reinterpret_cast<u_char*>(_ptBuffer), _un32Width * _un32Height * 3);
  else {
    un32PackedSize = _un32Width * _un32Height * 3;
    copy_b(_ptBuffer, _un32Width * _un32Height * 3, pucPacked);
  }

  // Compress image
  uint32_t un32CompressedSize = _CompressImageBuffer_(&pucCompressed[0], &pucPacked[0], un32PackedSize, eCompression);

  // Make header
  _TVIBHeader110 tHeader;
  memset(&tHeader, 0, sizeof(tHeader));
  tHeader.acID[0]         = _VIB_ID1_;
  tHeader.acID[1]         = _VIB_ID2_;
  tHeader.acID[2]         = _VIB_ID3_;
  tHeader.un8Version      = _VIB_OV2_;
  tHeader.un32ImageWidth  = _un32Width;
  tHeader.un32ImageHeight = _un32Height;
  tHeader.un32SizePacked  = un32PackedSize;
  if(un32CompressedSize) {
    tHeader.un8Compression     = uint8_t(eCompression);
    tHeader.un32SizeCompressed = un32CompressedSize;
  }
  else {
    tHeader.un8Compression     = _VIB_NONE_;
    tHeader.un32SizeCompressed = un32PackedSize;
  }
  if(_tImgInfo.un8Flag) memcpy(&tHeader.tImgInfo, &_tImgInfo, sizeof(_tImgInfo));

  // Get the stream
  const char *strmName = sink.Name().c_str();
  ostream    &os       = sink.Stream();

  // Write header
  os.write(reinterpret_cast<char*>(&tHeader), sizeof(tHeader));
  if(os.fail())
    VError(VERR_WriteStream, _NFL_, "Couldn't write VIB file header data to stream \"%s\"", strmName);  

  // Write image data
  if(un32CompressedSize)
    os.write(reinterpret_cast<char*>(&pucCompressed[0]), un32CompressedSize);
  else
    os.write(reinterpret_cast<char*>(&pucPacked[0]), un32PackedSize);
  if(os.fail())
    VError(VERR_WriteStream, _NFL_, "Couldn't write image data to stream \"%s\"", strmName);  
}


//
// Local functions implementation
//
inline bool _GetByte_(u_char &pucDst, u_char *&pucSrc, u_char *&pucSrcMax)
{
  if(pucSrc > pucSrcMax) {
    pucDst = 0;
    return(false);
  }
  pucDst = *pucSrc++;
  return(true);
}

static uint32_t _PackImage_(u_char *pucDstBuff, const u_char *pucSrcBuff, uint32_t un32SrcSize)
{
  // Temporary variables
  uint32_t un32Size;
  u_char  aucP[5], aucU[8];
  u_char  *pucDstTmp = pucDstBuff;
  u_char  *pucSrcTmp = const_cast<u_char*>(pucSrcBuff);
  u_char  *pucSrcMax = pucSrcTmp + un32SrcSize - 1;
  bool    bRun;

  // Pack image
  un32Size = 0; bRun = true;
  while(bRun) {
    for(uint32_t i = 0; i < 8; i++) bRun = _GetByte_(aucU[i], pucSrcTmp, pucSrcMax);
    _PackBytes_(aucP, aucU);
    for(uint32_t i = 0; i < 5; i++) *pucDstTmp++ = aucP[i];
    un32Size += 5;
  }
  return(un32Size);
}

static bool _UnpackImage_(u_char *pucDstBuff, const u_char *pucSrcBuff, uint32_t un32DstSize, uint32_t un32SrcSize)
{
  // Temporary variables
  uint32_t un32Size, un32Cnt;
  u_char  aucP[5], aucU[8];
  u_char  *pucDstTmp = pucDstBuff;
  u_char  *pucSrcTmp = const_cast<u_char*>(pucSrcBuff);
  u_char  *pucSrcMax = pucSrcTmp + un32SrcSize - 1;
  bool   bRun;

  // Unpack image
  un32Size = 0; bRun = true;
  while(bRun) {
    for(uint32_t i = 0; i < 5; i++) bRun = _GetByte_(aucP[i], pucSrcTmp, pucSrcMax);
    _UnpackBytes_(aucU, aucP);
    for(un32Cnt = 0; un32Cnt < 8; un32Cnt++) {
      *pucDstTmp++ = aucU[un32Cnt];
      un32Size++;
      if(un32Size >= un32DstSize) return(true);
    }
  }
  return(false);
}

static void _PackBytes_(u_char aucDst[5], const u_char aucSrc[8])
{
  aucDst[0] = (aucSrc[0] << 3) | (aucSrc[1] >> 2);
  aucDst[1] = (aucSrc[1] << 6) | (aucSrc[2] << 1) | (aucSrc[3] >> 4);
  aucDst[2] = (aucSrc[3] << 4) | (aucSrc[4] >> 1);
  aucDst[3] = (aucSrc[4] << 7) | (aucSrc[5] << 2) | (aucSrc[6] >> 3);
  aucDst[4] = (aucSrc[6] << 5) |  aucSrc[7];
}

static void _UnpackBytes_(u_char aucDst[8], const u_char aucSrc[5])
{
  aucDst[0] =   aucSrc[0] >> 3;
  aucDst[1] = ((aucSrc[0] << 2) & 0x1F) | (aucSrc[1] >> 6);
  aucDst[2] =  (aucSrc[1] >> 1) & 0x1F;
  aucDst[3] = ((aucSrc[1] << 4) & 0x1F) | (aucSrc[2] >> 4);
  aucDst[4] = ((aucSrc[2] << 1) & 0x1F) | (aucSrc[3] >> 7);
  aucDst[5] =  (aucSrc[3] >> 2) & 0x1F;
  aucDst[6] = ((aucSrc[3] << 3) & 0x1F) | (aucSrc[4] >> 5);
  aucDst[7] =   aucSrc[4]       & 0x1F;
}

static uint32_t _CompressImageBuffer_(u_char *pucDstBuff, const u_char *pucSrcBuff, uint32_t un32SrcBuffSize, EICMode eCompression)
{
  // No compression
  if(eCompression == EICM_None)
    return(0);
  // ZLIB_DEFLATE
  else if(eCompression == EICM_ZlibDeflate) {
    uLongf ulLen = un32SrcBuffSize;
    if(compress2(reinterpret_cast<Bytef*>(pucDstBuff), &ulLen,
                 reinterpret_cast<Bytef*>(const_cast<u_char*>(pucSrcBuff)),
                 static_cast<uLong>(un32SrcBuffSize), 9) == Z_OK) return(ulLen);
      VWarning(_NFL_, "ZLIB_DEFLATE failed compressing the image\nThe uncompressed image is kept");
    return(0);
  }

  // Unknown compression mode
  VWarning(_NFL_, "Unknown compression mode\nImage will not be compressed");
  return(0);
}

static uint32_t _UncompressImageBuffer_(u_char *pucDstBuff, uint32_t un32DstBuffSize, const u_char *pucSrcBuff, uint32_t un32SrcBuffSize, EICMode eCompression)
{
  // No compression
  if(eCompression == EICM_None)
    return(0);
  // ZLIB_DEFLATE
  else if(eCompression == EICM_ZlibDeflate) {
    uLongf ulLen = un32DstBuffSize;
    if(uncompress(reinterpret_cast<Bytef*>(pucDstBuff), &ulLen,
                  reinterpret_cast<Bytef*>(const_cast<u_char*>(pucSrcBuff)),
                  static_cast<uLong>(un32SrcBuffSize)) == Z_OK) return(ulLen);
    return(0);
  }

  // Unknown compression mode
  VWarning(_NFL_, "Unknown compression mode\nThe loaded image may contain an invalid image data");
  return(0);
}

} // namespace vge
