// ############################################################################################
//  FoggedCeilingLineSegment.cc
//  Impelementation of fogged ceiling-line-segment classes
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  Some of the source codes in this file were written based on the book :
//    Cutting-Edge 3D Game Programming with C++
//    John De Goes
//    Coriolis Group, Inc.
//    USA, 1996
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "FixedPointConsts.h"
#include "FoggedCeilingLineSegment.h"


namespace vge {

//
// The fogged-untextured-flat-shaded ceiling-line-segment class
//
_CFogUFCLineSeg::_CFogUFCLineSeg()
#ifdef __USE_AVEC__
: _n32X1   (0), _n32Y1   (0), _n32X2   (0), _n32Y2   (0),
  _n32X    (0), _n32Y    (0), _n32Z    (0), _n32F    (0),
  _n32XStep(0), _n32YStep(1), _n32ZStep(0), _n32FStep(0), _n32EdgeHeight(0)
#else
: _n32X1   (0), _n32Y1   (0), _n32X2   (0), _n32Y2   (0),
  _n32X    (0), _n32Y    (0), _n32Z    (0), _n32F    (0),
  _n32XStep(0),               _n32ZStep(0), _n32FStep(0), _n32EdgeHeight(0)
#endif
{
}

void _CFogUFCLineSeg::Init(const _TPoint2D &tP1, const _TPoint2D &tP2)
{
  // Calculate the initial values for polygon edge
  _n32X1 = tP1.n32X; _n32Y1 = tP1.n32Y;
  _n32X2 = tP2.n32X; _n32Y2 = tP2.n32Y;

  _n32EdgeHeight = _n32Y2 - _n32Y1;

  _n32X = (_n32X1 << _XSTEP_PREC_) + _XCEIL_FRAC_;
  _n32Y = _n32Y1;
  _n32Z = tP1.n32Z + C3DEnv::Active3DEnv()->_tRB->n32ZTrans;
  _n32F = tP1.n32F;

  if(_n32EdgeHeight) {
    _n32XStep = ((_n32X2 - _n32X1) << _XSTEP_PREC_) / _n32EdgeHeight;
    _n32ZStep = (tP2.n32Z - tP1.n32Z) / _n32EdgeHeight;
    _n32FStep = (tP2.n32F - tP1.n32F) / _n32EdgeHeight;
  }
  else {
    _n32XStep = 0;
    _n32ZStep = 0;
    _n32FStep = 0;
  }
}

void _CFogUFCLineSeg::ClipTop(const int32_t n32Top)
{
  // Clip the polygon edge againts the top of the viewport,
  // must be called directly after calling Init()
  const int32_t n32Height = _n32Y2 - _n32Y1;
  if((_n32Y < n32Top) && (n32Height)) {
    const int32_t n32Step = n32Top - _n32Y;
    const float   fXSlope = float(_n32X2 - _n32X1) / float(n32Height);

    _n32X = (_n32X1 << _XSTEP_PREC_) +
             (int32_t(fXSlope * float(n32Step)) << _XSTEP_PREC_) +
             _XCEIL_FRAC_;
    _n32Y = n32Top;
    _n32Z += (_n32ZStep * n32Step);
    _n32F += (_n32FStep * n32Step);
    _n32EdgeHeight -= n32Step;
  }
}


//
// The fogged-untextured-gouraud-shaded ceiling-line-segment class
//
_CFogUGCLineSeg::_CFogUGCLineSeg()
#ifdef __USE_AVEC__
: _n32X1   (0), _n32Y1   (0), _n32X2   (0), _n32Y2   (0),
  _n32X    (0), _n32Y    (0), _n32Z    (0), _n32F    (0),
  _n32R    (0), _n32G    (0), _n32B    (0), ___dummy1(1),
  _n32XStep(0), _n32YStep(1), _n32ZStep(0), _n32FStep(0),
  _n32RStep(0), _n32GStep(0), _n32BStep(0), ___dummy2(1), _n32EdgeHeight(0)
#else
: _n32X1   (0), _n32Y1   (0), _n32X2   (0), _n32Y2   (0),
  _n32X    (0), _n32Y    (0), _n32Z    (0), _n32F    (0),
  _n32R    (0), _n32G    (0), _n32B    (0),
  _n32XStep(0),               _n32ZStep(0), _n32FStep(0),
  _n32RStep(0), _n32GStep(0), _n32BStep(0),               _n32EdgeHeight(0)
#endif
{
}

void _CFogUGCLineSeg::Init(const _TPoint2D &tP1, const _TPoint2D &tP2)
{
  // Calculate the initial values for polygon edge
  _n32X1 = tP1.n32X; _n32Y1 = tP1.n32Y;
  _n32X2 = tP2.n32X; _n32Y2 = tP2.n32Y;

  _n32EdgeHeight = _n32Y2 - _n32Y1;

  _n32X = (_n32X1 << _XSTEP_PREC_) + _XCEIL_FRAC_;
  _n32Y = _n32Y1;
  _n32Z = tP1.n32Z + C3DEnv::Active3DEnv()->_tRB->n32ZTrans;
  _n32F = tP1.n32F;
  _n32R = tP1.n32R;
  _n32G = tP1.n32G;
  _n32B = tP1.n32B;

  if(_n32EdgeHeight) {
    _n32XStep = ((_n32X2 - _n32X1) << _XSTEP_PREC_) / _n32EdgeHeight;
    _n32ZStep = (tP2.n32Z - tP1.n32Z) / _n32EdgeHeight;
    _n32FStep = (tP2.n32F - tP1.n32F) / _n32EdgeHeight;
    _n32RStep = (tP2.n32R - tP1.n32R) / _n32EdgeHeight;
    _n32GStep = (tP2.n32G - tP1.n32G) / _n32EdgeHeight;
    _n32BStep = (tP2.n32B - tP1.n32B) / _n32EdgeHeight;
  }
  else {
    _n32XStep = 0;
    _n32ZStep = 0;
    _n32FStep = 0;
    _n32RStep = 0;
    _n32GStep = 0;
    _n32BStep = 0;
  }
}

void _CFogUGCLineSeg::ClipTop(const int32_t n32Top)
{
  // Clip the polygon edge againts the top of the viewport,
  // must be called directly after calling Init()
  const int32_t n32Height = _n32Y2 - _n32Y1;
  if((_n32Y < n32Top) && (n32Height)) {
    const int32_t n32Step = n32Top - _n32Y;
    const float   fXSlope = float(_n32X2 - _n32X1) / float(n32Height);

    _n32X = (_n32X1 << _XSTEP_PREC_) +
             (int32_t(fXSlope * float(n32Step)) << _XSTEP_PREC_) +
             _XCEIL_FRAC_;
    _n32Y = n32Top;
    _n32Z += (_n32ZStep * n32Step);
    _n32F += (_n32FStep * n32Step);
    _n32R += (_n32RStep * n32Step);
    _n32G += (_n32GStep * n32Step);
    _n32B += (_n32BStep * n32Step);
    _n32EdgeHeight -= n32Step;
  }
}


//
// The fogged-textured-flat-shaded ceiling-line-segment class
//
_CFogTFCLineSeg::_CFogTFCLineSeg()
#ifdef __USE_AVEC__
: _n32X1   (0), _n32Y1   (0), _n32X2   (0), _n32Y2   (0),
  _n32X    (0), _n32Y    (0), _n32Z    (0), _n32F    (0),
  _n32U    (0), _n32V    (0), ___dummy1(1), ___dummy2(1),
  _n32XStep(0), _n32YStep(1), _n32ZStep(0), _n32FStep(0),
  _n32UStep(0), _n32VStep(0), ___dummy3(1), ___dummy4(1), _n32EdgeHeight(0)
#else
: _n32X1   (0), _n32Y1   (0), _n32X2   (0), _n32Y2   (0),
  _n32X    (0), _n32Y    (0), _n32Z    (0), _n32F    (0),
  _n32U    (0), _n32V    (0),
  _n32XStep(0),               _n32ZStep(0), _n32FStep(0),
  _n32UStep(0), _n32VStep(0),                             _n32EdgeHeight(0)
#endif
{
}

void _CFogTFCLineSeg::Init(const _TPoint2D &tP1, const _TPoint2D &tP2)
{
  // Calculate the initial values for polygon edge
  _n32X1 = tP1.n32X; _n32Y1 = tP1.n32Y;
  _n32X2 = tP2.n32X; _n32Y2 = tP2.n32Y;

  _n32EdgeHeight = _n32Y2 - _n32Y1;

  _n32X = (_n32X1 << _XSTEP_PREC_) + _XCEIL_FRAC_;
  _n32Y = _n32Y1;
  _n32Z = tP1.n32Z + C3DEnv::Active3DEnv()->_tRB->n32ZTrans;
  _n32F = tP1.n32F;
  _n32U = tP1.n32U;
  _n32V = tP1.n32V;

  if(_n32EdgeHeight) {
    _n32XStep = ((_n32X2 - _n32X1) << _XSTEP_PREC_) / _n32EdgeHeight;
#ifdef __USE_AVEC__
    //VPrint("%p %p %p %p\n", &tP1.n32Z, &tP1.n32F, &tP1.n32U, &tP1.n32V);
    const v4si  &p1 = *I32_TO_V4SI(&tP1.n32Z);
    const v4si  &p2 = *I32_TO_V4SI(&tP2.n32Z);
    const v4si   pd = MAKE_V4SI(_n32EdgeHeight);
    const v4si_u ps( (p2-p1) / pd );
    _n32ZStep = ps.i[0]; // NOTE: the newly calculated Z step is in the lowest
    _n32FStep = ps.i[1]; //       memory address because the n32Z member of the
    _n32UStep = ps.i[2]; //       _TPoint2D structure is in the lower memory
    _n32VStep = ps.i[3]; //       address compared to the n32F, n32U and n32V
                         //       members
#else
    _n32ZStep = (tP2.n32Z - tP1.n32Z) / _n32EdgeHeight;
    _n32FStep = (tP2.n32F - tP1.n32F) / _n32EdgeHeight;
    _n32UStep = (tP2.n32U - tP1.n32U) / _n32EdgeHeight;
    _n32VStep = (tP2.n32V - tP1.n32V) / _n32EdgeHeight;
#endif
  }
  else {
    _n32XStep = 0;
    _n32ZStep = 0;
    _n32FStep = 0;
    _n32UStep = 0;
    _n32VStep = 0;
  }
}

void _CFogTFCLineSeg::ClipTop(const int32_t n32Top)
{
  // Clip the polygon edge againts the top of the viewport,
  // must be called directly after calling Init()
  const int32_t n32Height = _n32Y2 - _n32Y1;
  if((_n32Y < n32Top) && (n32Height)) {
    const int32_t n32Step = n32Top - _n32Y;
    const float   fXSlope = float(_n32X2 - _n32X1) / float(n32Height);

    _n32X = (_n32X1 << _XSTEP_PREC_) +
             (int32_t(fXSlope * float(n32Step)) << _XSTEP_PREC_) +
             _XCEIL_FRAC_;
    _n32Y = n32Top;
#ifdef __USE_AVEC__
    const v4si   mf = MAKE_V4SI(n32Step);
    const v4si   cv = MAKE_V4SI(_n32Z, _n32F, _n32U, _n32V);
    const v4si   sv = MAKE_V4SI(_n32ZStep, _n32FStep, _n32UStep, _n32VStep);
    const v4si_u nv( cv + sv*mf );
    _n32Z = nv.i[3]; // NOTE: the newly calculated Z is in the highest memory
    _n32F = nv.i[2]; //       address because MAKE_V4SI() always makes the first
    _n32U = nv.i[1]; //       function parameter as the left most elements of
    _n32V = nv.i[0]; //       the v4si vector
#else
    _n32Z += (_n32ZStep * n32Step);
    _n32F += (_n32FStep * n32Step);
    _n32U += (_n32UStep * n32Step);
    _n32V += (_n32VStep * n32Step);
#endif
    _n32EdgeHeight -= n32Step;
  }
}


//
// The fogged-textured-gouraud-shaded ceiling-line-segment class
//
_CFogTGCLineSeg::_CFogTGCLineSeg()
#ifdef __USE_AVEC__
: _n32X1   (0), _n32Y1   (0), _n32X2   (0), _n32Y2   (0),
  _n32X    (0), _n32Y    (0), _n32Z    (0), _n32F    (0),
  _n32U    (0), _n32V    (0), ___dummy1(1), ___dummy2(1),
  _n32R    (0), _n32G    (0), _n32B    (0), ___dummy3(1),
  _n32XStep(0), _n32YStep(1), _n32ZStep(0), _n32FStep(0),
  _n32UStep(0), _n32VStep(0), ___dummy4(1), ___dummy5(1),
  _n32RStep(0), _n32GStep(0), _n32BStep(0), ___dummy6(1), _n32EdgeHeight(0)
#else
: _n32X1   (0), _n32Y1   (0), _n32X2   (0), _n32Y2   (0),
  _n32X    (0), _n32Y    (0), _n32Z    (0), _n32F    (0),
  _n32U    (0), _n32V    (0),
  _n32R    (0), _n32G    (0), _n32B    (0),
  _n32XStep(0),               _n32ZStep(0), _n32FStep(0),
  _n32UStep(0), _n32VStep(0),
  _n32RStep(0), _n32GStep(0), _n32BStep(0),               _n32EdgeHeight(0)
#endif
{
}

void _CFogTGCLineSeg::Init(const _TPoint2D &tP1, const _TPoint2D &tP2)
{
  // Calculate the initial values for polygon edge
  _n32X1 = tP1.n32X; _n32Y1 = tP1.n32Y;
  _n32X2 = tP2.n32X; _n32Y2 = tP2.n32Y;

  _n32EdgeHeight = _n32Y2 - _n32Y1;

  _n32X = (_n32X1 << _XSTEP_PREC_) + _XCEIL_FRAC_;
  _n32Y = _n32Y1;
  _n32Z = tP1.n32Z + C3DEnv::Active3DEnv()->_tRB->n32ZTrans;
  _n32F = tP1.n32F;
  _n32U = tP1.n32U;
  _n32V = tP1.n32V;
  _n32R = tP1.n32R;
  _n32G = tP1.n32G;
  _n32B = tP1.n32B;

  if(_n32EdgeHeight) {
    _n32XStep = ((_n32X2 - _n32X1) << _XSTEP_PREC_) / _n32EdgeHeight;
    _n32ZStep = (tP2.n32Z - tP1.n32Z) / _n32EdgeHeight;
    _n32FStep = (tP2.n32F - tP1.n32F) / _n32EdgeHeight;
    _n32UStep = (tP2.n32U - tP1.n32U) / _n32EdgeHeight;
    _n32VStep = (tP2.n32V - tP1.n32V) / _n32EdgeHeight;
    _n32RStep = (tP2.n32R - tP1.n32R) / _n32EdgeHeight;
    _n32GStep = (tP2.n32G - tP1.n32G) / _n32EdgeHeight;
    _n32BStep = (tP2.n32B - tP1.n32B) / _n32EdgeHeight;
  }
  else {
    _n32XStep = 0;
    _n32ZStep = 0;
    _n32FStep = 0;
    _n32UStep = 0;
    _n32VStep = 0;
    _n32RStep = 0;
    _n32GStep = 0;
    _n32BStep = 0;
  }
}

void _CFogTGCLineSeg::ClipTop(const int32_t n32Top)
{
  // Clip the polygon edge againts the top of the viewport,
  // must be called directly after calling Init()
  const int32_t n32Height = _n32Y2 - _n32Y1;
  if((_n32Y < n32Top) && (n32Height)) {
    const int32_t n32Step = n32Top - _n32Y;
    const float   fXSlope = float(_n32X2 - _n32X1) / float(n32Height);

    _n32X = (_n32X1 << _XSTEP_PREC_) +
             (int32_t(fXSlope * float(n32Step)) << _XSTEP_PREC_) +
             _XCEIL_FRAC_;
    _n32Y = n32Top;
    _n32Z += (_n32ZStep * n32Step);
    _n32F += (_n32FStep * n32Step);
    _n32U += (_n32UStep * n32Step);
    _n32V += (_n32VStep * n32Step);
    _n32R += (_n32RStep * n32Step);
    _n32G += (_n32GStep * n32Step);
    _n32B += (_n32BStep * n32Step);
    _n32EdgeHeight -= n32Step;
  }
}

} // namespace vge
