/** \file Panel3D.h */
// ############################################################################################
//  Panel3D.h
//  Definition & implementation of CPanel3D class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  Some of the source codes in this file were written based on the book :
//    Cutting-Edge 3D Game Programming with C++
//    John De Goes
//    Coriolis Group, Inc.
//    USA, 1996
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _3D_ENVIRONMENT___PANEL_3D_H_
#define _3D_ENVIRONMENT___PANEL_3D_H_

// VortexGE headers
#include "../3DEnvironment/3DEnvironment.h"
#include "Matrix3D.h"


namespace vge {

  // Foward reference from "lib/2DEnvironment/ImageClasses.h"
  class CTexture;
  class CAnimatedTexture;


  //!
  //! \brief The 3D panel class
  //!
  //! One shouldn't need to use the clipping, projection and rendering
  //! functions of this class directly (since one will use them indirectly
  //! from within the CPanel3DBuffer class).
  //!
  //! \api pretty stable
  class VGE_EXPORT CPanel3D : VCLASS_UNCOPYABLE {
    public:
      inline CPanel3D()
      { memset(&_tP, 0, sizeof(_tP)); }

      // Texture coordinates functions
      //!
      //! Set the texture coordinates of this panel using the given parameter.
      void SetTextureCoordinates(TTexUV atUV[4]);
      //!
      //! Set the texture coordinates of this panel using the given parameters.
      void SetTextureCoordinates(float fU1, float fV1, float fU2, float fV2, float fU3, float fV3, float fU4, float fV4);

      // Clipping functions
      //!
      //! Check if the panel is visible in the 2D (screen) space.
      //! \n\n \sntc
      bool IsVisible2D();
      //!
      //! Check if the panel is visible in the 3D space.
      //! \n\n \sntc
      inline bool IsVisible3D();
      //!
      //! Check if the panel is totally invisible.
      //! \n\n \sntc
      inline uint32_t Invisible() const
      { return(_tP.un32Invisible); }
      //!
      //! Must be called on each frame update.
      //! \n\n \sntc
      inline void UpdateFrame()
      { if(_tP.un32Invisible) --_tP.un32Invisible; }
    
      // Normal projection functions
      //!
      //! Project the panel's points for untextured/uncolored, flat shaded rendering.
      //! \n\n \sntc
      void UFProject();
      //!
      //! Project the panel's points for untextured/uncolored, gouraud shaded rendering.
      //! \n\n \sntc
      void UGProject();
      //!
      //! Project the panel's points for textured, not/flat shaded rendering.
      //! \n\n \sntc
      void TFProject(const CTexture &csTexture); 
      //!
      //! Project the panel's points for textured, gouraud shaded rendering.
      //! \n\n \sntc
      void TGProject(const CTexture &csTexture); 

      // Fogged projection functions
      //!
      //! Project the panel's points for fogged, untextured/uncolored, flat shaded rendering.
      //! \n\n \sntc
      void FogUFProject();
      //!
      //! Project the panel's points for fogged, untextured/uncolored, gouraud shaded rendering.
      //! \n\n \sntc
      void FogUGProject();
      //!
      //! Project the panel's for points fogged, textured, not/flat shaded rendering.
      //! \n\n \sntc
      void FogTFProject(const CTexture &csTexture);
      //!
      //! Project the panel's for points fogged, textured, gouraud shaded rendering.
      //! \n\n \sntc
      void FogTGProject(const CTexture &csTexture);

      // Normal rendering functions
      //!
      //! Render the panel: untextured, flat shaded.
      //! \n\n \sntc
      void UFRender();
      //!
      //! Render the panel: untextured, gouraud shaded.
      //! \n\n \sntc
      void UGRender();
      //!
      //! Render the panel: colored, not shaded.
      //! \n\n \sntc
      void CNRender(uint32_t &un32R, uint32_t &un32G, uint32_t &un32B);
      //!
      //! Render the panel: colored, flat shaded.
      //! \n\n \sntc
      void CFRender(uint32_t &un32R, uint32_t &un32G, uint32_t &un32B);
      //!
      //! Render the panel: colored, gouraud shaded.
      //! \n\n \sntc
      void CGRender(uint32_t &un32R, uint32_t &un32G, uint32_t &un32B);
      //!
      //! Render the panel: textured, not shaded.
      //! \n\n \sntc
      void TNRender(const CTexture &csTexture);
      //!
      //! Render the panel: textured, colored.
      //! \n\n \sntc
      void TCRender(const CTexture &csTexture, uint32_t &un32R, uint32_t &un32G, uint32_t &un32B);
      //!
      //! Render the panel: textured, flat shaded.
      //! \n\n \sntc
      void TFRender(const CTexture &csTexture);
      //!
      //! Render the panel: textured, gouraud shaded.
      //! \n\n \sntc
      void TGRender(const CTexture &csTexture);

      // Fogged rendering functions
      //!
      //! Render the panel: fogged, untextured, flat shaded.
      //! \n\n \sntc
      void FogUFRender();
      //!
      //! Render the panel: fogged, untextured, gouraud shaded.
      //! \n\n \sntc
      void FogUGRender();
      //!
      //! Render the panel: fogged, colored, not shaded.
      //! \n\n \sntc
      void FogCNRender(uint32_t &un32R, uint32_t &un32G, uint32_t &un32B);
      //!
      //! Render the panel: fogged, colored, flat shaded.
      //! \n\n \sntc
      void FogCFRender(uint32_t &un32R, uint32_t &un32G, uint32_t &un32B);
      //!
      //! Render the panel: fogged, colored, gouraud shaded.
      //! \n\n \sntc
      void FogCGRender(uint32_t &un32R, uint32_t &un32G, uint32_t &un32B);
      //!
      //! Render the panel: fogged, textured, not shaded.
      //! \n\n \sntc
      void FogTNRender(const CTexture &csTexture);
      //!
      //! Render the panel: fogged, textured, colored.
      //! \n\n \sntc
      void FogTCRender(const CTexture &csTexture, uint32_t &un32R, uint32_t &un32G, uint32_t &un32B);
      //!
      //! Render the panel: fogged, textured, flat shaded.
      //! \n\n \sntc
      void FogTFRender(const CTexture &csTexture);
      //!
      //! Render the panel: fogged, textured, gouraud shaded.
      //! \n\n \sntc
      void FogTGRender(const CTexture &csTexture);

      // Transparent rendering functions
      //!
      //! Render the panel: transparent, colored,  flat shaded.
      //! \n\n \sntc
      void CFTransRender(uint32_t &un32R, uint32_t &un32G, uint32_t &un32B);
      //!
      //! Render the panel: transparent, textured, flat shaded.
      //! \n\n \sntc
      void TFTransRender(const CTexture &csTexture);
      //!
      //! Render the panel: transparent, textured, not shaded.
      //! \n\n \sntc
      void TNTransRender(const CTexture &csTexture);
      //!
      //! Render the panel: transparent, fogged, colored,  flat shaded.
      //! \n\n \sntc
      void FogCFTransRender(uint32_t &un32R, uint32_t &un32G, uint32_t &un32B);
      //!
      //! Render the panel: transparent, fogged, textured, flat shaded.
      //! \n\n \sntc
      void FogTFTransRender(const CTexture &csTexture);
      //!
      //! Render the panel: transparent, fogged, textured, not shaded.
      //! \n\n \sntc
      void FogTNTransRender(const CTexture &csTexture);

      // Support for SSE
      DECLARE_SSE_ENABLED_NEW_AND_DELETE_OPERATORS()

      // Friend classes
      friend class CPanel3DBuffer;

    protected:
      // Panel functions
      void _CalcNormal_();
      void _CalcRadius_();
      // Lighting functions
      void _FSCalcIntenOmni_(const TOmniLight *ptVOmniLight);
      void _GSCalcIntenOmni_(const TOmniLight *ptVOmniLight, bool bUseAvgNormal = false);
      void _DynFSCalcIntenOmni_(const TOmniLight *ptVOmniLight);
      void _DynGSCalcIntenOmni_(const TOmniLight *ptVOmniLight);
      inline void _CalcIntenAmbient_(const TAmbientLight *ptVAmbientLight);
      inline void _FSCalcInten_();
      void _GSCalcInten_();
      // Clipping functions
      bool _CalcBFace_();
      bool _CheckExtents_();
      inline bool _HasVertex_(const CPoint3D *pcsVPoint3D) const;

      // Data
      struct _TPanelData {
        CPoint3D  *apcsVPoint[4]; // Pointers to vertex list
        TTexUV     atUV[4];       // The texture coordinates
        _TPoint2D  atSPoint[5];   // Maximum 5 on-screen points
        _TVector3D tNormal;       // The panel normal
        _TVector3D tBNormal;      // The panel normal (for backface culling)
        float      afCenter[3];   // The center of panel (in local coordinates)
        float      fRadius;       // The panel's radius
        uint8_t    un8SPCount;    // Screen point count
        TUN8RGB    tAmbient;      // The panel's color (from ambient light)
        TUN8RGB    tOmni;         // The panel's color (from omni light)
        TUN8RGB    tTotal;        // The panel's color (total)
        TUN8RGB    atOmni[4];     // The vertex's color (from omni light)
        TUN8RGB    atTotal[4];    // The vertex's color (total)
        uint32_t   un32Invisible; // Invisible flag
      } _tP;
  };

  void CPanel3D::_CalcIntenAmbient_(const TAmbientLight *ptVAmbientLight)
  {
    // Use the supplied ambient light
    if(ptVAmbientLight) {
      _tP.tAmbient.un8R = ptVAmbientLight->un8R;
      _tP.tAmbient.un8G = ptVAmbientLight->un8G;
      _tP.tAmbient.un8B = ptVAmbientLight->un8B;
    }
    // Use the default ambient light from the active 3D environment
    else {
      _tP.tAmbient.un8R = C3DEnv::Active3DEnv()->_tDefALight.un8R;
      _tP.tAmbient.un8G = C3DEnv::Active3DEnv()->_tDefALight.un8G;
      _tP.tAmbient.un8B = C3DEnv::Active3DEnv()->_tDefALight.un8B;
    }
  }

  void CPanel3D::_FSCalcInten_()
  {
    _tP.tTotal.un8R = (_tP.tAmbient.un8R + _tP.tOmni.un8R) >> 1;
    _tP.tTotal.un8G = (_tP.tAmbient.un8G + _tP.tOmni.un8G) >> 1;
    _tP.tTotal.un8B = (_tP.tAmbient.un8B + _tP.tOmni.un8B) >> 1;
  }

  bool CPanel3D::IsVisible3D()
  {
    if(_CalcBFace_()) return(_CheckExtents_());
    return(false);
  }

  bool CPanel3D::_HasVertex_(const CPoint3D *pcsVPoint3D) const
  {
    if(_tP.apcsVPoint[0] == pcsVPoint3D) return(true);
    if(_tP.apcsVPoint[1] == pcsVPoint3D) return(true);
    if(_tP.apcsVPoint[2] == pcsVPoint3D) return(true);
    if(_tP.apcsVPoint[3] == pcsVPoint3D) return(true);
    return(false);
  }

} // namespace vge
#endif
