/** \file Panel3DBuffer.h */
// ############################################################################################
//  Panel3DBuffer.h
//  Definition & implementation of CPanel3DBuffer class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  Some of the source codes in this file were written based on the book :
//    Cutting-Edge 3D Game Programming with C++
//    John De Goes
//    Coriolis Group, Inc.
//    USA, 1996
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _3D_ENVIRONMENT___PANEL_3D_BUFFER_H_
#define _3D_ENVIRONMENT___PANEL_3D_BUFFER_H_

// VortexGE headers
#include "../System/DataSinkSource.h"
#include "Panel3D.h"
#include "Point3D.h"


namespace vge {

  //!
  //! \brief Collision detection result
  //!
  //! \api pretty stable
  enum ECollType { ECT_None, ECT_Floor, ECT_Ceiling, ECT_Wall };

  //!
  //! \brief Local region translation mode
  //!
  //! \api pretty stable
  enum ELRTMode {
    ELRTM_Original, //!< Do not translate the coordinates.
    ELRTM_Balanced, //!< Balance all the coordinates in the middle of the negative and positive quadrant.
    ELRTM_Negative, //!< Make all the coordinates become in the negative quadrant.
    ELRTM_Positive  //!< Make all the coordinates become in the positive quadrant.
  };


  //!
  //! \brief The 3D panel buffer class
  //!
  //! \api pretty stable
  class VGE_EXPORT CPanel3DBuffer : VCLASS_UNCOPYABLE {
    public:
      CPanel3DBuffer();
      ~CPanel3DBuffer() throw();

      // Lighting functions
      //!
      //! (Re-)set the custom ambient light to be used when rendering
      //! with flat shading.
      void FSResetAmbientLight(TAmbientLight *ptVAmbientLight = NULL);
      //!
      //! (Re-)set the custom ambient light to be used when rendering
      //! with gouraud shading.
      void GSResetAmbientLight(TAmbientLight *ptVAmbientLight = NULL);
      //!
      //! (Re-)set the custom omni light to be used when rendering
      //! with flat shading.
      void FSResetOmniLight(TOmniLight *ptVOmniLight = NULL);
      //!
      //! (Re-)set the custom omni light to be used when rendering
      //! with gouraud shading.
      void GSResetOmniLight(TOmniLight *ptVOmniLight = NULL);
      //!
      //! (Re-)set the custom ambient and omni lights to be used when rendering
      //! with flat shading.
      void FSResetLights(TAmbientLight *ptVAmbientLight = NULL, TOmniLight *ptVOmniLight = NULL);
      //!
      //! (Re-)set the custom ambient and omni lights to be used when rendering
      //! with gouraud shading.
      void GSResetLights(TAmbientLight *ptVAmbientLight = NULL, TOmniLight *ptVOmniLight = NULL);
      //!
      //! (Re-)set the custom ambient light to be used when rendering
      //! with flat and gouraud shadings.
      void FSGSResetAmbientLight(TAmbientLight *ptVAmbientLight = NULL);
      //!
      //! (Re-)set the custom omni light to be used when rendering
      //! with flat and gouraud shadings.
      void FSGSResetOmniLight(TOmniLight *ptVOmniLight = NULL);
      //!
      //! (Re-)set the custom ambient and omni lights to be used when rendering
      //! with flat and gouraud shadings.
      void FSGSResetLights(TAmbientLight *ptVAmbientLight = NULL, TOmniLight *ptVOmniLight = NULL);

      // Texturing functions
      //!
      //! Set the color to be used when rendering.
      inline void SetColor(uint8_t un8R, uint8_t un8G, uint8_t un8B);
      //!
      //! Set the texture to be used when rendering.
      inline void SetTexture(CTexture *pcsTexture)
       { _pcsTexture = pcsTexture; }
      //!
      //! Set the animated texture to be used when rendering.
      inline void SetTexture(CAnimatedTexture *pcsAnimatedTexture)
      { _pcsTexture = reinterpret_cast<CTexture*>(pcsAnimatedTexture); }
      //!
      //! Get the current texture which will be used when rendering.
      inline CTexture *GetTexture() const 
      { return(_pcsTexture); }
      //!
      //! Get the width of the current texture.
      inline uint32_t GetTextureWidth() const;
      //!
      //! Get the width of the current texture.
      inline uint32_t GetTextureHeight() const;
      //!
      //! Calculate default texture coordinates for all panels.
      void CalcDefTexCoords();

      // Transformation functions
      //!
      //! Transform the vertices using the given matrix.
      //! \n\n
      //! \sntc
      void Transform(const CMatrix3D &csTMatrix);
      //!
      //! Dynamic transform the vertices using the given matrix.
      //! \n\n
      //! \sntc
      void DynTransform(const CMatrix3D &csCameraMatrix);

      // Pre-dynamic rendering functions
      //!
      //! In dynamic rendering, this function must be called before actually
      //! calling the dynamic rendering functions without shading.
      //! \param[in] csDynMatrix      Dynamic matrix that usually comes from CHNode::Matrix().
      //! \param[in] ptVOmniLightUsed Custom omni light to be used when rendering.
      void PreDynNSRender(const CMatrix3D &csDynMatrix, const TOmniLight *ptVOmniLightUsed = NULL);
      //!
      //! In dynamic rendering, this function must be called before actually
      //! calling the dynamic rendering functions with flat shading.
      //! \param[in] csDynMatrix      Dynamic matrix that usually comes from CHNode::Matrix().
      //! \param[in] ptVOmniLightUsed Custom omni light to be used when rendering.
      void PreDynFSRender(const CMatrix3D &csDynMatrix, const TOmniLight *ptVOmniLightUsed = NULL);
      //!
      //! In dynamic rendering, this function must be called before actually
      //! calling the dynamic rendering functions with gouraud shading.
      //! \param[in] csDynMatrix      Dynamic matrix that usually comes from CHNode::Matrix().
      //! \param[in] ptVOmniLightUsed Custom omni light to be used when rendering.
      void PreDynGSRender(const CMatrix3D &csDynMatrix, const TOmniLight *ptVOmniLightUsed = NULL);

      // Normal rendering functions
      //!
      //! Render the panels: untextured, flat shaded.
      void UFRender(const CMatrix3D &csTMatrix);
      //!
      //! Render the panels: untextured, gouraud shaded.
      void UGRender(const CMatrix3D &csTMatrix);
      //!
      //! Render the panels: colored, not shaded.
      void CNRender(const CMatrix3D &csTMatrix);
      //!
      //! Render the panels: colored, flat shaded.
      void CFRender(const CMatrix3D &csTMatrix);
      //!
      //! Render the panels: colored, gouraud shaded.
      void CGRender(const CMatrix3D &csTMatrix);
      //!
      //! Render the panels: textured, not shaded.
      void TNRender(const CMatrix3D &csTMatrix);
      //!
      //! Render the panels: textured, colored.
      void TCRender(const CMatrix3D &csTMatrix);
      //!
      //! Render the panels: textured, flat shaded.
      void TFRender(const CMatrix3D &csTMatrix);
      //!
      //! Render the panels: textured, gouraud shaded.
      void TGRender(const CMatrix3D &csTMatrix);
      //!
      //! Dynamic render the panels: untextured, flat shaded.
      void DynUFRender(const CMatrix3D &csTMatrix);
      //!
      //! Dynamic render the panels: untextured, gouraud shaded.
      void DynUGRender(const CMatrix3D &csTMatrix);
      //!
      //! Dynamic render the panels: colored, flat shaded.
      void DynCFRender(const CMatrix3D &csTMatrix);
      //!
      //! Dynamic render the panels: colored, gouraud shaded.
      void DynCGRender(const CMatrix3D &csTMatrix);
      //!
      //! Dynamic render the panels: textured, not shaded.
      void DynTNRender(const CMatrix3D &csTMatrix);
      //!
      //! Dynamic render the panels: textured, flat shaded.
      void DynTFRender(const CMatrix3D &csTMatrix);
      //!
      //! Dynamic render the panels: textured, gouraud shaded.
      void DynTGRender(const CMatrix3D &csTMatrix);

      // Fogged rendering functions
      //!
      //! Render the panels: fogged, untextured, flat shaded.
      void FogUFRender(const CMatrix3D &csTMatrix);
      //!
      //! Render the panels: fogged, untextured, gouraud shaded.
      void FogUGRender(const CMatrix3D &csTMatrix);
      //!
      //! Render the panels: fogged, colored, not shaded.
      void FogCNRender(const CMatrix3D &csTMatrix);
      //!
      //! Render the panels: fogged, colored, flat shaded.
      void FogCFRender(const CMatrix3D &csTMatrix);
      //!
      //! Render the panels: fogged, colored, gouraud shaded.
      void FogCGRender(const CMatrix3D &csTMatrix);
      //!
      //! Render the panels: fogged, textured, not shaded.
      void FogTNRender(const CMatrix3D &csTMatrix);
      //!
      //! Render the panels: fogged, textured, colored.
      void FogTCRender(const CMatrix3D &csTMatrix);
      //!
      //! Render the panels: fogged, textured, flat shaded.
      void FogTFRender(const CMatrix3D &csTMatrix);
      //!
      //! Render the panels: fogged, textured, gouraud shaded.
      void FogTGRender(const CMatrix3D &csTMatrix);
      //!
      //! Dynamic render the panels: fogged, untextured, flat shaded.
      void DynFogUFRender(const CMatrix3D &csTMatrix);
      //!
      //! Dynamic render the panels: fogged, untextured, gouraud shaded.
      void DynFogUGRender(const CMatrix3D &csTMatrix);
      //!
      //! Dynamic render the panels: fogged, colored, flat shaded.
      void DynFogCFRender(const CMatrix3D &csTMatrix);
      //!
      //! Dynamic render the panels: fogged, colored, gouraud shaded.
      void DynFogCGRender(const CMatrix3D &csTMatrix);
      //!
      //! Dynamic render the panels: fogged, textured, not shaded.
      void DynFogTNRender(const CMatrix3D &csTMatrix);
      //!
      //! Dynamic render the panels: fogged, textured, flat shaded.
      void DynFogTFRender(const CMatrix3D &csTMatrix);
      //!
      //! Dynamic render the panels: fogged, textured, gouraud shaded.
      void DynFogTGRender(const CMatrix3D &csTMatrix);

      // Transparent rendering functions
      //!
      //! Render the panels: transparent, colored, flat shaded.
      void CFTransRender(const CMatrix3D &csTMatrix);
      //!
      //! Render the panels: transparent, textured, flat shaded.
      void TFTransRender(const CMatrix3D &csTMatrix);
      //!
      //! Render the panels: transparent, textured, not shaded.
      void TNTransRender(const CMatrix3D &csTMatrix);
      //!
      //! Render the panels: transparent, fogged, colored, flat shaded.
      void FogCFTransRender(const CMatrix3D &csTMatrix);
      //!
      //! Render the panels: transparent, fogged, textured, flat shaded.
      void FogTFTransRender(const CMatrix3D &csTMatrix);
      //!
      //! Render the panels: transparent, fogged, textured, not shaded.
      void FogTNTransRender(const CMatrix3D &csTMatrix);

      // Collision functions
      //!
      //! Check collision between all panels with a point with the given
      //! position and movement vector.
      ECollType PlaneCollide(float afNewXPos, float afNewYPos, float afNewZPos, float fViewRadius = 10.0f) const;
      //!
      //! Check collision between all panels with a point with the given
      //! position and movement vector.
      inline ECollType PlaneCollide(const float afNewPosXYZ[3], float fViewRadius = 10.0f) const;
      //!
      //! Calculate bounding box of all the panels in 3D world coordinates.
      void GetBoundingBox(TFloatXYZ &minCoord, TFloatXYZ &maxCoord) const;

      // Builder functions
      //!
      //! Build panels from the data given in the parameters.
      //! \param[in] pfVertices                    Pointer to buffer containing the XYZ vertices.
      //! \param[in] un32VertexCount               Number of vertices in the pfVertices buffer.
      //! \param[in] pfPanelPoints                 Pointer to buffer containing the panels' 4 points (indices).
      //! \param[in] un32PanelCount                Number of points (indices) in the pfPanelPoints buffer.
      //! \param[in] bEnableSmoothShading          Set this to false if the panel buffer will never be rendered using gouraud shading (it may speed up loading a bit).
      //! \param[in] bUseAvgNormalForSmoothShading Set this to true to use the average panel's normal to calculate shading (rather than using vertices' normals).
      //! \param[in] ptVAmbientLight               Set this to an instance of TAmbientLight to use it as the ambient light (instead of using the ambient light from the active 3D environment).
      //! \param[in] ptVOmniLight                  Set this to an instance of TOmniLight to use it as the omni light (instead of using the omni light from the active 3D environment).
      void InitFromMemory(const float *pfVertices, uint32_t un32VertexCount,
                          const uint32_t *pfPanelPoints, uint32_t un32PanelCount,
                          bool bEnableSmoothShading = true,
                          bool bUseAvgNormalForSmoothShading = false,
                          const TAmbientLight *ptVAmbientLight = NULL,
                          const TOmniLight *ptVOmniLight = NULL);
      //!
      //! Recalculate the panels's shading using the data given in the parameters.
      //! \param[in] bEnableSmoothShading          Set this to false if the panel buffer will never be rendered using gouraud shading (it may speed up loading a bit).
      //! \param[in] bUseAvgNormalForSmoothShading Set this to true to use the average panel's normal to calculate shading (rather than using vertices' normals).
      //! \param[in] ptVAmbientLight               Set this to an instance of TAmbientLight to use it as the ambient light (instead of using the ambient light from the active 3D environment).
      //! \param[in] ptVOmniLight                  Set this to an instance of TOmniLight to use it as the omni light (instead of using the omni light from the active 3D environment).
      void ReinitShading(bool bEnableSmoothShading = true,
                         bool bUseAvgNormalForSmoothShading = false,
                         const TAmbientLight *ptVAmbientLight = NULL,
                         const TOmniLight *ptVOmniLight = NULL);
      //!
      //! Direct access to the CPanel3D instance with the given index.
      inline CPanel3D *GetPanel(uint32_t un32Index)
      { return(_GetPanel_(un32Index)); }
      //!
      //! Direct access to the CPanel3D instance with the given index.
      inline const CPanel3D *GetPanel(uint32_t un32Index) const
      { return(_GetPanel_(un32Index)); }
      //!
      //! Scale all vertices' local X coordinates by the given amount.
      void LocalXScale(float fXAmount);
      //!
      //! Scale all vertices' local Y coordinates by the given amount.
      void LocalYScale(float fYAmount);
      //!
      //! Scale all vertices' local Z coordinates by the given amount.
      void LocalZScale(float fZAmount);
      //!
      //! Scale all vertices' local XY coordinates by the given amount.
      void LocalXYScale(float fXAmount, float fYAmount);
      //!
      //! Scale all vertices' local XZ coordinates by the given amount.
      void LocalXZScale(float fXAmount, float fZAmount);
      //!
      //! Scale all vertices' local YZ coordinates by the given amount.
      void LocalYZScale(float fYAmount, float fZAmount);
      //!
      //! Scale all vertices' local XYZ coordinates by the given amount.
      void LocalXYZScale(float fXAmount, float fYAmount, float fZAmount);
      //!
      //! Translate all vertices' local X coordinates by the given amount.
      void LocalXTranslate(float fXAmount);
      //!
      //! Translate all vertices' local Y coordinates by the given amount.
      void LocalYTranslate(float fYAmount);
      //!
      //! Translate all vertices' local Z coordinates by the given amount.
      void LocalZTranslate(float fZAmount);
      //!
      //! Translate all vertices' local XY coordinates by the given amount.
      void LocalXYTranslate(float fXAmount, float fYAmount);
      //!
      //! Translate all vertices' local XZ coordinates by the given amount.
      void LocalXZTranslate(float fXAmount, float fZAmount);
      //!
      //! Translate all vertices' local YZ coordinates by the given amount.
      void LocalYZTranslate(float fYAmount, float fZAmount);
      //!
      //! Translate all vertices' local XYZ coordinates by the given amount.
      void LocalXYZTranslate(float fXAmount, float fYAmount, float fZAmount);
      //!
      //! Translate all vertices' local XYZ coordinates using the given mode.
      void LocalRegionTranslate(ELRTMode eXMode, ELRTMode eYMode, ELRTMode eZMode);
      //!
      //! Translate all vertices' local XYZ coordinates using the ELRTM_Balanced mode.
      void LocalRegionTranslateAllBalanced();
      //!
      //! Translate all vertices' local XYZ coordinates so that the center of weight
      //! is in the middle.
      void LocalRegionTranslateCenterOfWeight();

      // I/O functions
      //!
      //! Save to VortexGE's native VPD format.
      //! \param[in] sink The name of the file (.vpd) to be saved to or an instance of ostream derived class.
      void SaveVPD(const CDataSink &sink) const;
      //!
      //! Load from VortexGE's native VPD format.
      //! \param[in] source                        The name of the file (.vpd) to be loaded or an instance of istream derived class.
      //! \param[in] bEnableSmoothShading          Set this to false if the panel buffer will never be rendered using gouraud shading (it may speed up loading a bit).
      //! \param[in] bUseAvgNormalForSmoothShading Set this to true to use the average panel's normal to calculate shading (rather than using vertices' normals).
      //! \param[in] ptVAmbientLight               Set this to an instance of TAmbientLight to use it as the ambient light (instead of using the ambient light from the active 3D environment).
      //! \param[in] ptVOmniLight                  Set this to an instance of TOmniLight to use it as the omni light (instead of using the omni light from the active 3D environment).
      void LoadVPD(const CDataSource &source,
                   bool bEnableSmoothShading = true,
                   bool bUseAvgNormalForSmoothShading = false,
                   TAmbientLight *ptVAmbientLight = NULL,
                   TOmniLight *ptVOmniLight = NULL);

      // Old I/O functions
      //!
      //! Used for saving with older VPD format.
      //! Use only if compatibility with older versions of VortexGE is needed.
      void __SaveVPD100(const CDataSink &sink) const;
      //!
      //! Used for loading from older VPD format.
      //! Shouldn't need to call this function directly since LoadVPD() is able to load from older VIB format.
      void __LoadVPD100(const CDataSource &source);
      //!
      //! Used for saving with older VPD format.
      //! Use only if compatibility with older versions of VortexGE is needed.
      void __SaveVPD110(const CDataSink &sink) const;
      //!
      //! Used for loading from older VPD format.
      //! Shouldn't need to call this function directly since LoadVPD() is able to load from older VIB format.
      void __LoadVPD110(const CDataSource &source);
      //!
      //! Used for saving with older VPD format.
      //! Use only if compatibility with older versions of VortexGE is needed.
      void __SaveVPD111(const CDataSink &sink) const;
      //!
      //! Used for loading from older VPD format.
      //! Shouldn't need to call this function directly since LoadVPD() is able to load from older VIB format.
      void __LoadVPD111(const CDataSource &source, bool bEnableSmoothShading = true, bool bUseAvgNormalForSmoothShading = false, TAmbientLight *ptVAmbientLight = NULL, TOmniLight *ptVOmniLight = NULL);

    protected:
      // Helper functions
      CPanel3D *_GetPanel_(uint32_t un32Index) const;
      void _CalcAvgNormal_(uint32_t un32VCount, uint32_t un32PCount);
      void _FindCoordMinMax_(float &xmin, float &xmax, float &ymin, float &ymax, float &zmin, float &zmax);

      // Data
      CTexture  *_pcsTexture; // The panel's texture
      CPoint3D  *_pcsVList;   // The list of vertices
      CPanel3D  *_pcsPList;   // The list of panels
      uint32_t   _un32VCount; // The vertex count
      uint32_t   _un32PCount; // The panel count
      uint32_t   _un32PCR;    // The R component of the panel's color
      uint32_t   _un32PCG;    // The G component of the panel's color
      uint32_t   _un32PCB;    // The B component of the panel's color
 };

  void CPanel3DBuffer::SetColor(uint8_t un8R, uint8_t un8G, uint8_t un8B)
  {
    _un32PCR = uint32_t(un8R);
    _un32PCG = uint32_t(un8G);
    _un32PCB = uint32_t(un8B);
  }

  uint32_t CPanel3DBuffer::GetTextureWidth() const
  {
    if(_pcsTexture) return(_pcsTexture->_un32Width);
    return(0);
  }

  uint32_t CPanel3DBuffer::GetTextureHeight() const
  {
    if(_pcsTexture) return(_pcsTexture->_un32Height);
    return(0);
  }

  // Collision functions
  ECollType CPanel3DBuffer::PlaneCollide(const float afNewPosXYZ[3], float fViewRadius) const
  {
    return(PlaneCollide(afNewPosXYZ[0], afNewPosXYZ[1], afNewPosXYZ[2], fViewRadius));
  }

} // namespace vge
#endif
