// ############################################################################################
//  Sprite2D.cc
//  Implementation of CSprite2D class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  Some of the source codes in this file were written based on the book :
//    Cutting-Edge 3D Game Programming with C++
//    John De Goes
//    Coriolis Group, Inc.
//    USA, 1996
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "../Base/MessageDisplayer.h"
#include "FixedPointConsts.h"
#include "Sprite2D.h"


namespace vge {

//
// Local inline functions implementation
//
inline static bool _InsideZBoundaries_(float fZ)
{
  if(fZ > C3DEnv::Active3DEnv()->ZMax()) return(false);
  //if(fZ < C3DEnv::Active3DEnv()->ZMin()) return(false);
  if(fZ < 1.0f) return(false); 	//! \bug TEMPORARY HACK!!!
  return(true);
}

inline static void _ProjectSprite_(float &fX, float &fY, float &fW, float &fH,
                                   int32_t &n32L, int32_t &n32R, int32_t &n32T, int32_t &n32B,
                                   float fWX, float fWY, float fWZ, uint32_t un32Width, uint32_t un32Height)
{
    fX = fWX * C3DEnv::Active3DEnv()->XScale() / fWZ + C3DEnv::Active3DEnv()->XShift();
    fY = fWY * C3DEnv::Active3DEnv()->YScale() / fWZ + C3DEnv::Active3DEnv()->YShift();
    fW = float(un32Width)  * C3DEnv::Active3DEnv()->XScale4() / fWZ;
    fH = float(un32Height) * C3DEnv::Active3DEnv()->YScale4() / (-fWZ);
    n32L = int32_t(fX - (fW / 2.0f));
    n32R = int32_t(fX + (fW / 2.0f));
    n32T = int32_t(fY - (fH / 2.0f));
    n32B = int32_t(fY + (fH / 2.0f));
}

inline static bool _Perform2DClip_(int32_t &n32TexLeft, int32_t &n32TexTop,
                                   int32_t &n32L, int32_t &n32R, int32_t &n32T, int32_t &n32B,
                                   uint32_t un32Width, uint32_t un32Height, float fW, float fH)
{
  n32TexLeft = n32TexTop = 0;

  if(n32R < C3DEnv::Active3DEnv()->XMin()) return(false);
  if(n32B < C3DEnv::Active3DEnv()->YMin()) return(false);
  if(n32L > C3DEnv::Active3DEnv()->XMax()) return(false);
  if(n32T > C3DEnv::Active3DEnv()->YMax()) return(false);

  if(n32R > C3DEnv::Active3DEnv()->XMax()) n32R = C3DEnv::Active3DEnv()->XMax();
  if(n32B > C3DEnv::Active3DEnv()->YMax()) n32B = C3DEnv::Active3DEnv()->YMax();
  if(n32L < C3DEnv::Active3DEnv()->XMin()) {
    n32TexLeft = int32_t(float(C3DEnv::Active3DEnv()->XMin() - n32L) * float(un32Width) / fW);
    n32L = int32_t(C3DEnv::Active3DEnv()->XMin());
  }
  if(n32T < C3DEnv::Active3DEnv()->YMin()) {
    n32TexTop = int32_t(float(C3DEnv::Active3DEnv()->YMin() - n32T) * float(un32Height) / fH);
    n32T = int32_t(C3DEnv::Active3DEnv()->YMin());
  }
  return(true);
}

inline static int32_t _CalcZValue_(float fZ)
{
  return(int32_t(1.0f / fZ * float(1 << _ZSTEP_PREC_)) + C3DEnv::Active3DEnv()->ZTrans());
}

inline int32_t _CalcFValue_(float fZ)
{
  int32_t n32Fog;
  n32Fog = int32_t(fZ - C3DEnv::Active3DEnv()->ZMin()) >> C3DEnv::Active3DEnv()->FogShift();
  if(n32Fog < 0)               return(0);
  if(n32Fog > MAX_COLOR_VALUE) return(MAX_COLOR_VALUE << 5);
  return(n32Fog << 5);
}


//
// The 2D sprite class
//
CSprite2D::CSprite2D()
: _ptData(NULL), _un8TotData(0), _pcsImageSprite(NULL), _ptOmni(NULL)
{
}

CSprite2D::~CSprite2D() throw()
{
  delete[] _ptData;
}

void CSprite2D::Initialize(const float *pfPosXYZs, uint8_t un8PosCount, const CImage *pcsImageSprite,
                           EPivot ePivot, uint32_t un32ExcludedColorVRGB,
                           TAmbientLight *ptVAmbientLight, TOmniLight *ptVOmniLight,
                           float fCollisionRadius)
{
  Clear();

  // Allocate memory for data buffer
  _ptData = _nothrow_new _TSSpriteData[un8PosCount];
  if(_ptData == NULL)
    VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for sprite's data buffer");

  // Copy variables
  _pcsImageSprite = pcsImageSprite;
  if(un32ExcludedColorVRGB == EXCLUDED_COLOR_FIRST_PIXEL)
    _un32ExclColor = *(reinterpret_cast<uint32_t*>(_pcsImageSprite->_pun8Buffer));
  else
    _un32ExclColor = un32ExcludedColorVRGB;
  _un8TotData = un8PosCount;
  _ptOmni     = ptVOmniLight;

  // Calculate position offset for pivoting
  float w  = _pcsImageSprite->_un32Width >> 2;
  float h  = _pcsImageSprite->_un32Height >> 2;
  float w2 = w / 2.0f;
  float h2 = h / 2.0f;
  
  _fXo = _fYo = 0.0f;
  if(fCollisionRadius > 0.0f)
    _fRadius = fCollisionRadius;
  else
    _fRadius = sqrt(w2*w2 + h2*h2) * 0.5f;

  if(ePivot == EP_TopLeft) {
    _fXo = w2;
    _fYo = -h2;
  }
  else if(ePivot == EP_TopCenter)
    _fYo = -h2;
  else if(ePivot == EP_TopRight) {
    _fXo = -w2;
    _fYo = -h2;
  }
  else if(ePivot == EP_CenterLeft)
    _fXo = w2;
  else if(ePivot == EP_CenterCenter)
    ;
  else if(ePivot == EP_CenterRight)
    _fXo = -w2;
  else if(ePivot == EP_BottomLeft) {
    _fXo = w2;
    _fYo = h2;
  }
  else if(ePivot == EP_BottomCenter)
    _fYo = h2;
  else if(ePivot == EP_BottomRight) {
    _fXo = -w2;
    _fYo = h2;
  }

  // Set initial properties
  const float *pfTmp = pfPosXYZs;
  for(uint8_t un8Cnt = 0; un8Cnt < un8PosCount; un8Cnt++) {
    // Exclusion flag
    _ptData[un8Cnt].bExcluded = false;

    // Set normal
    _ptData[un8Cnt].fX = 0.0f;
    _ptData[un8Cnt].fY = 0.316228f;
    _ptData[un8Cnt].fZ = 0.948683f;

    // Set position
    _ptData[un8Cnt].csPos._tLocal.fX = (*pfTmp++) + _fXo;
    _ptData[un8Cnt].csPos._tLocal.fY = (*pfTmp++) + _fYo;
    _ptData[un8Cnt].csPos._tLocal.fZ = *pfTmp++;
  }

  // Calc light intensities
  _CalcIntenAmbient_(ptVAmbientLight);
  _CalcIntenOmni_(ptVOmniLight);
  _CalcInten_();
}

void CSprite2D::Clear()
{
  if(!_ptData) return;

  delete[] _ptData;
  _ptData         = NULL;
  _un8TotData     = 0;
  _pcsImageSprite = 0;
  _ptOmni         = NULL;
}

void CSprite2D::_CalcIntenOmni_(TOmniLight *ptVOmniLight)
{
  // Temporary variables
  float fXm, fYm, fZm, fMag, fCos;

  // Calculate omni light intensity
  for(uint8_t un8Cnt = 0; un8Cnt < _un8TotData; un8Cnt++) {
    // Use the supplied omni light
    if(ptVOmniLight) {
      // Create the normalized light vector from the panel center to the omni light position
      fXm = ptVOmniLight->fX - _ptData[un8Cnt].csPos._tLocal.fX;
      fYm = ptVOmniLight->fY - _ptData[un8Cnt].csPos._tLocal.fY;
      fZm = ptVOmniLight->fZ - _ptData[un8Cnt].csPos._tLocal.fZ;
      fMag = sqrt(fXm * fXm + fYm * fYm + fZm * fZm);
      fYm /= fMag;
      fZm /= fMag;

      // Calculate the angle between sprite's normal (fixed) and the light vector
      fCos = _ptData[un8Cnt].fX * fXm + _ptData[un8Cnt].fY * fYm - _ptData[un8Cnt].fZ * fZm;
      fCos = (fCos < 0) ? 0 : fCos;

      // Calculate the color value
      _ptData[un8Cnt].tOmni.un8R = uint8_t(ptVOmniLight->un8R * fCos);
      _ptData[un8Cnt].tOmni.un8G = uint8_t(ptVOmniLight->un8G * fCos);
      _ptData[un8Cnt].tOmni.un8B = uint8_t(ptVOmniLight->un8B * fCos);
    }
    // Use the default omni light from the active 3D environment
    else {
      // Create the normalized light vector from the panel center to the omni light position
      fXm = C3DEnv::Active3DEnv()->_tDefOLight.fX - _ptData[un8Cnt].csPos._tLocal.fX;
      fYm = C3DEnv::Active3DEnv()->_tDefOLight.fY - _ptData[un8Cnt].csPos._tLocal.fY;
      fZm = C3DEnv::Active3DEnv()->_tDefOLight.fZ - _ptData[un8Cnt].csPos._tLocal.fZ;
      fMag = sqrt(fXm * fXm + fYm * fYm + fZm * fZm);
      fYm /= fMag;
      fZm /= fMag;

      // Calculate the angle between sprite's normal (fixed) and the light vector
      fCos = _ptData[un8Cnt].fX * fXm + _ptData[un8Cnt].fY * fYm - _ptData[un8Cnt].fZ * fZm;
      fCos = (fCos < 0) ? 0 : fCos;

      // Calculate the color value
      _ptData[un8Cnt].tOmni.un8R = uint8_t(C3DEnv::Active3DEnv()->_tDefOLight.un8R * fCos);
      _ptData[un8Cnt].tOmni.un8G = uint8_t(C3DEnv::Active3DEnv()->_tDefOLight.un8G * fCos);
      _ptData[un8Cnt].tOmni.un8B = uint8_t(C3DEnv::Active3DEnv()->_tDefOLight.un8B * fCos);
    }
  }
}

void CSprite2D::_CalcIntenOmni_(uint8_t un8Index, TOmniLight *ptVOmniLight)
{
  // Temporary variables
  float fXm, fYm, fZm, fMag, fCos;

  // Use the supplied omni light
  if(ptVOmniLight) {
    // Create the normalized light vector from the panel center to the omni light position
    fXm = ptVOmniLight->fX - _ptData[un8Index].csPos._tLocal.fX;
    fYm = ptVOmniLight->fY - _ptData[un8Index].csPos._tLocal.fY;
    fZm = ptVOmniLight->fZ - _ptData[un8Index].csPos._tLocal.fZ;
    fMag = sqrt(fXm * fXm + fYm * fYm + fZm * fZm);
    fYm /= fMag;
    fZm /= fMag;

    // Calculate the angle between sprite's normal (fixed) and the light vector
    fCos = _ptData[un8Index].fX * fXm + _ptData[un8Index].fY * fYm - _ptData[un8Index].fZ * fZm;
    fCos = (fCos < 0) ? 0 : fCos;

    // Calculate the color value
    _ptData[un8Index].tOmni.un8R = uint8_t(ptVOmniLight->un8R * fCos);
    _ptData[un8Index].tOmni.un8G = uint8_t(ptVOmniLight->un8G * fCos);
    _ptData[un8Index].tOmni.un8B = uint8_t(ptVOmniLight->un8B * fCos);
  }
  // Use the default omni light from the active 3D environment
  else {
    // Create the normalized light vector from the panel center to the omni light position
    fXm = C3DEnv::Active3DEnv()->_tDefOLight.fX - _ptData[un8Index].csPos._tLocal.fX;
    fYm = C3DEnv::Active3DEnv()->_tDefOLight.fY - _ptData[un8Index].csPos._tLocal.fY;
    fZm = C3DEnv::Active3DEnv()->_tDefOLight.fZ - _ptData[un8Index].csPos._tLocal.fZ;
    fMag = sqrt(fXm * fXm + fYm * fYm + fZm * fZm);
    fYm /= fMag;
    fZm /= fMag;

    // Calculate the angle between sprite's normal (fixed) and the light vector
    fCos = _ptData[un8Index].fX * fXm + _ptData[un8Index].fY * fYm - _ptData[un8Index].fZ * fZm;
    fCos = (fCos < 0) ? 0 : fCos;

    // Calculate the color value
    _ptData[un8Index].tOmni.un8R = uint8_t(C3DEnv::Active3DEnv()->_tDefOLight.un8R * fCos);
    _ptData[un8Index].tOmni.un8G = uint8_t(C3DEnv::Active3DEnv()->_tDefOLight.un8G * fCos);
    _ptData[un8Index].tOmni.un8B = uint8_t(C3DEnv::Active3DEnv()->_tDefOLight.un8B * fCos);
  }
}

void CSprite2D::_CalcIntenOmni_(uint8_t un8StartIndex, uint8_t un8EndIndex, TOmniLight *ptVOmniLight)
{
  // Temporary variables
  float fXm, fYm, fZm, fMag, fCos;

  // Calculate omni light intensity
  for(uint8_t un8Cnt = un8StartIndex; un8Cnt <= un8EndIndex; un8Cnt++) {
    // Use the supplied omni light
    if(ptVOmniLight) {
      // Create the normalized light vector from the panel center to the omni light position
      fXm = ptVOmniLight->fX - _ptData[un8Cnt].csPos._tLocal.fX;
      fYm = ptVOmniLight->fY - _ptData[un8Cnt].csPos._tLocal.fY;
      fZm = ptVOmniLight->fZ - _ptData[un8Cnt].csPos._tLocal.fZ;
      fMag = sqrt(fXm * fXm + fYm * fYm + fZm * fZm);
      fYm /= fMag;
      fZm /= fMag;

      // Calculate the angle between sprite's normal (fixed) and the light vector
      fCos = _ptData[un8Cnt].fX * fXm + _ptData[un8Cnt].fY * fYm - _ptData[un8Cnt].fZ * fZm;
      fCos = (fCos < 0) ? 0 : fCos;

      // Calculate the color value
      _ptData[un8Cnt].tOmni.un8R = uint8_t(ptVOmniLight->un8R * fCos);
      _ptData[un8Cnt].tOmni.un8G = uint8_t(ptVOmniLight->un8G * fCos);
      _ptData[un8Cnt].tOmni.un8B = uint8_t(ptVOmniLight->un8B * fCos);
    }
    // Use the default omni light from the active 3D environment
    else {
      // Create the normalized light vector from the panel center to the omni light position
      fXm = C3DEnv::Active3DEnv()->_tDefOLight.fX - _ptData[un8Cnt].csPos._tLocal.fX;
      fYm = C3DEnv::Active3DEnv()->_tDefOLight.fY - _ptData[un8Cnt].csPos._tLocal.fY;
      fZm = C3DEnv::Active3DEnv()->_tDefOLight.fZ - _ptData[un8Cnt].csPos._tLocal.fZ;
      fMag = sqrt(fXm * fXm + fYm * fYm + fZm * fZm);
      fYm /= fMag;
      fZm /= fMag;

      // Calculate the angle between sprite's normal (fixed) and the light vector
      fCos = _ptData[un8Cnt].fX * fXm + _ptData[un8Cnt].fY * fYm - _ptData[un8Cnt].fZ * fZm;
      fCos = (fCos < 0) ? 0 : fCos;

      // Calculate the color value
      _ptData[un8Cnt].tOmni.un8R = uint8_t(C3DEnv::Active3DEnv()->_tDefOLight.un8R * fCos);
      _ptData[un8Cnt].tOmni.un8G = uint8_t(C3DEnv::Active3DEnv()->_tDefOLight.un8G * fCos);
      _ptData[un8Cnt].tOmni.un8B = uint8_t(C3DEnv::Active3DEnv()->_tDefOLight.un8B * fCos);
    }
  }
}

void CSprite2D::_CalcIntenAmbient_(TAmbientLight *ptVAmbientLight)
{
  // Use the supplied ambient light
  if(ptVAmbientLight) {
    _tAmbient.un8R  = ptVAmbientLight->un8R;
    _tAmbient.un8G  = ptVAmbientLight->un8G;
    _tAmbient.un8B  = ptVAmbientLight->un8B;
  }
  else {
  // Use the default ambient light from the active 3D environment
    _tAmbient.un8R  = C3DEnv::Active3DEnv()->_tDefALight.un8R;
    _tAmbient.un8G  = C3DEnv::Active3DEnv()->_tDefALight.un8G;
    _tAmbient.un8B  = C3DEnv::Active3DEnv()->_tDefALight.un8B;
  }
}

void CSprite2D::_CalcInten_()
{
  for(uint8_t un8Cnt = 0; un8Cnt < _un8TotData; un8Cnt++) {
    _ptData[un8Cnt].tTotal.un8R = (_ptData[un8Cnt].tOmni.un8R + _tAmbient.un8R) >> 1;
    _ptData[un8Cnt].tTotal.un8G = (_ptData[un8Cnt].tOmni.un8G + _tAmbient.un8G) >> 1;
    _ptData[un8Cnt].tTotal.un8B = (_ptData[un8Cnt].tOmni.un8B + _tAmbient.un8B) >> 1;
  }
}

void CSprite2D::_CalcInten_(uint8_t un8Index)
{
  _ptData[un8Index].tTotal.un8R = (_ptData[un8Index].tOmni.un8R + _tAmbient.un8R) >> 1;
  _ptData[un8Index].tTotal.un8G = (_ptData[un8Index].tOmni.un8G + _tAmbient.un8G) >> 1;
  _ptData[un8Index].tTotal.un8B = (_ptData[un8Index].tOmni.un8B + _tAmbient.un8B) >> 1;
}

void CSprite2D::_CalcInten_(uint8_t un8StartIndex, uint8_t un8EndIndex)
{
  for(uint8_t un8Cnt = un8StartIndex; un8Cnt <= un8EndIndex; un8Cnt++) {
    _ptData[un8Cnt].tTotal.un8R = (_ptData[un8Cnt].tOmni.un8R + _tAmbient.un8R) >> 1;
    _ptData[un8Cnt].tTotal.un8G = (_ptData[un8Cnt].tOmni.un8G + _tAmbient.un8G) >> 1;
    _ptData[un8Cnt].tTotal.un8B = (_ptData[un8Cnt].tOmni.un8B + _tAmbient.un8B) >> 1;
  }
}

void CSprite2D::SetPosition(float fNewX, float fNewY, float fNewZ)
{
  for(uint8_t un8Cnt = 0; un8Cnt < _un8TotData; un8Cnt++) {
    _ptData[un8Cnt].csPos._tLocal.fX = fNewX + _fXo;
    _ptData[un8Cnt].csPos._tLocal.fY = fNewY + _fYo;
    _ptData[un8Cnt].csPos._tLocal.fZ = fNewZ;
  }
}

void CSprite2D::SetPosition(uint8_t un8StartIndex, uint8_t un8EndIndex, float fNewX, float fNewY, float fNewZ)
{
  for(uint8_t un8Cnt = un8StartIndex; un8Cnt <= un8EndIndex; un8Cnt++) {
    _ptData[un8Cnt].csPos._tLocal.fX = fNewX + _fXo;
    _ptData[un8Cnt].csPos._tLocal.fY = fNewY + _fYo;
    _ptData[un8Cnt].csPos._tLocal.fZ = fNewZ;
  }
}

void CSprite2D::Translate(float fAmountX, float fAmountY, float fAmountZ)
{
  for(uint8_t un8Cnt = 0; un8Cnt < _un8TotData; un8Cnt++) {
    _ptData[un8Cnt].csPos._tLocal.fX += fAmountX;
    _ptData[un8Cnt].csPos._tLocal.fY += fAmountY;
    _ptData[un8Cnt].csPos._tLocal.fZ += fAmountZ;
  }
}

void CSprite2D::Translate(uint8_t un8StartIndex, uint8_t un8EndIndex, float fAmountX, float fAmountY, float fAmountZ)
{
  for(uint8_t un8Cnt = un8StartIndex; un8Cnt <= un8EndIndex; un8Cnt++) {
    _ptData[un8Cnt].csPos._tLocal.fX += fAmountX;
    _ptData[un8Cnt].csPos._tLocal.fY += fAmountY;
    _ptData[un8Cnt].csPos._tLocal.fZ += fAmountZ;
  }
}

void CSprite2D::SetDirection(float fDirX, float fDirY, float fDirZ)
{
  float fMag = sqrt(fDirX * fDirX + fDirY * fDirY + fDirZ * fDirZ);
  fDirX /= fMag; fDirY /= fMag; fDirZ /= fMag;
  for(uint8_t un8Cnt = 0; un8Cnt < _un8TotData; un8Cnt++) {
    _ptData[un8Cnt].fX = fDirX;
    _ptData[un8Cnt].fY = fDirY;
    _ptData[un8Cnt].fZ = fDirZ;
  }
}

void CSprite2D::SetDirection(uint8_t un8StartIndex, uint8_t un8EndIndex, float fDirX, float fDirY, float fDirZ)
{
  float fMag = sqrt(fDirX * fDirX + fDirY * fDirY + fDirZ * fDirZ);
  fDirX /= fMag; fDirY /= fMag; fDirZ /= fMag;
  for(uint8_t un8Cnt = un8StartIndex; un8Cnt <= un8EndIndex; un8Cnt++) {
    _ptData[un8Cnt].fX = fDirX;
    _ptData[un8Cnt].fY = fDirY;
    _ptData[un8Cnt].fZ = fDirZ;
  }
}

void CSprite2D::Exclude(uint8_t un8StartIndex, uint8_t un8EndIndex, bool bExcluded)
{
  for(uint8_t un8Cnt = un8StartIndex; un8Cnt <= un8EndIndex; un8Cnt++)
    _ptData[un8Cnt].bExcluded = bExcluded;
}

void CSprite2D::InvertExclusions()
{
  for(uint8_t un8Cnt = 0; un8Cnt < _un8TotData; un8Cnt++)
    _ptData[un8Cnt].bExcluded = !_ptData[un8Cnt].bExcluded;
}

void CSprite2D::SNRender(const CMatrix3D &csTMatrix) // Solid, not shaded
{
  // Temporary variables
  uint8_t  un8Cnt;
  int32_t  n32Z, n32L, n32R, n32T, n32B, n32TexLeft, n32TexTop;
  int32_t  n32U, n32UStep, n32V, n32VStep, n32TY, n32TX, *pn32Z;
  int32_t  n32SLSize, n32Ur;
  uint32_t un32YIndex, *pun32Buff, *pun32Tex;
  float    fX, fY, fW, fH;

  // Render all sprites
  n32SLSize  = int32_t(_pcsImageSprite->_un32Width << 2);
  for(un8Cnt = 0; un8Cnt < _un8TotData; un8Cnt++) {
    // Preparation
    if(_ptData[un8Cnt].bExcluded) continue;
    csTMatrix.Transform(_ptData[un8Cnt].csPos);
    if(!_InsideZBoundaries_(_ptData[un8Cnt].csPos._tWorld.fZ)) continue;
    _ProjectSprite_(fX, fY, fW, fH, n32L, n32R, n32T, n32B, _ptData[un8Cnt].csPos._tWorld.fX, _ptData[un8Cnt].csPos._tWorld.fY, _ptData[un8Cnt].csPos._tWorld.fZ, _pcsImageSprite->_un32Width, _pcsImageSprite->_un32Height);
    if(!_Perform2DClip_(n32TexLeft, n32TexTop, n32L, n32R, n32T, n32B, _pcsImageSprite->_un32Width, _pcsImageSprite->_un32Height, fW, fH)) continue;
    n32Z = _CalcZValue_(_ptData[un8Cnt].csPos._tWorld.fZ);
    // Render
    n32V       = n32TexTop << _SSTEP_PREC_;
    n32UStep   = int32_t(float(_pcsImageSprite->_un32Width << _SSTEP_PREC_) / fW);
    n32VStep   = int32_t(float(_pcsImageSprite->_un32Height << _SSTEP_PREC_) / fH);
    un32YIndex = n32T * C3DEnv::Active3DEnv()->_tRB->un32Width;

    for(n32TY = n32T; n32TY < n32B; n32TY++) {
      n32U      = n32TexLeft << _SSTEP_PREC_;
      pn32Z     = &C3DEnv::Active3DEnv()->_tRB->pn32ZBuffer[un32YIndex + uint32_t(n32L)];
      pun32Buff = reinterpret_cast<uint32_t*>(&C3DEnv::Active3DEnv()->_tRB->pun8Buffer[(un32YIndex + uint32_t(n32L)) << 2]);
      pun32Tex  = reinterpret_cast<uint32_t*>(&_pcsImageSprite->_pun8Buffer[(n32V >> _SSTEP_PREC_) * n32SLSize]);

      for(n32TX = n32L; n32TX < n32R; n32TX++) {
        if(*pn32Z < n32Z) {
          n32Ur = n32U >> _SSTEP_PREC_;
          if(*(pun32Tex + n32Ur) != _un32ExclColor) {
            *pn32Z = n32Z;
            *pun32Buff = *(pun32Tex + n32Ur);
          }
        }
        n32U += n32UStep;
        pn32Z++;
        pun32Buff++;
      }
      n32V += n32VStep;
      un32YIndex += C3DEnv::Active3DEnv()->_tRB->un32Width;
    }
  }
}

void CSprite2D::SFRender(const CMatrix3D &csTMatrix) // Solid, flat shaded
{
  // Temporary variables
  uint8_t  un8Cnt, *pun8Tex;
  int32_t  n32Z, n32L, n32R, n32T, n32B, n32TexLeft, n32TexTop;
  int32_t  n32U, n32UStep, n32V, n32VStep, n32TY, n32TX, *pn32Z;
  int32_t  n32SLSize, n32Ur;
  uint32_t un32YIndex, *pun32Buff;
  float    fX, fY, fW, fH;

  // Render all sprites
  n32SLSize  = int32_t(_pcsImageSprite->_un32Width << 2);
  for(un8Cnt = 0; un8Cnt < _un8TotData; un8Cnt++) {
    // Preparation
    if(_ptData[un8Cnt].bExcluded) continue;
    csTMatrix.Transform(_ptData[un8Cnt].csPos);
    if(!_InsideZBoundaries_(_ptData[un8Cnt].csPos._tWorld.fZ)) continue;
    _ProjectSprite_(fX, fY, fW, fH, n32L, n32R, n32T, n32B, _ptData[un8Cnt].csPos._tWorld.fX, _ptData[un8Cnt].csPos._tWorld.fY, _ptData[un8Cnt].csPos._tWorld.fZ, _pcsImageSprite->_un32Width, _pcsImageSprite->_un32Height);
    if(!_Perform2DClip_(n32TexLeft, n32TexTop, n32L, n32R, n32T, n32B, _pcsImageSprite->_un32Width, _pcsImageSprite->_un32Height, fW, fH)) continue;
    n32Z = _CalcZValue_(_ptData[un8Cnt].csPos._tWorld.fZ);
    // Render
    n32V       = n32TexTop << _SSTEP_PREC_;
    n32UStep   = int32_t(float(_pcsImageSprite->_un32Width << _SSTEP_PREC_) / fW);
    n32VStep   = int32_t(float(_pcsImageSprite->_un32Height << _SSTEP_PREC_) / fH);
    un32YIndex = n32T * C3DEnv::Active3DEnv()->_tRB->un32Width;

    for(n32TY = n32T; n32TY < n32B; n32TY++) {
      n32U      = n32TexLeft << _SSTEP_PREC_;
      pn32Z     = &C3DEnv::Active3DEnv()->_tRB->pn32ZBuffer[un32YIndex + uint32_t(n32L)];
      pun32Buff = reinterpret_cast<uint32_t*>(&C3DEnv::Active3DEnv()->_tRB->pun8Buffer[(un32YIndex + uint32_t(n32L)) << 2]);
      pun8Tex   = &_pcsImageSprite->_pun8Buffer[(n32V >> _SSTEP_PREC_) * n32SLSize];

      for(n32TX = n32L; n32TX < n32R; n32TX++) {
        if(*pn32Z < n32Z) {
          n32Ur = n32U >> _SSTEP_PREC_ << 2;
          if(*(reinterpret_cast<uint32_t*>(pun8Tex + n32Ur)) != _un32ExclColor) {
            *pn32Z = n32Z;
            *pun32Buff = xprv::_aun32RLightTable_[(uint32_t(pun8Tex[n32Ur + 2]) << 2) | _ptData[un8Cnt].tTotal.un8R] |
              xprv::_aun32GLightTable_[(uint32_t(pun8Tex[n32Ur + 1]) << 2) | _ptData[un8Cnt].tTotal.un8G] |
              xprv::_aun32BLightTable_[(uint32_t(pun8Tex[n32Ur + 0]) << 2) | _ptData[un8Cnt].tTotal.un8B];
          }
        }
        n32U += n32UStep;
        pn32Z++;
        pun32Buff++;
      }
      n32V += n32VStep;
      un32YIndex += C3DEnv::Active3DEnv()->_tRB->un32Width;
    }
  }
}

void CSprite2D::SBRender(const CMatrix3D &csTMatrix) // Solid, color blended
{
  // Temporary variables
  uint8_t  un8Cnt, *pun8Tex;
  int32_t  n32Z, n32L, n32R, n32T, n32B, n32TexLeft, n32TexTop;
  int32_t  n32U, n32UStep, n32V, n32VStep, n32TY, n32TX, *pn32Z;
  int32_t  n32SLSize, n32Ur;
  uint32_t un32YIndex, *pun32Buff;
  float    fX, fY, fW, fH;

  // Render all sprites
  n32SLSize  = int32_t(_pcsImageSprite->_un32Width << 2);
  for(un8Cnt = 0; un8Cnt < _un8TotData; un8Cnt++) {
    // Preparation
    if(_ptData[un8Cnt].bExcluded) continue;
    csTMatrix.Transform(_ptData[un8Cnt].csPos);
    if(!_InsideZBoundaries_(_ptData[un8Cnt].csPos._tWorld.fZ)) continue;
    _ProjectSprite_(fX, fY, fW, fH, n32L, n32R, n32T, n32B, _ptData[un8Cnt].csPos._tWorld.fX, _ptData[un8Cnt].csPos._tWorld.fY, _ptData[un8Cnt].csPos._tWorld.fZ, _pcsImageSprite->_un32Width, _pcsImageSprite->_un32Height);
    if(!_Perform2DClip_(n32TexLeft, n32TexTop, n32L, n32R, n32T, n32B, _pcsImageSprite->_un32Width, _pcsImageSprite->_un32Height, fW, fH)) continue;
    n32Z = _CalcZValue_(_ptData[un8Cnt].csPos._tWorld.fZ);
    // Render
    n32V       = n32TexTop << _SSTEP_PREC_;
    n32UStep   = int32_t(float(_pcsImageSprite->_un32Width << _SSTEP_PREC_) / fW);
    n32VStep   = int32_t(float(_pcsImageSprite->_un32Height << _SSTEP_PREC_) / fH);
    un32YIndex = n32T * C3DEnv::Active3DEnv()->_tRB->un32Width;

    for(n32TY = n32T; n32TY < n32B; n32TY++) {
      n32U      = n32TexLeft << _SSTEP_PREC_;
      pn32Z     = &C3DEnv::Active3DEnv()->_tRB->pn32ZBuffer[un32YIndex + uint32_t(n32L)];
      pun32Buff = reinterpret_cast<uint32_t*>(&C3DEnv::Active3DEnv()->_tRB->pun8Buffer[(un32YIndex + uint32_t(n32L)) << 2]);
      pun8Tex   = &_pcsImageSprite->_pun8Buffer[(n32V >> _SSTEP_PREC_) * n32SLSize];

      for(n32TX = n32L; n32TX < n32R; n32TX++) {
        if(*pn32Z < n32Z) {
          n32Ur = n32U >> _SSTEP_PREC_ << 2;
          if(*(reinterpret_cast<uint32_t*>(pun8Tex + n32Ur)) != _un32ExclColor) {
            *pn32Z = n32Z;
            *pun32Buff = xprv::_aun32RLightTable_[(uint32_t(pun8Tex[n32Ur + 2]) << 2) | _un32PCR] |
              xprv::_aun32GLightTable_[(uint32_t(pun8Tex[n32Ur + 1]) << 2) | _un32PCG] |
              xprv::_aun32BLightTable_[(uint32_t(pun8Tex[n32Ur + 0]) << 2) | _un32PCB];
          }
        }
        n32U += n32UStep;
        pn32Z++;
        pun32Buff++;
      }
      n32V += n32VStep;
      un32YIndex += C3DEnv::Active3DEnv()->_tRB->un32Width;
    }
  }
}

void CSprite2D::SCRender(const CMatrix3D &csTMatrix) // Solid, one color
{
  // Temporary variables
  uint8_t      un8Cnt;
  int32_t      n32Z, n32L, n32R, n32T, n32B, n32TexLeft, n32TexTop;
  int32_t      n32U, n32UStep, n32V, n32VStep, n32TY, n32TX, *pn32Z;
  int32_t      n32SLSize, n32Ur;
  uint32_t     un32YIndex, *pun32Buff, *pun32Tex;
  float        fX, fY, fW, fH;
  XColor15 xCol;

  // Render all sprites
  xCol = V_RGB(_un32PCR, _un32PCG, _un32PCB);
  n32SLSize  = int32_t(_pcsImageSprite->_un32Width << 2);
  for(un8Cnt = 0; un8Cnt < _un8TotData; un8Cnt++) {
    // Preparation
    if(_ptData[un8Cnt].bExcluded) continue;
    csTMatrix.Transform(_ptData[un8Cnt].csPos);
    if(!_InsideZBoundaries_(_ptData[un8Cnt].csPos._tWorld.fZ)) continue;
    _ProjectSprite_(fX, fY, fW, fH, n32L, n32R, n32T, n32B, _ptData[un8Cnt].csPos._tWorld.fX, _ptData[un8Cnt].csPos._tWorld.fY, _ptData[un8Cnt].csPos._tWorld.fZ, _pcsImageSprite->_un32Width, _pcsImageSprite->_un32Height);
    if(!_Perform2DClip_(n32TexLeft, n32TexTop, n32L, n32R, n32T, n32B, _pcsImageSprite->_un32Width, _pcsImageSprite->_un32Height, fW, fH)) continue;
    n32Z = _CalcZValue_(_ptData[un8Cnt].csPos._tWorld.fZ);
    // Render
    n32V       = n32TexTop << _SSTEP_PREC_;
    n32UStep   = int32_t(float(_pcsImageSprite->_un32Width << _SSTEP_PREC_) / fW);
    n32VStep   = int32_t(float(_pcsImageSprite->_un32Height << _SSTEP_PREC_) / fH);
    un32YIndex = n32T * C3DEnv::Active3DEnv()->_tRB->un32Width;

    for(n32TY = n32T; n32TY < n32B; n32TY++) {
      n32U      = n32TexLeft << _SSTEP_PREC_;
      pn32Z     = &C3DEnv::Active3DEnv()->_tRB->pn32ZBuffer[un32YIndex + uint32_t(n32L)];
      pun32Buff = reinterpret_cast<uint32_t*>(&C3DEnv::Active3DEnv()->_tRB->pun8Buffer[(un32YIndex + uint32_t(n32L)) << 2]);
      pun32Tex  = reinterpret_cast<uint32_t*>(&_pcsImageSprite->_pun8Buffer[(n32V >> _SSTEP_PREC_) * n32SLSize]);

      for(n32TX = n32L; n32TX < n32R; n32TX++) {
        if(*pn32Z < n32Z) {
          n32Ur = n32U >> _SSTEP_PREC_;
          if(*(pun32Tex + n32Ur) != _un32ExclColor) {
            *pn32Z = n32Z;
            *pun32Buff = xCol;
          }
        }
        n32U += n32UStep;
        pn32Z++;
        pun32Buff++;
      }
      n32V += n32VStep;
      un32YIndex += C3DEnv::Active3DEnv()->_tRB->un32Width;
    }
  }
}

void CSprite2D::TNRender(const CMatrix3D &csTMatrix) // Transparent, not shaded
{
  // Temporary variables
  uint8_t  un8Cnt, *pun8Tex, *pun8Buff;
  int32_t  n32Z, n32L, n32R, n32T, n32B, n32TexLeft, n32TexTop;
  int32_t  n32U, n32UStep, n32V, n32VStep, n32TY, n32TX, *pn32Z;
  int32_t  n32SLSize, n32Ur;
  uint32_t un32YIndex;
  float    fX, fY, fW, fH;

  // Render all sprites
  n32SLSize  = int32_t(_pcsImageSprite->_un32Width << 2);
  for(un8Cnt = 0; un8Cnt < _un8TotData; un8Cnt++) {
    // Preparation
    if(_ptData[un8Cnt].bExcluded) continue;
    csTMatrix.Transform(_ptData[un8Cnt].csPos);
    if(!_InsideZBoundaries_(_ptData[un8Cnt].csPos._tWorld.fZ)) continue;
    _ProjectSprite_(fX, fY, fW, fH, n32L, n32R, n32T, n32B, _ptData[un8Cnt].csPos._tWorld.fX, _ptData[un8Cnt].csPos._tWorld.fY, _ptData[un8Cnt].csPos._tWorld.fZ, _pcsImageSprite->_un32Width, _pcsImageSprite->_un32Height);
    if(!_Perform2DClip_(n32TexLeft, n32TexTop, n32L, n32R, n32T, n32B, _pcsImageSprite->_un32Width, _pcsImageSprite->_un32Height, fW, fH)) continue;
    n32Z = _CalcZValue_(_ptData[un8Cnt].csPos._tWorld.fZ);
    // Render
    n32V       = n32TexTop << _SSTEP_PREC_;
    n32UStep   = int32_t(float(_pcsImageSprite->_un32Width << _SSTEP_PREC_) / fW);
    n32VStep   = int32_t(float(_pcsImageSprite->_un32Height << _SSTEP_PREC_) / fH);
    un32YIndex = n32T * C3DEnv::Active3DEnv()->_tRB->un32Width;
    for(n32TY = n32T; n32TY < n32B; n32TY++) {

    n32U     = n32TexLeft << _SSTEP_PREC_;
      pn32Z    = &C3DEnv::Active3DEnv()->_tRB->pn32ZBuffer[un32YIndex + uint32_t(n32L)];
      pun8Buff = &C3DEnv::Active3DEnv()->_tRB->pun8Buffer[(un32YIndex + uint32_t(n32L)) << 2];
      pun8Tex  = &_pcsImageSprite->_pun8Buffer[(n32V >> _SSTEP_PREC_) * n32SLSize];

      for(n32TX = n32L; n32TX < n32R; n32TX++) {
        if(*pn32Z < n32Z) {
          n32Ur = n32U >> _SSTEP_PREC_ << 2;
          if(*(reinterpret_cast<uint32_t*>(pun8Tex + n32Ur)) != _un32ExclColor) {
            *pn32Z = n32Z;
            *(pun8Buff + 2) = xprv::_aun8CMeanTable_[(uint32_t(pun8Tex[n32Ur + 2]) << 2) | (*(pun8Buff + 2) >> 3)];
            *(pun8Buff + 1) = xprv::_aun8CMeanTable_[(uint32_t(pun8Tex[n32Ur + 1]) << 2) | (*(pun8Buff + 1) >> 3)];
            *(pun8Buff + 0) = xprv::_aun8CMeanTable_[(uint32_t(pun8Tex[n32Ur + 0]) << 2) | (*(pun8Buff + 0) >> 3)];
          }
        }

        n32U += n32UStep;
        pn32Z++;
        pun8Buff += 4;
      }
      n32V += n32VStep;
      un32YIndex += C3DEnv::Active3DEnv()->_tRB->un32Width;
    }
  }
}

void CSprite2D::TFRender(const CMatrix3D &csTMatrix) // Transparent, flat shaded
{
  // Temporary variables
  uint8_t  un8Cnt, *pun8Tex, *pun8Buff;
  int32_t  n32Z, n32L, n32R, n32T, n32B, n32TexLeft, n32TexTop;
  int32_t  n32U, n32UStep, n32V, n32VStep, n32TY, n32TX, *pn32Z;
  int32_t  n32SLSize, n32Ur;
  uint32_t un32YIndex;
  float    fX, fY, fW, fH;

  // Render all sprites
  n32SLSize  = int32_t(_pcsImageSprite->_un32Width << 2);
  for(un8Cnt = 0; un8Cnt < _un8TotData; un8Cnt++) {
    // Preparation
    if(_ptData[un8Cnt].bExcluded) continue;
    csTMatrix.Transform(_ptData[un8Cnt].csPos);
    if(!_InsideZBoundaries_(_ptData[un8Cnt].csPos._tWorld.fZ)) continue;
    _ProjectSprite_(fX, fY, fW, fH, n32L, n32R, n32T, n32B, _ptData[un8Cnt].csPos._tWorld.fX, _ptData[un8Cnt].csPos._tWorld.fY, _ptData[un8Cnt].csPos._tWorld.fZ, _pcsImageSprite->_un32Width, _pcsImageSprite->_un32Height);
    if(!_Perform2DClip_(n32TexLeft, n32TexTop, n32L, n32R, n32T, n32B, _pcsImageSprite->_un32Width, _pcsImageSprite->_un32Height, fW, fH)) continue;
    n32Z = _CalcZValue_(_ptData[un8Cnt].csPos._tWorld.fZ);
    // Render
    n32V       = n32TexTop << _SSTEP_PREC_;
    n32UStep   = int32_t(float(_pcsImageSprite->_un32Width << _SSTEP_PREC_) / fW);
    n32VStep   = int32_t(float(_pcsImageSprite->_un32Height << _SSTEP_PREC_) / fH);
    un32YIndex = n32T * C3DEnv::Active3DEnv()->_tRB->un32Width;

    for(n32TY = n32T; n32TY < n32B; n32TY++) {
      n32U     = n32TexLeft << _SSTEP_PREC_;
      pn32Z    = &C3DEnv::Active3DEnv()->_tRB->pn32ZBuffer[un32YIndex + uint32_t(n32L)];
      pun8Buff = &C3DEnv::Active3DEnv()->_tRB->pun8Buffer[(un32YIndex + uint32_t(n32L)) << 2];
      pun8Tex  = &_pcsImageSprite->_pun8Buffer[(n32V >> _SSTEP_PREC_) * n32SLSize];

      for(n32TX = n32L; n32TX < n32R; n32TX++) {
        if(*pn32Z < n32Z) {
          n32Ur = n32U >> _SSTEP_PREC_ << 2;
          if(*(reinterpret_cast<uint32_t*>(pun8Tex + n32Ur)) != _un32ExclColor) {
            *pn32Z = n32Z;
            *(pun8Buff + 2) = xprv::_aun8TLightTable_[(*(pun8Buff + 2) << 7) | (uint32_t(pun8Tex[n32Ur + 2]) << 2) | _ptData[un8Cnt].tTotal.un8R];
            *(pun8Buff + 1) = xprv::_aun8TLightTable_[(*(pun8Buff + 1) << 7) | (uint32_t(pun8Tex[n32Ur + 1]) << 2) | _ptData[un8Cnt].tTotal.un8G];
            *(pun8Buff + 0) = xprv::_aun8TLightTable_[(*(pun8Buff + 0) << 7) | (uint32_t(pun8Tex[n32Ur + 0]) << 2) | _ptData[un8Cnt].tTotal.un8B];
          }
        }

        n32U += n32UStep;
        pn32Z++;
        pun8Buff += 4;
      }
      n32V += n32VStep;
      un32YIndex += C3DEnv::Active3DEnv()->_tRB->un32Width;
    }
  }
}

void CSprite2D::TBRender(const CMatrix3D &csTMatrix) // Transparent, color blended
{
  // Temporary variables
  uint8_t  un8Cnt, *pun8Tex, *pun8Buff;
  int32_t  n32Z, n32L, n32R, n32T, n32B, n32TexLeft, n32TexTop;
  int32_t  n32U, n32UStep, n32V, n32VStep, n32TY, n32TX, *pn32Z;
  int32_t  n32SLSize, n32Ur;
  uint32_t un32YIndex;
  float    fX, fY, fW, fH;

  // Render all sprites
  n32SLSize  = int32_t(_pcsImageSprite->_un32Width << 2);
  for(un8Cnt = 0; un8Cnt < _un8TotData; un8Cnt++) {
    // Preparation
    if(_ptData[un8Cnt].bExcluded) continue;
    csTMatrix.Transform(_ptData[un8Cnt].csPos);
    if(!_InsideZBoundaries_(_ptData[un8Cnt].csPos._tWorld.fZ)) continue;
    _ProjectSprite_(fX, fY, fW, fH, n32L, n32R, n32T, n32B, _ptData[un8Cnt].csPos._tWorld.fX, _ptData[un8Cnt].csPos._tWorld.fY, _ptData[un8Cnt].csPos._tWorld.fZ, _pcsImageSprite->_un32Width, _pcsImageSprite->_un32Height);
    if(!_Perform2DClip_(n32TexLeft, n32TexTop, n32L, n32R, n32T, n32B, _pcsImageSprite->_un32Width, _pcsImageSprite->_un32Height, fW, fH)) continue;
    n32Z = _CalcZValue_(_ptData[un8Cnt].csPos._tWorld.fZ);
    // Render
    n32V       = n32TexTop << _SSTEP_PREC_;
    n32UStep   = int32_t(float(_pcsImageSprite->_un32Width << _SSTEP_PREC_) / fW);
    n32VStep   = int32_t(float(_pcsImageSprite->_un32Height << _SSTEP_PREC_) / fH);
    un32YIndex = n32T * C3DEnv::Active3DEnv()->_tRB->un32Width;

    for(n32TY = n32T; n32TY < n32B; n32TY++) {
      n32U     = n32TexLeft << _SSTEP_PREC_;
      pn32Z    = &C3DEnv::Active3DEnv()->_tRB->pn32ZBuffer[un32YIndex + uint32_t(n32L)];
      pun8Buff = &C3DEnv::Active3DEnv()->_tRB->pun8Buffer[(un32YIndex + uint32_t(n32L)) << 2];
      pun8Tex  = &_pcsImageSprite->_pun8Buffer[(n32V >> _SSTEP_PREC_) * n32SLSize];

      for(n32TX = n32L; n32TX < n32R; n32TX++) {
        if(*pn32Z < n32Z) {
          n32Ur = n32U >> _SSTEP_PREC_ << 2;
          if(*(reinterpret_cast<uint32_t*>(pun8Tex + n32Ur)) != _un32ExclColor) {
            *pn32Z = n32Z;
            *(pun8Buff + 2) = xprv::_aun8TLightTable_[(*(pun8Buff + 2) << 7) | (uint32_t(pun8Tex[n32Ur + 2]) << 2) | _un32PCR];
            *(pun8Buff + 1) = xprv::_aun8TLightTable_[(*(pun8Buff + 1) << 7) | (uint32_t(pun8Tex[n32Ur + 1]) << 2) | _un32PCG];
            *(pun8Buff + 0) = xprv::_aun8TLightTable_[(*(pun8Buff + 0) << 7) | (uint32_t(pun8Tex[n32Ur + 0]) << 2) | _un32PCB];
          }
        }

        n32U += n32UStep;
        pn32Z++;
        pun8Buff += 4;
      }
      n32V += n32VStep;
      un32YIndex += C3DEnv::Active3DEnv()->_tRB->un32Width;
    }
  }
}

void CSprite2D::TCRender(const CMatrix3D &csTMatrix) // Transparent, one color
{
  // Temporary variables
  uint8_t  un8Cnt, *pun8Tex, *pun8Buff;
  int32_t  n32Z, n32L, n32R, n32T, n32B, n32TexLeft, n32TexTop;
  int32_t  n32U, n32UStep, n32V, n32VStep, n32TY, n32TX, *pn32Z;
  int32_t  n32SLSize, n32Ur;
  uint32_t un32YIndex;
  float    fX, fY, fW, fH;

  // Render all sprites
  n32SLSize  = int32_t(_pcsImageSprite->_un32Width << 2);
  for(un8Cnt = 0; un8Cnt < _un8TotData; un8Cnt++) {
    // Preparation
    if(_ptData[un8Cnt].bExcluded) continue;
    csTMatrix.Transform(_ptData[un8Cnt].csPos);
    if(!_InsideZBoundaries_(_ptData[un8Cnt].csPos._tWorld.fZ)) continue;
    _ProjectSprite_(fX, fY, fW, fH, n32L, n32R, n32T, n32B, _ptData[un8Cnt].csPos._tWorld.fX, _ptData[un8Cnt].csPos._tWorld.fY, _ptData[un8Cnt].csPos._tWorld.fZ, _pcsImageSprite->_un32Width, _pcsImageSprite->_un32Height);
    if(!_Perform2DClip_(n32TexLeft, n32TexTop, n32L, n32R, n32T, n32B, _pcsImageSprite->_un32Width, _pcsImageSprite->_un32Height, fW, fH)) continue;
    n32Z = _CalcZValue_(_ptData[un8Cnt].csPos._tWorld.fZ);
    // Render
    n32V       = n32TexTop << _SSTEP_PREC_;
    n32UStep   = int32_t(float(_pcsImageSprite->_un32Width << _SSTEP_PREC_) / fW);
    n32VStep   = int32_t(float(_pcsImageSprite->_un32Height << _SSTEP_PREC_) / fH);
    un32YIndex = n32T * C3DEnv::Active3DEnv()->_tRB->un32Width;

    for(n32TY = n32T; n32TY < n32B; n32TY++) {
      n32U     = n32TexLeft << _SSTEP_PREC_;
      pn32Z    = &C3DEnv::Active3DEnv()->_tRB->pn32ZBuffer[un32YIndex + uint32_t(n32L)];
      pun8Buff = &C3DEnv::Active3DEnv()->_tRB->pun8Buffer[(un32YIndex + uint32_t(n32L)) << 2];
      pun8Tex  = &_pcsImageSprite->_pun8Buffer[(n32V >> _SSTEP_PREC_) * n32SLSize];

      for(n32TX = n32L; n32TX < n32R; n32TX++) {
        if(*pn32Z < n32Z) {
          n32Ur = n32U >> _SSTEP_PREC_ << 2;
          if(*(reinterpret_cast<uint32_t*>(pun8Tex + n32Ur)) != _un32ExclColor) {
            *pn32Z = n32Z;
            *(pun8Buff + 2) = xprv::_aun8CMeanTable_[(_un32PCR << 2) | (*(pun8Buff + 2) >> 3)];
            *(pun8Buff + 1) = xprv::_aun8CMeanTable_[(_un32PCG << 2) | (*(pun8Buff + 1) >> 3)];
            *(pun8Buff + 0) = xprv::_aun8CMeanTable_[(_un32PCB << 2) | (*(pun8Buff + 0) >> 3)];
          }
        }

        n32U += n32UStep;
        pn32Z++;
        pun8Buff += 4;
      }
      n32V += n32VStep;
      un32YIndex += C3DEnv::Active3DEnv()->_tRB->un32Width;
    }
  }
}

void CSprite2D::FogSNRender(const CMatrix3D &csTMatrix) // Fogged, solid, not shaded
{
  // Temporary variables
  uint8_t  un8Cnt, *pun8Tex;
  int32_t  n32Z, n32L, n32R, n32T, n32B, n32TexLeft, n32TexTop;
  int32_t  n32U, n32UStep, n32V, n32VStep, n32TY, n32TX, *pn32Z;
  int32_t  n32SLSize, n32Ur, n32Fog;
  uint32_t un32YIndex, *pun32Buff;
  float  fX, fY, fW, fH;

  // Render all sprites
  n32SLSize  = int32_t(_pcsImageSprite->_un32Width << 2);
  for(un8Cnt = 0; un8Cnt < _un8TotData; un8Cnt++) {
    // Preparation
    if(_ptData[un8Cnt].bExcluded) continue;
    csTMatrix.Transform(_ptData[un8Cnt].csPos);
    if(!_InsideZBoundaries_(_ptData[un8Cnt].csPos._tWorld.fZ)) continue;
    _ProjectSprite_(fX, fY, fW, fH, n32L, n32R, n32T, n32B, _ptData[un8Cnt].csPos._tWorld.fX, _ptData[un8Cnt].csPos._tWorld.fY, _ptData[un8Cnt].csPos._tWorld.fZ, _pcsImageSprite->_un32Width, _pcsImageSprite->_un32Height);
    if(!_Perform2DClip_(n32TexLeft, n32TexTop, n32L, n32R, n32T, n32B, _pcsImageSprite->_un32Width, _pcsImageSprite->_un32Height, fW, fH)) continue;
    n32Z   = _CalcZValue_(_ptData[un8Cnt].csPos._tWorld.fZ);
    n32Fog = _CalcFValue_(_ptData[un8Cnt].csPos._tWorld.fZ);
    // Render
    n32V       = n32TexTop << _SSTEP_PREC_;
    n32UStep   = int32_t(float(_pcsImageSprite->_un32Width << _SSTEP_PREC_) / fW);
    n32VStep   = int32_t(float(_pcsImageSprite->_un32Height << _SSTEP_PREC_) / fH);
    un32YIndex = n32T * C3DEnv::Active3DEnv()->_tRB->un32Width;

    for(n32TY = n32T; n32TY < n32B; n32TY++) {
      n32U      = n32TexLeft << _SSTEP_PREC_;
      pn32Z     = &C3DEnv::Active3DEnv()->_tRB->pn32ZBuffer[un32YIndex + uint32_t(n32L)];
      pun32Buff = reinterpret_cast<uint32_t*>(&C3DEnv::Active3DEnv()->_tRB->pun8Buffer[(un32YIndex + uint32_t(n32L)) << 2]);
      pun8Tex   = &_pcsImageSprite->_pun8Buffer[(n32V >> _SSTEP_PREC_) * n32SLSize];

      for(n32TX = n32L; n32TX < n32R; n32TX++) {
        if(*pn32Z < n32Z) {
          n32Ur = n32U >> _SSTEP_PREC_ << 2;
          if(*(reinterpret_cast<uint32_t*>(pun8Tex + n32Ur)) != _un32ExclColor) {
            *pn32Z = n32Z;
            *pun32Buff = xprv::_pun32RFogColorTable_[n32Fog | (uint32_t(pun8Tex[n32Ur + 2]) >> 3)] |
              xprv::_pun32GFogColorTable_[n32Fog | (uint32_t(pun8Tex[n32Ur + 1]) >> 3)] |
              xprv::_pun32BFogColorTable_[n32Fog | (uint32_t(pun8Tex[n32Ur + 0]) >> 3)];
          }
        }
        n32U += n32UStep;
        pn32Z++;
        pun32Buff++;
      }
      n32V += n32VStep;
      un32YIndex += C3DEnv::Active3DEnv()->_tRB->un32Width;
    }
  }
}

void CSprite2D::FogSFRender(const CMatrix3D &csTMatrix) // Fogged, solid, flat shaded
{
  // Temporary variables
  uint8_t  un8Cnt, *pun8Tex;
  int32_t  n32Z, n32L, n32R, n32T, n32B, n32TexLeft, n32TexTop;
  int32_t  n32U, n32UStep, n32V, n32VStep, n32TY, n32TX, *pn32Z;
  int32_t  n32SLSize, n32Ur, n32Fog;
  uint32_t un32YIndex, *pun32Buff;
  float    fX, fY, fW, fH;

  // Render all sprites
  n32SLSize  = int32_t(_pcsImageSprite->_un32Width << 2);
  for(un8Cnt = 0; un8Cnt < _un8TotData; un8Cnt++) {
    // Preparation
    if(_ptData[un8Cnt].bExcluded) continue;
    csTMatrix.Transform(_ptData[un8Cnt].csPos);
    if(!_InsideZBoundaries_(_ptData[un8Cnt].csPos._tWorld.fZ)) continue;
    _ProjectSprite_(fX, fY, fW, fH, n32L, n32R, n32T, n32B, _ptData[un8Cnt].csPos._tWorld.fX, _ptData[un8Cnt].csPos._tWorld.fY, _ptData[un8Cnt].csPos._tWorld.fZ, _pcsImageSprite->_un32Width, _pcsImageSprite->_un32Height);
    if(!_Perform2DClip_(n32TexLeft, n32TexTop, n32L, n32R, n32T, n32B, _pcsImageSprite->_un32Width, _pcsImageSprite->_un32Height, fW, fH)) continue;
    n32Z   = _CalcZValue_(_ptData[un8Cnt].csPos._tWorld.fZ);
    n32Fog = _CalcFValue_(_ptData[un8Cnt].csPos._tWorld.fZ) << 5;
    // Render
    n32V       = n32TexTop << _SSTEP_PREC_;
    n32UStep   = int32_t(float(_pcsImageSprite->_un32Width << _SSTEP_PREC_) / fW);
    n32VStep   = int32_t(float(_pcsImageSprite->_un32Height << _SSTEP_PREC_) / fH);
    un32YIndex = n32T * C3DEnv::Active3DEnv()->_tRB->un32Width;

    for(n32TY = n32T; n32TY < n32B; n32TY++) {
      n32U      = n32TexLeft << _SSTEP_PREC_;
      pn32Z     = &C3DEnv::Active3DEnv()->_tRB->pn32ZBuffer[un32YIndex + uint32_t(n32L)];
      pun32Buff = reinterpret_cast<uint32_t*>(&C3DEnv::Active3DEnv()->_tRB->pun8Buffer[(un32YIndex + uint32_t(n32L)) << 2]);
      pun8Tex   = &_pcsImageSprite->_pun8Buffer[(n32V >> _SSTEP_PREC_) * n32SLSize];

      for(n32TX = n32L; n32TX < n32R; n32TX++) {
        if(*pn32Z < n32Z) {
          n32Ur = n32U >> _SSTEP_PREC_ << 2;
          if(*(reinterpret_cast<uint32_t*>(pun8Tex + n32Ur)) != _un32ExclColor) {
            *pn32Z = n32Z;
            *pun32Buff = xprv::_pun32RFogLightTable_[n32Fog | (uint32_t(pun8Tex[n32Ur + 2]) << 2) | _ptData[un8Cnt].tTotal.un8R] |
              xprv::_pun32GFogLightTable_[n32Fog | (uint32_t(pun8Tex[n32Ur + 1]) << 2) | _ptData[un8Cnt].tTotal.un8G] |
              xprv::_pun32BFogLightTable_[n32Fog | (uint32_t(pun8Tex[n32Ur + 0]) << 2) | _ptData[un8Cnt].tTotal.un8B];
          }
        }
        n32U += n32UStep;
        pn32Z++;
        pun32Buff++;
      }
      n32V += n32VStep;
      un32YIndex += C3DEnv::Active3DEnv()->_tRB->un32Width;
    }
  }
}

void CSprite2D::FogSBRender(const CMatrix3D &csTMatrix) // Fogged, solid, color blended
{
  // Temporary variables
  uint8_t  un8Cnt, *pun8Tex;
  int32_t  n32Z, n32L, n32R, n32T, n32B, n32TexLeft, n32TexTop;
  int32_t  n32U, n32UStep, n32V, n32VStep, n32TY, n32TX, *pn32Z;
  int32_t  n32SLSize, n32Ur, n32Fog;
  uint32_t un32YIndex, *pun32Buff;
  float    fX, fY, fW, fH;

  // Render all sprites
  n32SLSize  = int32_t(_pcsImageSprite->_un32Width << 2);
  for(un8Cnt = 0; un8Cnt < _un8TotData; un8Cnt++) {
    // Preparation
    if(_ptData[un8Cnt].bExcluded) continue;
    csTMatrix.Transform(_ptData[un8Cnt].csPos);
    if(!_InsideZBoundaries_(_ptData[un8Cnt].csPos._tWorld.fZ)) continue;
    _ProjectSprite_(fX, fY, fW, fH, n32L, n32R, n32T, n32B, _ptData[un8Cnt].csPos._tWorld.fX, _ptData[un8Cnt].csPos._tWorld.fY, _ptData[un8Cnt].csPos._tWorld.fZ, _pcsImageSprite->_un32Width, _pcsImageSprite->_un32Height);
    if(!_Perform2DClip_(n32TexLeft, n32TexTop, n32L, n32R, n32T, n32B, _pcsImageSprite->_un32Width, _pcsImageSprite->_un32Height, fW, fH)) continue;
    n32Z   = _CalcZValue_(_ptData[un8Cnt].csPos._tWorld.fZ);
    n32Fog = _CalcFValue_(_ptData[un8Cnt].csPos._tWorld.fZ) << 5;
    // Render
    n32V       = n32TexTop << _SSTEP_PREC_;
    n32UStep   = int32_t(float(_pcsImageSprite->_un32Width << _SSTEP_PREC_) / fW);
    n32VStep   = int32_t(float(_pcsImageSprite->_un32Height << _SSTEP_PREC_) / fH);
    un32YIndex = n32T * C3DEnv::Active3DEnv()->_tRB->un32Width;

    for(n32TY = n32T; n32TY < n32B; n32TY++) {
      n32U      = n32TexLeft << _SSTEP_PREC_;
      pn32Z     = &C3DEnv::Active3DEnv()->_tRB->pn32ZBuffer[un32YIndex + uint32_t(n32L)];
      pun32Buff = reinterpret_cast<uint32_t*>(&C3DEnv::Active3DEnv()->_tRB->pun8Buffer[(un32YIndex + uint32_t(n32L)) << 2]);
      pun8Tex   = &_pcsImageSprite->_pun8Buffer[(n32V >> _SSTEP_PREC_) * n32SLSize];

      for(n32TX = n32L; n32TX < n32R; n32TX++) {
        if(*pn32Z < n32Z) {
          n32Ur = n32U >> _SSTEP_PREC_ << 2;
          if(*(reinterpret_cast<uint32_t*>(pun8Tex + n32Ur)) != _un32ExclColor) {
            *pn32Z = n32Z;
            *pun32Buff = xprv::_pun32RFogLightTable_[n32Fog | (uint32_t(pun8Tex[n32Ur + 2]) << 2) | _un32PCR] |
              xprv::_pun32GFogLightTable_[n32Fog | (uint32_t(pun8Tex[n32Ur + 1]) << 2) | _un32PCG] |
              xprv::_pun32BFogLightTable_[n32Fog | (uint32_t(pun8Tex[n32Ur + 0]) << 2) | _un32PCB];
          }
        }
        n32U += n32UStep;
        pn32Z++;
        pun32Buff++;
      }
      n32V += n32VStep;
      un32YIndex += C3DEnv::Active3DEnv()->_tRB->un32Width;
    }
  }
}

void CSprite2D::FogSCRender(const CMatrix3D &csTMatrix) // Fogged, solid, one color
{
  // Temporary variables
  uint8_t  un8Cnt, *pun8Tex;
  int32_t  n32Z, n32L, n32R, n32T, n32B, n32TexLeft, n32TexTop;
  int32_t  n32U, n32UStep, n32V, n32VStep, n32TY, n32TX, *pn32Z;
  int32_t  n32SLSize, n32Ur, n32Fog;
  uint32_t un32YIndex, *pun32Buff;
  float    fX, fY, fW, fH;

  // Render all sprites
  n32SLSize  = int32_t(_pcsImageSprite->_un32Width << 2);
  for(un8Cnt = 0; un8Cnt < _un8TotData; un8Cnt++) {
    // Preparation
    if(_ptData[un8Cnt].bExcluded) continue;
    csTMatrix.Transform(_ptData[un8Cnt].csPos);
    if(!_InsideZBoundaries_(_ptData[un8Cnt].csPos._tWorld.fZ)) continue;
    _ProjectSprite_(fX, fY, fW, fH, n32L, n32R, n32T, n32B, _ptData[un8Cnt].csPos._tWorld.fX, _ptData[un8Cnt].csPos._tWorld.fY, _ptData[un8Cnt].csPos._tWorld.fZ, _pcsImageSprite->_un32Width, _pcsImageSprite->_un32Height);
    if(!_Perform2DClip_(n32TexLeft, n32TexTop, n32L, n32R, n32T, n32B, _pcsImageSprite->_un32Width, _pcsImageSprite->_un32Height, fW, fH)) continue;
    n32Z   = _CalcZValue_(_ptData[un8Cnt].csPos._tWorld.fZ);
    n32Fog = _CalcFValue_(_ptData[un8Cnt].csPos._tWorld.fZ);
    // Render
    n32V       = n32TexTop << _SSTEP_PREC_;
    n32UStep   = int32_t(float(_pcsImageSprite->_un32Width << _SSTEP_PREC_) / fW);
    n32VStep   = int32_t(float(_pcsImageSprite->_un32Height << _SSTEP_PREC_) / fH);
    un32YIndex = n32T * C3DEnv::Active3DEnv()->_tRB->un32Width;

    for(n32TY = n32T; n32TY < n32B; n32TY++) {
      n32U      = n32TexLeft << _SSTEP_PREC_;
      pn32Z     = &C3DEnv::Active3DEnv()->_tRB->pn32ZBuffer[un32YIndex + uint32_t(n32L)];
      pun32Buff = reinterpret_cast<uint32_t*>(&C3DEnv::Active3DEnv()->_tRB->pun8Buffer[(un32YIndex + uint32_t(n32L)) << 2]);
      pun8Tex   = &_pcsImageSprite->_pun8Buffer[(n32V >> _SSTEP_PREC_) * n32SLSize];

      for(n32TX = n32L; n32TX < n32R; n32TX++) {
        if(*pn32Z < n32Z) {
          n32Ur = n32U >> _SSTEP_PREC_ << 2;
          if(*(reinterpret_cast<uint32_t*>(pun8Tex + n32Ur)) != _un32ExclColor) {
            *pn32Z = n32Z;
            *pun32Buff = xprv::_pun32RFogColorTable_[n32Fog | _un32PCR] |
              xprv::_pun32GFogColorTable_[n32Fog | _un32PCG] |
              xprv::_pun32BFogColorTable_[n32Fog | _un32PCB];
          }
        }
        n32U += n32UStep;
        pn32Z++;
        pun32Buff++;
      }
      n32V += n32VStep;
      un32YIndex += C3DEnv::Active3DEnv()->_tRB->un32Width;
    }
  }
}

void CSprite2D::FogTNRender(const CMatrix3D &csTMatrix) // Fogged, transparent, not shaded
{
  // Temporary variables
  uint8_t  un8Cnt, *pun8Tex, *pun8Buff;
  int32_t  n32Z, n32L, n32R, n32T, n32B, n32TexLeft, n32TexTop;
  int32_t  n32U, n32UStep, n32V, n32VStep, n32TY, n32TX, *pn32Z;
  int32_t  n32SLSize, n32Ur, n32Fog;
  uint32_t un32YIndex;
  float    fX, fY, fW, fH;

  // Render all sprites
  n32SLSize  = int32_t(_pcsImageSprite->_un32Width << 2);
  for(un8Cnt = 0; un8Cnt < _un8TotData; un8Cnt++) {
    // Preparation
    if(_ptData[un8Cnt].bExcluded) continue;
    csTMatrix.Transform(_ptData[un8Cnt].csPos);
    if(!_InsideZBoundaries_(_ptData[un8Cnt].csPos._tWorld.fZ)) continue;
    _ProjectSprite_(fX, fY, fW, fH, n32L, n32R, n32T, n32B, _ptData[un8Cnt].csPos._tWorld.fX, _ptData[un8Cnt].csPos._tWorld.fY, _ptData[un8Cnt].csPos._tWorld.fZ, _pcsImageSprite->_un32Width, _pcsImageSprite->_un32Height);
    if(!_Perform2DClip_(n32TexLeft, n32TexTop, n32L, n32R, n32T, n32B, _pcsImageSprite->_un32Width, _pcsImageSprite->_un32Height, fW, fH)) continue;
    n32Z   = _CalcZValue_(_ptData[un8Cnt].csPos._tWorld.fZ);
    n32Fog = _CalcFValue_(_ptData[un8Cnt].csPos._tWorld.fZ);
    // Render
    n32V       = n32TexTop << _SSTEP_PREC_;
    n32UStep   = int32_t(float(_pcsImageSprite->_un32Width << _SSTEP_PREC_) / fW);
    n32VStep   = int32_t(float(_pcsImageSprite->_un32Height << _SSTEP_PREC_) / fH);
    un32YIndex = n32T * C3DEnv::Active3DEnv()->_tRB->un32Width;

    for(n32TY = n32T; n32TY < n32B; n32TY++) {
      n32U     = n32TexLeft << _SSTEP_PREC_;
      pn32Z    = &C3DEnv::Active3DEnv()->_tRB->pn32ZBuffer[un32YIndex + uint32_t(n32L)];
      pun8Buff = &C3DEnv::Active3DEnv()->_tRB->pun8Buffer[(un32YIndex + uint32_t(n32L)) << 2];
      pun8Tex  = &_pcsImageSprite->_pun8Buffer[(n32V >> _SSTEP_PREC_) * n32SLSize];

      for(n32TX = n32L; n32TX < n32R; n32TX++) {
        if(*pn32Z < n32Z) {
          n32Ur = n32U >> _SSTEP_PREC_ << 2;
          if(*(reinterpret_cast<uint32_t*>(pun8Tex + n32Ur)) != _un32ExclColor) {
            *pn32Z = n32Z;
            *(pun8Buff + 2) = xprv::_pun8TRFogColorTable_[(*(pun8Buff + 2) << 7) | n32Fog | (uint32_t(pun8Tex[n32Ur + 2]) >> 3)];
            *(pun8Buff + 1) = xprv::_pun8TGFogColorTable_[(*(pun8Buff + 1) << 7) | n32Fog | (uint32_t(pun8Tex[n32Ur + 1]) >> 3)];
            *(pun8Buff + 0) = xprv::_pun8TBFogColorTable_[(*(pun8Buff + 0) << 7) | n32Fog | (uint32_t(pun8Tex[n32Ur + 0]) >> 3)];
          }
        }

        n32U += n32UStep;
        pn32Z++;
        pun8Buff += 4;
      }
      n32V += n32VStep;
      un32YIndex += C3DEnv::Active3DEnv()->_tRB->un32Width;
    }
  }
}

void CSprite2D::FogTFRender(const CMatrix3D &csTMatrix) // Fogged, transparent, flat shaded
{
  // Temporary variables
  uint8_t  un8Cnt, *pun8Tex, *pun8Buff;
  int32_t  n32Z, n32L, n32R, n32T, n32B, n32TexLeft, n32TexTop;
  int32_t  n32U, n32UStep, n32V, n32VStep, n32TY, n32TX, *pn32Z;
  int32_t  n32SLSize, n32Ur, n32Fog;
  uint32_t un32YIndex;
  float    fX, fY, fW, fH;

  // Render all sprites
  n32SLSize  = int32_t(_pcsImageSprite->_un32Width << 2);
  for(un8Cnt = 0; un8Cnt < _un8TotData; un8Cnt++) {
    // Preparation
    if(_ptData[un8Cnt].bExcluded) continue;
    csTMatrix.Transform(_ptData[un8Cnt].csPos);
    if(!_InsideZBoundaries_(_ptData[un8Cnt].csPos._tWorld.fZ)) continue;
    _ProjectSprite_(fX, fY, fW, fH, n32L, n32R, n32T, n32B, _ptData[un8Cnt].csPos._tWorld.fX, _ptData[un8Cnt].csPos._tWorld.fY, _ptData[un8Cnt].csPos._tWorld.fZ, _pcsImageSprite->_un32Width, _pcsImageSprite->_un32Height);
    if(!_Perform2DClip_(n32TexLeft, n32TexTop, n32L, n32R, n32T, n32B, _pcsImageSprite->_un32Width, _pcsImageSprite->_un32Height, fW, fH)) continue;
    n32Z   = _CalcZValue_(_ptData[un8Cnt].csPos._tWorld.fZ);
    n32Fog = _CalcFValue_(_ptData[un8Cnt].csPos._tWorld.fZ) << 5;
    // Render
    n32V       = n32TexTop << _SSTEP_PREC_;
    n32UStep   = int32_t(float(_pcsImageSprite->_un32Width << _SSTEP_PREC_) / fW);
    n32VStep   = int32_t(float(_pcsImageSprite->_un32Height << _SSTEP_PREC_) / fH);
    un32YIndex = n32T * C3DEnv::Active3DEnv()->_tRB->un32Width;

    for(n32TY = n32T; n32TY < n32B; n32TY++) {
      n32U     = n32TexLeft << _SSTEP_PREC_;
      pn32Z    = &C3DEnv::Active3DEnv()->_tRB->pn32ZBuffer[un32YIndex + uint32_t(n32L)];
      pun8Buff = &C3DEnv::Active3DEnv()->_tRB->pun8Buffer[(un32YIndex + uint32_t(n32L)) << 2];
      pun8Tex  = &_pcsImageSprite->_pun8Buffer[(n32V >> _SSTEP_PREC_) * n32SLSize];

      for(n32TX = n32L; n32TX < n32R; n32TX++) {
        if(*pn32Z < n32Z) {
          n32Ur = n32U >> _SSTEP_PREC_ << 2;
          if(*(reinterpret_cast<uint32_t*>(pun8Tex + n32Ur)) != _un32ExclColor) {
            *pn32Z = n32Z;
            *(pun8Buff + 2) = xprv::_pun8TRFogLightTable_[(*(pun8Buff + 2) << 12) | n32Fog | (uint32_t(pun8Tex[n32Ur + 2]) << 2) | _ptData[un8Cnt].tTotal.un8R];
            *(pun8Buff + 1) = xprv::_pun8TGFogLightTable_[(*(pun8Buff + 1) << 12) | n32Fog | (uint32_t(pun8Tex[n32Ur + 1]) << 2) | _ptData[un8Cnt].tTotal.un8G];
            *(pun8Buff + 0) = xprv::_pun8TBFogLightTable_[(*(pun8Buff + 0) << 12) | n32Fog | (uint32_t(pun8Tex[n32Ur + 0]) << 2) | _ptData[un8Cnt].tTotal.un8B];
          }
        }

        n32U += n32UStep;
        pn32Z++;
        pun8Buff += 4;
      }
      n32V += n32VStep;
      un32YIndex += C3DEnv::Active3DEnv()->_tRB->un32Width;
    }
  }
}

void CSprite2D::FogTBRender(const CMatrix3D &csTMatrix) // Fogged, transparent, color blended
{
  // Temporary variables
  uint8_t  un8Cnt, *pun8Tex, *pun8Buff;
  int32_t  n32Z, n32L, n32R, n32T, n32B, n32TexLeft, n32TexTop;
  int32_t  n32U, n32UStep, n32V, n32VStep, n32TY, n32TX, *pn32Z;
  int32_t  n32SLSize, n32Ur, n32Fog;
  uint32_t un32YIndex;
  float    fX, fY, fW, fH;

  // Render all sprites
  n32SLSize  = int32_t(_pcsImageSprite->_un32Width << 2);
  for(un8Cnt = 0; un8Cnt < _un8TotData; un8Cnt++) {
    // Preparation
    if(_ptData[un8Cnt].bExcluded) continue;
    csTMatrix.Transform(_ptData[un8Cnt].csPos);
    if(!_InsideZBoundaries_(_ptData[un8Cnt].csPos._tWorld.fZ)) continue;
    _ProjectSprite_(fX, fY, fW, fH, n32L, n32R, n32T, n32B, _ptData[un8Cnt].csPos._tWorld.fX, _ptData[un8Cnt].csPos._tWorld.fY, _ptData[un8Cnt].csPos._tWorld.fZ, _pcsImageSprite->_un32Width, _pcsImageSprite->_un32Height);
    if(!_Perform2DClip_(n32TexLeft, n32TexTop, n32L, n32R, n32T, n32B, _pcsImageSprite->_un32Width, _pcsImageSprite->_un32Height, fW, fH)) continue;
    n32Z   = _CalcZValue_(_ptData[un8Cnt].csPos._tWorld.fZ);
    n32Fog = _CalcFValue_(_ptData[un8Cnt].csPos._tWorld.fZ) << 5;
    // Render
    n32V       = n32TexTop << _SSTEP_PREC_;
    n32UStep   = int32_t(float(_pcsImageSprite->_un32Width << _SSTEP_PREC_) / fW);
    n32VStep   = int32_t(float(_pcsImageSprite->_un32Height << _SSTEP_PREC_) / fH);
    un32YIndex = n32T * C3DEnv::Active3DEnv()->_tRB->un32Width;

    for(n32TY = n32T; n32TY < n32B; n32TY++) {
      n32U     = n32TexLeft << _SSTEP_PREC_;
      pn32Z    = &C3DEnv::Active3DEnv()->_tRB->pn32ZBuffer[un32YIndex + uint32_t(n32L)];
      pun8Buff = &C3DEnv::Active3DEnv()->_tRB->pun8Buffer[(un32YIndex + uint32_t(n32L)) << 2];
      pun8Tex  = &_pcsImageSprite->_pun8Buffer[(n32V >> _SSTEP_PREC_) * n32SLSize];

      for(n32TX = n32L; n32TX < n32R; n32TX++) {
        if(*pn32Z < n32Z) {
          n32Ur = n32U >> _SSTEP_PREC_ << 2;
          if(*(reinterpret_cast<uint32_t*>(pun8Tex + n32Ur)) != _un32ExclColor) {
            *pn32Z = n32Z;
            *(pun8Buff + 2) = xprv::_pun8TRFogLightTable_[(*(pun8Buff + 2) << 12) | n32Fog | (uint32_t(pun8Tex[n32Ur + 2]) << 2) | _un32PCR];
            *(pun8Buff + 1) = xprv::_pun8TGFogLightTable_[(*(pun8Buff + 1) << 12) | n32Fog | (uint32_t(pun8Tex[n32Ur + 1]) << 2) | _un32PCG];
            *(pun8Buff + 0) = xprv::_pun8TBFogLightTable_[(*(pun8Buff + 0) << 12) | n32Fog | (uint32_t(pun8Tex[n32Ur + 0]) << 2) | _un32PCB];
          }
        }

        n32U += n32UStep;
        pn32Z++;
        pun8Buff += 4;
      }
      n32V += n32VStep;
      un32YIndex += C3DEnv::Active3DEnv()->_tRB->un32Width;
    }
  }
}

void CSprite2D::FogTCRender(const CMatrix3D &csTMatrix) // Fogged, transparent, one color
{
  // Temporary variables
  uint8_t  un8Cnt, *pun8Tex, *pun8Buff;
  int32_t  n32Z, n32L, n32R, n32T, n32B, n32TexLeft, n32TexTop;
  int32_t  n32U, n32UStep, n32V, n32VStep, n32TY, n32TX, *pn32Z;
  int32_t  n32SLSize, n32Ur, n32Fog;
  uint32_t un32YIndex;
  float    fX, fY, fW, fH;

  // Render all sprites
  n32SLSize  = int32_t(_pcsImageSprite->_un32Width << 2);
  for(un8Cnt = 0; un8Cnt < _un8TotData; un8Cnt++) {
    // Preparation
    if(_ptData[un8Cnt].bExcluded) continue;
    csTMatrix.Transform(_ptData[un8Cnt].csPos);
    if(!_InsideZBoundaries_(_ptData[un8Cnt].csPos._tWorld.fZ)) continue;
    _ProjectSprite_(fX, fY, fW, fH, n32L, n32R, n32T, n32B, _ptData[un8Cnt].csPos._tWorld.fX, _ptData[un8Cnt].csPos._tWorld.fY, _ptData[un8Cnt].csPos._tWorld.fZ, _pcsImageSprite->_un32Width, _pcsImageSprite->_un32Height);
    if(!_Perform2DClip_(n32TexLeft, n32TexTop, n32L, n32R, n32T, n32B, _pcsImageSprite->_un32Width, _pcsImageSprite->_un32Height, fW, fH)) continue;
    n32Z   = _CalcZValue_(_ptData[un8Cnt].csPos._tWorld.fZ);
    n32Fog = _CalcFValue_(_ptData[un8Cnt].csPos._tWorld.fZ);
    // Render
    n32V       = n32TexTop << _SSTEP_PREC_;
    n32UStep   = int32_t(float(_pcsImageSprite->_un32Width << _SSTEP_PREC_) / fW);
    n32VStep   = int32_t(float(_pcsImageSprite->_un32Height << _SSTEP_PREC_) / fH);
    un32YIndex = n32T * C3DEnv::Active3DEnv()->_tRB->un32Width;

    for(n32TY = n32T; n32TY < n32B; n32TY++) {
      n32U     = n32TexLeft << _SSTEP_PREC_;
      pn32Z    = &C3DEnv::Active3DEnv()->_tRB->pn32ZBuffer[un32YIndex + uint32_t(n32L)];
      pun8Buff = &C3DEnv::Active3DEnv()->_tRB->pun8Buffer[(un32YIndex + uint32_t(n32L)) << 2];
      pun8Tex  = &_pcsImageSprite->_pun8Buffer[(n32V >> _SSTEP_PREC_) * n32SLSize];

      for(n32TX = n32L; n32TX < n32R; n32TX++) {
        if(*pn32Z < n32Z) {
          n32Ur = n32U >> _SSTEP_PREC_ << 2;
          if(*(reinterpret_cast<uint32_t*>(pun8Tex + n32Ur)) != _un32ExclColor) {
            *pn32Z = n32Z;
            *(pun8Buff + 2) = xprv::_pun8TRFogColorTable_[(*(pun8Buff + 2) << 7) | n32Fog | _un32PCR];
            *(pun8Buff + 1) = xprv::_pun8TGFogColorTable_[(*(pun8Buff + 1) << 7) | n32Fog | _un32PCG];
            *(pun8Buff + 0) = xprv::_pun8TBFogColorTable_[(*(pun8Buff + 0) << 7) | n32Fog | _un32PCB];
          }
        }

        n32U += n32UStep;
        pn32Z++;
        pun8Buff += 4;
      }
      n32V += n32VStep;
      un32YIndex += C3DEnv::Active3DEnv()->_tRB->un32Width;
    }
  }
}

void CSprite2D::ARender(const CMatrix3D &csTMatrix) // Additive, transparent
{
  // Temporary variables
  uint8_t  un8Cnt, *pun8Tex, *pun8Buff;
  int32_t  n32Z, n32L, n32R, n32T, n32B, n32TexLeft = 0, n32TexTop = 0;
  int32_t  n32U, n32UStep, n32V, n32VStep, n32TY, n32TX, *pn32Z;
  int32_t  n32SLSize, n32Ur;
  uint32_t un32YIndex;
  float    fX, fY, fW, fH;

  // Render all sprites
  n32SLSize  = int32_t(_pcsImageSprite->_un32Width << 2);
  for(un8Cnt = 0; un8Cnt < _un8TotData; un8Cnt++) {
    // Preparation
    if(_ptData[un8Cnt].bExcluded) continue;
    csTMatrix.Transform(_ptData[un8Cnt].csPos);
    if(!_InsideZBoundaries_(_ptData[un8Cnt].csPos._tWorld.fZ)) continue;
    _ProjectSprite_(fX, fY, fW, fH, n32L, n32R, n32T, n32B, _ptData[un8Cnt].csPos._tWorld.fX, _ptData[un8Cnt].csPos._tWorld.fY, _ptData[un8Cnt].csPos._tWorld.fZ, _pcsImageSprite->_un32Width, _pcsImageSprite->_un32Height);
    if(!_Perform2DClip_(n32TexLeft, n32TexTop, n32L, n32R, n32T, n32B, _pcsImageSprite->_un32Width, _pcsImageSprite->_un32Height, fW, fH)) continue;
    n32Z = _CalcZValue_(_ptData[un8Cnt].csPos._tWorld.fZ);
    // Render
    n32V       = n32TexTop << _SSTEP_PREC_;
    n32UStep   = int32_t(float(_pcsImageSprite->_un32Width << _SSTEP_PREC_) / fW);
    n32VStep   = int32_t(float(_pcsImageSprite->_un32Height << _SSTEP_PREC_) / fH);
    un32YIndex = n32T * C3DEnv::Active3DEnv()->_tRB->un32Width;

    for(n32TY = n32T; n32TY < n32B; n32TY++) {
      n32U     = n32TexLeft << _SSTEP_PREC_;
      pn32Z    = &C3DEnv::Active3DEnv()->_tRB->pn32ZBuffer[un32YIndex + uint32_t(n32L)];
      pun8Buff = &C3DEnv::Active3DEnv()->_tRB->pun8Buffer[(un32YIndex + uint32_t(n32L)) << 2];
      pun8Tex  = &_pcsImageSprite->_pun8Buffer[(n32V >> _SSTEP_PREC_) * n32SLSize];

      for(n32TX = n32L; n32TX < n32R; n32TX++) {
        if(*pn32Z < n32Z) {
          n32Ur = n32U >> _SSTEP_PREC_ << 2;
          if(*(reinterpret_cast<uint32_t*>(pun8Tex + n32Ur)) != _un32ExclColor) {
            *pn32Z = n32Z;
            *(pun8Buff + 2) = xprv::_aun8CAddTable_[(uint32_t(pun8Tex[n32Ur + 2]) << 2) | (*(pun8Buff + 2) >> 3)];
            *(pun8Buff + 1) = xprv::_aun8CAddTable_[(uint32_t(pun8Tex[n32Ur + 1]) << 2) | (*(pun8Buff + 1) >> 3)];
            *(pun8Buff + 0) = xprv::_aun8CAddTable_[(uint32_t(pun8Tex[n32Ur + 0]) << 2) | (*(pun8Buff + 0) >> 3)];
          }
        }

        n32U += n32UStep;
        pn32Z++;
        pun8Buff += 4;
      }
      n32V += n32VStep;
      un32YIndex += C3DEnv::Active3DEnv()->_tRB->un32Width;
    }
  }
}

bool CSprite2D::Collide(const float afCurPosXYZ[3], float afTransFactorXYZ[3]) const
{
  // Temporary variables
  uint8_t un8Cnt;
  float   fX, fY, fZ, fSX, fSY, fSZ, fDX, fDY, fDZ, fC;

  // Check with all sprites
  for(un8Cnt = 0; un8Cnt < _un8TotData; un8Cnt++) {
    // Check if excluded
    if(_ptData[un8Cnt].bExcluded) continue;
    // Calculate the new position
    fX = afCurPosXYZ[0] + afTransFactorXYZ[0];
    fY = afCurPosXYZ[1] + afTransFactorXYZ[1];
    fZ = afCurPosXYZ[2] + afTransFactorXYZ[2];
    // Calculate sprite position
    fSX = _ptData[un8Cnt].csPos._tLocal.fX; // - _fXo;
    fSY = _ptData[un8Cnt].csPos._tLocal.fY; // - _fYo;
    fSZ = _ptData[un8Cnt].csPos._tLocal.fZ;
    // Calculate distance
    fDX = fabs(fSX - fX);
    fDY = fabs(fSY - fY);
    fDZ = fabs(fSZ - fZ);
    // Check if close not enough
    if((fDX > _fRadius) && (fDY > _fRadius) && (fDZ > _fRadius)) continue;
    // Check collision
    fC  = sqrt(fDX * fDX + fDY * fDY + fDZ * fDZ);
    if(fC < _fRadius) return(true);
  }
  return(false);
}

void CSprite2D::GetEquivalentGeometry(TFloatXY &min, TFloatXY &max) const
{
  const float sw = _pcsImageSprite->_un32Width  / 8.0f;
  const float sh = _pcsImageSprite->_un32Height / 8.0f;

  const float x1 = _fXo - sw;
  const float x2 = _fXo + sw;
  const float y1 = _fYo - sh;
  const float y2 = _fYo + sh;

  min.x = std::min(x1, x2); min.y = std::min(y1, y2);
  max.x = std::max(x1, x2); max.y = std::max(y1, y2);
}

} // namespace vge
