/** \file Sprite2D.h */
// ############################################################################################
//  Sprite2D.h
//  Definition & implementation of _CSprite2D class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  Some of the source codes in this file were written based on the book :
//    Cutting-Edge 3D Game Programming with C++
//    John De Goes
//    Coriolis Group, Inc.
//    USA, 1996
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _3D_ENVIRONMENT___SPRITE_2D_H_
#define _3D_ENVIRONMENT___SPRITE_2D_H_

// VortexGE headers
#include "../2DManipulation/ImageClasses.h"
#include "3DEnvironment.h"
#include "Matrix3D.h"
#include "Point3D.h"


namespace vge {

  //!
  //! \brief Pivot position for CSprite2D
  //!
  //! \api pretty stable
  enum EPivot {
    EP_TopLeft,    EP_TopCenter,    EP_TopRight,
    EP_CenterLeft, EP_CenterCenter, EP_CenterRight,
    EP_BottomLeft, EP_BottomCenter, EP_BottomRight
  };


  //! \internal
  //! \brief Static sprite data
  //!
  //! \api pretty stable
  struct _TSSpriteData {
    CPoint3D csPos;      // The sprite's position in 3D world
    float    fX, fY, fZ; // The sprite's normal
    TUN8RGB  tOmni;      // The sprite's color (from omni light)
    TUN8RGB  tTotal;     // The sprite's color (total)
    bool     bExcluded;  // Flag to indicate whether this sprite is excluded

    // Support for SSE
    DECLARE_SSE_ENABLED_NEW_AND_DELETE_OPERATORS()
  };


  //!
  //! \brief 2D sprite class
  //!
  //! \api pretty stable
  class VGE_EXPORT CSprite2D : VCLASS_UNCOPYABLE {
    public:
      CSprite2D();
      ~CSprite2D() throw();

      // Initialization functions
      //!
      //! Initialize the sprite 2D with the given parameters.
      //! \param[in] pfPosXYZs            Pointer to buffer containing the XYZ
      //!                                 position of the sprites in 3D world.
      //! \param[in] un8PosCount          Number of XYZ pairs in the above buffer.
      //! \param[in] pcsImageSprite       Image to be used as the sprite's image.
      //! \param[in] ePivot               Origin of the sprite in the image.
      //! \param[in] n32ExcludedColorVRGB Which color to be assumed as transparent?
      //! \param[in] ptVAmbientLight      User ambient light to be used. If NULL,
      //!                                 the default ambient light of the active
      //!                                 3D environment (rendering context) will
      //!                                 be used.
      //! \param[in] ptVOmniLight         User omni light to be used. If NULL,
      //!                                 the default omni light of the active
      //!                                 3D environment (rendering context) will
      //!                                 be used.
      //! \param[in] fCollisionRadius     Collision radius of the sprite. If set
      //!                                 to -1, will be calculated automatically.
      void Initialize(const float   *pfPosXYZs,
                      uint8_t        un8PosCount,
                      const CImage  *pcsImageSprite,
                      EPivot         ePivot                = EP_BottomCenter,
                      uint32_t       un32ExcludedColorVRGB = EXCLUDED_COLOR_FIRST_PIXEL,
                      TAmbientLight *ptVAmbientLight       = NULL,
                      TOmniLight    *ptVOmniLight          = NULL,
                      float          fCollisionRadius      = -1.0f);
      //!
      //! Initialize the sprite 2D with the given parameters.
      //! \param[in] afPosXYZs            Array containing the XYZ position of the
      //!                                 sprites in 3D world.
      //! \param[in] pcsImageSprite       Image to be used as the sprite's image.
      //! \param[in] ePivot               Origin of the sprite in the image.
      //! \param[in] n32ExcludedColorVRGB Which color to be assumed as transparent?
      //! \param[in] ptVAmbientLight      User ambient light to be used. If NULL,
      //!                                 the default ambient light of the active
      //!                                 3D environment (rendering context) will
      //!                                 be used.
      //! \param[in] ptVOmniLight         User omni light to be used. If NULL,
      //!                                 the default omni light of the active
      //!                                 3D environment (rendering context) will
      //!                                 be used.
      //! \param[in] fCollisionRadius     Collision radius of the sprite. If set
      //!                                 to -1, will be calculated automatically.
      inline void Initialize(const vector<float> &afPosXYZs,
                             const CImage        *pcsImageSprite,
                             EPivot               ePivot                = EP_BottomCenter,
                             uint32_t             un32ExcludedColorVRGB = EXCLUDED_COLOR_FIRST_PIXEL,
                             TAmbientLight       *ptVAmbientLight       = NULL,
                             TOmniLight          *ptVOmniLight          = NULL,
                             float                fCollisionRadius      = -1.0f)
      { Initialize(&afPosXYZs[0], afPosXYZs.size()/3, pcsImageSprite, ePivot, un32ExcludedColorVRGB, ptVAmbientLight, ptVOmniLight, fCollisionRadius); }
      //!
      //! Clear all internal data of this sprite.
      void Clear();

      // Lighting functions
      //!
      //! (Re-)set the custom ambient light to be used with this sprite.
      inline void ResetAmbientLight(TAmbientLight *ptVAmbientLight = NULL);
      //!
      //! (Re-)set the custom omni light to be used with this sprite.
      inline void ResetOmniLight(TOmniLight *ptVOmniLight = NULL);
      //!
      //! (Re-)set the custom ambient and omni lights to be used with this sprite.
      inline void ResetLights(TAmbientLight *ptVAmbientLight = NULL, TOmniLight *ptVOmniLight = NULL);

      // Dynamic control functions
      //!
      //! Set the color of all the sprite.
      inline void SetColor(uint8_t un8R, uint8_t un8G, uint8_t un8B);
      //!
      //! Set the position of all the sprites in 3D world.
      void SetPosition(float fNewX, float fNewY, float fNewZ);
      //!
      //! Set the position of the sprite of the given index in 3D world.
      inline void SetPosition(uint8_t un8Index, float fNewX, float fNewY, float fNewZ);
      //!
      //! Set the position of the sprites of the given start and end
      //! indexes in 3D world.
      void SetPosition(uint8_t un8StartIndex, uint8_t un8EndIndex, float fNewX, float fNewY, float fNewZ);
      //!
      //! Translate the position of all the sprites in 3D world.
      void Translate(float fAmountX, float fAmountY, float fAmountZ);
      //!
      //! Translate the position of the sprite of the given index in 3D world.
      inline void Translate(uint8_t un8Index, float fAmountX, float fAmountY, float fAmountZ);
      //!
      //! Translate the position of the sprites of the given start and end
      //! indexes in 3D world.
      void Translate(uint8_t un8StartIndex, uint8_t un8EndIndex, float fAmountX, float fAmountY, float fAmountZ);
      //!
      //! Set the direction of all the sprites in 3D world.
      void SetDirection(float fDirX, float fDirY, float fDirZ);
      //!
      //! Set the direction of the sprite of the given index in 3D world.
      inline void SetDirection(uint8_t un8Index, float fDirX, float fDirY, float fDirZ);
      //!
      //! Set the direction of the sprites of the given start and end
      //! indexes in 3D world.
      void SetDirection(uint8_t un8StartIndex, uint8_t un8EndIndex, float fDirX, float fDirY, float fDirZ);
      //!
      //! Recalculate the lighting of all the sprites.
      inline void RecalculateLighting();
      //!
      //! Recalculate the lighting of the sprite of the given index.
      inline void RecalculateLighting(uint8_t un8Index);
      //!
      //! Recalculate the lighting of the sprites of the given start and
      //! end indexes.
      inline void RecalculateLighting(uint8_t un8StartIndex, uint8_t un8EndIndex);
      //!
      //! Exclude (or include) the sprite of the given index in rendering.
      inline void Exclude(uint8_t un8Index, bool bExcluded);
      //!
      //! Exclude (or include) the sprite of the given start and end indexes
      //! in rendering.
      void Exclude(uint8_t un8StartIndex, uint8_t un8EndIndex, bool bExcluded);
      //!
      //! Invert the excluded sprite indexes with the non-excluded sprite indexes.
      void InvertExclusions();

      // Normal rendering functions
      //!
      //! Render all the sprites: solid, normal sprite, not shaded.
      void SNRender(const CMatrix3D &csTMatrix);
      //!
      //! Render all the sprites: solid, normal sprite, flat shaded.
      void SFRender(const CMatrix3D &csTMatrix);
      //!
      //! Render all the sprites: solid, blended color, not shaded.
      void SBRender(const CMatrix3D &csTMatrix);
      //!
      //! Render all the sprites: solid, single color, not shaded.
      void SCRender(const CMatrix3D &csTMatrix);
      //!
      //! Render all the sprites: transparent, normal sprite, not shaded.
      void TNRender(const CMatrix3D &csTMatrix);
      //!
      //! Render all the sprites: transparent, blended color, not shaded.
      void TBRender(const CMatrix3D &csTMatrix);
      //!
      //! Render all the sprites: transparent, normal sprite, flat shaded.
      void TFRender(const CMatrix3D &csTMatrix);
      //!
      //! Render all the sprites: transparent, single color, not shaded.
      void TCRender(const CMatrix3D &csTMatrix);

      // Fogged rendering functions
      //!
      // Render all the sprites: fogged, solid, normal sprite, not shaded.
      void FogSNRender(const CMatrix3D &csTMatrix);
      //!
      // Render all the sprites: fogged, solid, normal sprite, flat shaded.
      void FogSFRender(const CMatrix3D &csTMatrix);
      //!
      // Render all the sprites: fogged, solid, blended color, flat shaded.
      void FogSBRender(const CMatrix3D &csTMatrix);
      //!
      // Render all the sprites: fogged, solid, single color, not shaded.
      void FogSCRender(const CMatrix3D &csTMatrix);
      //!
      // Render all the sprites: fogged, transparent, normal sprite, not shaded.
      void FogTNRender(const CMatrix3D &csTMatrix);
      //!
      // Render all the sprites: fogged, transparent, blended color, flat shaded.
      void FogTBRender(const CMatrix3D &csTMatrix);
      //!
      // Render all the sprites: fogged, transparent, normal sprite, flat shaded.
      void FogTFRender(const CMatrix3D &csTMatrix);
      //!
      // Render all the sprites: fogged, transparent, single color, not shaded.
      void FogTCRender(const CMatrix3D &csTMatrix);

      // Special rendering functions
      //!
      // Render all the sprites: transparent, color addition, not shaded.
      void ARender(const CMatrix3D &csTMatrix);

      // Collision functions
      //!
      //! Check collision between all non-excluded sprites with a point with
      //! the given position and movement vector.
      bool Collide(const float afCurPosXYZ[3], float afTransFactorXYZ[3]) const;
      //!
      //! Calculate the equivalent geometry of the sprite in standard
      //! 3D world coordinates.
      void GetEquivalentGeometry(TFloatXY &min, TFloatXY &max) const;

    protected:
      // Initialization functions
      void _CalcIntenOmni_(TOmniLight *ptVOmniLight);
      void _CalcIntenOmni_(uint8_t un8Index, TOmniLight *ptVOmniLight);
      void _CalcIntenOmni_(uint8_t un8StartIndex, uint8_t un8EndIndex, TOmniLight *ptVOmniLight);
      void _CalcIntenAmbient_(TAmbientLight *ptVAmbientLight);
      void _CalcInten_();
      void _CalcInten_(uint8_t un8Index);
      void _CalcInten_(uint8_t un8StartIndex, uint8_t un8EndIndex);

      // Data
      _TSSpriteData *_ptData;         // The indivudial sprite datas
      uint8_t        _un8TotData;     // Total sprite datas
      TUN8RGB        _tAmbient;       // The sprites' color (from ambient light)
      const CImage  *_pcsImageSprite; // The sprites' image
      uint32_t       _un32ExclColor;  // The color that will not be drawn
      float          _fXo;            // Position offset for pivoting
      float          _fYo;            // Position offset for pivoting
      float          _fRadius;        // For collision detection
      TOmniLight    *_ptOmni;         // Used omni light
      uint32_t       _un32PCR;        // The R component of the sprite's color
      uint32_t       _un32PCG;        // The G component of the sprite's color
      uint32_t       _un32PCB;        // The B component of the sprite's color
  };

  void CSprite2D::ResetAmbientLight(TAmbientLight *ptVAmbientLight)
  {
    _CalcIntenAmbient_(ptVAmbientLight);
    _CalcInten_();
  }

  void CSprite2D::ResetOmniLight(TOmniLight *ptVOmniLight)
  {
    _CalcIntenOmni_(ptVOmniLight);
    _CalcInten_();
  }

  void CSprite2D::ResetLights(TAmbientLight *ptVAmbientLight, TOmniLight *ptVOmniLight)
  {
    _CalcIntenAmbient_(ptVAmbientLight);
    _CalcIntenOmni_(ptVOmniLight);
    _CalcInten_();
  }

  void CSprite2D::SetColor(uint8_t un8R, uint8_t un8G, uint8_t un8B)
  {
    _un32PCR = uint32_t(un8R);
    _un32PCG = uint32_t(un8G);
    _un32PCB = uint32_t(un8B);
  }

  void CSprite2D::SetPosition(uint8_t un8Index, float fNewX, float fNewY, float fNewZ)
  {
    _ptData[un8Index].csPos._tLocal.fX = fNewX + _fXo;
    _ptData[un8Index].csPos._tLocal.fY = fNewY + _fYo;
    _ptData[un8Index].csPos._tLocal.fZ = fNewZ;
  }

  void CSprite2D::Translate(uint8_t un8Index, float fAmountX, float fAmountY, float fAmountZ)
  {
    _ptData[un8Index].csPos._tLocal.fX += fAmountX;
    _ptData[un8Index].csPos._tLocal.fY += fAmountY;
    _ptData[un8Index].csPos._tLocal.fZ += fAmountZ;
  }

  void CSprite2D::SetDirection(uint8_t un8Index, float fDirX, float fDirY, float fDirZ)
  {
    float fMag = sqrt(fDirX * fDirX + fDirY * fDirY + fDirZ * fDirZ);
    _ptData[un8Index].fX = fDirX / fMag;
    _ptData[un8Index].fY = fDirY / fMag;
    _ptData[un8Index].fZ = fDirZ / fMag;
  }

  void CSprite2D::RecalculateLighting()
  {
    _CalcIntenOmni_(_ptOmni);
    _CalcInten_();
  }

  void CSprite2D::RecalculateLighting(uint8_t un8Index)
  {
    _CalcIntenOmni_(un8Index, _ptOmni);
    _CalcInten_(un8Index);
  }

  void CSprite2D::RecalculateLighting(uint8_t un8StartIndex, uint8_t un8EndIndex)
  {
    _CalcIntenOmni_(un8StartIndex, un8EndIndex, _ptOmni);
    _CalcInten_(un8StartIndex, un8EndIndex);
  }

  void CSprite2D::Exclude(uint8_t un8Index, bool bExcluded)
  {
    _ptData[un8Index].bExcluded = bExcluded;
  }

} // namespace vge
#endif
