/** \file LowLevelSoundBufferFunctions.h */
// ############################################################################################
//  LowLevelSoundBufferFunctions.h
//  Definition of structures and prototype of low-level sound buffer functions
//  used by all CAudioEnv and CSoundBuffer implementations
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _AUDIO_ENVIRONMENT___LOW_LEVEL_SOUND_BUFFER_FUNCTIONS_H_
#define _AUDIO_ENVIRONMENT___LOW_LEVEL_SOUND_BUFFER_FUNCTIONS_H_

// VortexGE headers
#include "../Base/VBasics.h"


namespace vge {

  // Fixed point constants
  const uint32_t _WSTEP_PREC_     = uint32_t(8);
  const uint32_t _MAX_WBUFF_SIZE_ = uint32_t(1 << (31 - _WSTEP_PREC_));

  //!
  //! \brief Raw data formats
  //!
  //! \api pretty stable
  enum ERawFormat {
    ERF_Invalid,
    ERF_U8Mono, ERF_U8Stereo, ERF_S8Mono, ERF_S8Stereo,
    ERF_U16LEMono, ERF_U16LEStereo, ERF_S16LEMono, ERF_S16LEStereo,
    ERF_U16BEMono, ERF_U16BEStereo, ERF_S16BEMono, ERF_S16BEStereo
  };

  //!
  //! \brief Audio buffer format
  //!
  //! \api pretty stable
  enum ESoundBufferFormat {
    ESB_Undefined, ESB_Unknown,
    ESB_Signed8, ESB_Unsigned8,
    ESB_Signed16LE, ESB_Unsigned16LE,
    ESB_Signed16BE, ESB_Unsigned16BE
  };

  //!
  //! \brief Audio settings
  //!
  //! \api pretty stable
  struct TAudioSettings {
    ESoundBufferFormat eFormat;      // The audio data format
    int32_t            n32Frequency; // The sampling rate
    int32_t            n32Channels;  // The number of channels
  };

  //!
  //! \internal
  //! \brief The sound buffer format
  //!
  //! \api pretty stable
  struct _TAudioFormat {
    uint8_t  un8Bits;       // Bit depth (8 or 16)
    uint8_t  un8Sign;       // 0 = unsigned, 1 = signed
    uint8_t  un8Endian;     // 0 = LE, 1 = BE
    uint8_t  un8Channels;   // 1 = mono, 2 = stereo
    uint32_t un32Frequency; // Sampling frequency
  };

  //!
  //! \internal
  //! \brief The sound buffer data
  //!
  //! \api pretty stable
  struct _TSBData {
    u_char   *pucLeft;     // The left/mono channel buffer
    u_char   *pucRight;    // The right channel buffer
    u_char   *pucLeftMax;  // The left/mono channel buffer maximum address
    u_char   *pucRightMax; // The right channel buffer maximum address
    uint32_t un32PosMax;   // The maximum position (for frequency scalling)
    uint32_t un32PosStep;  // The position step (for frequency scalling)
  };

  //!
  //! \internal
  //! \brief The played sound buffer data
  //!
  //! \api pretty stable
  struct _TPlayedSBData {
    u_char   *pucLeft;     // The left/mono channel buffer
    u_char   *pucRight;    // The right channel buffer
    u_char   *pucLeftMax;  // The left/mono channel buffer maximum address
    u_char   *pucRightMax; // The right channel buffer maximum address
    bool     bPlaying;     // The "playing" flag
    bool     bLooping;     // The "looping" flag
    u_char   *pucLeftCur;  // The left/mono channel buffer current address
    u_char   *pucRightCur; // The right channel buffer current address
    uint32_t un32PosCur;   // The current position (for frequency scalling)
    uint32_t un32PosMax;   // The maximum position (for frequency scalling)
    uint32_t un32PosStep;  // The position step (for frequency scalling)
  };


  //!
  //! Use this to get the string name for the ESoundBufferFormat enumeration
  //!
  //! \api pretty stable
  extern VGE_EXPORT const char *VSoundBufferFormatName(ESoundBufferFormat eFormat);


  // Converter functions without frequency scalling
  // All functions for 16-bit sound buffer will only work for little-endian computers
  extern void __intexp_8Mn8Mn_(_TPlayedSBData &tData, u_char *pucDstBuff);
  extern void __intexp_8Mn8St_(_TPlayedSBData &tData, u_char *pucDstBuff);
  extern void __intexp_8Mn16Mn_(_TPlayedSBData &tData, u_char *pucDstBuff);
  extern void __intexp_8Mn16St_(_TPlayedSBData &tData, u_char *pucDstBuff);
  extern void __intexp_8St8St_(_TPlayedSBData &tData, u_char *pucDstBuff);
  extern void __intexp_8St16St_(_TPlayedSBData &tData, u_char *pucDstBuff);
  extern void __intexp_16Mn16Mn_(_TPlayedSBData &tData, u_char *pucDstBuff);
  extern void __intexp_16Mn16St_(_TPlayedSBData &tData, u_char *pucDstBuff);
  extern void __intexp_16St16St_(_TPlayedSBData &tData, u_char *pucDstBuff);

  // Converter functions with frequency scalling
  // All functions for 16-bit sound buffer will only work for little-endian computers
  extern void __intexp_X8Mn8Mn_(_TPlayedSBData &tData, u_char *pucDstBuff);
  extern void __intexp_X8Mn8St_(_TPlayedSBData &tData, u_char *pucDstBuff);
  extern void __intexp_X8Mn16Mn_(_TPlayedSBData &tData, u_char *pucDstBuff);
  extern void __intexp_X8Mn16St_(_TPlayedSBData &tData, u_char *pucDstBuff);
  extern void __intexp_X8St8St_(_TPlayedSBData &tData, u_char *pucDstBuff);
  extern void __intexp_X8St16St_(_TPlayedSBData &tData, u_char *pucDstBuff);
  extern void __intexp_X16Mn16Mn_(_TPlayedSBData &tData, u_char *pucDstBuff);
  extern void __intexp_X16Mn16St_(_TPlayedSBData &tData, u_char *pucDstBuff);
  extern void __intexp_X16St16St_(_TPlayedSBData &tData, u_char *pucDstBuff);

} // namespace vge
#endif
