/** \file MidiPlayer.h */
// ############################################################################################
//  MidiPlayer.h
//  Definition of CMidiPlayer interface class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _AUDIO_ENVIRONMENT___MIDI_PLAYER_H_
#define _AUDIO_ENVIRONMENT___MIDI_PLAYER_H_

// VortexGE headers
#include "../System/DataSinkSource.h"
#include "../System/Thread.h"
#include "AudioEnvironment.h"


namespace vge {

  //!
  //! \brief MIDI player interface class
  //!
  //! Affecting environment variables :
  //!   - VGE_MIDI_SYSTEM
  //!
  //! See the section "Environment Variables" form the main doumentation (not the
  //! API documentation) for more details.
  //!
  //! \api pretty stable
  class VGE_EXPORT CMidiPlayer : VCLASS_UNCOPYABLE {
    public:
      virtual ~CMidiPlayer() throw();

      // Auto instantiation functions
      //!
      //! Automatically search for usable MIDI player and return the
      //! corresponding CMidiPlayer instance.
      static CMidiPlayer& AutoDetect();

      // Initialization functions
      //!
      //! Initialize the MIDI player and attach it to the given
      //! audio environment instance.
      virtual void InitializeDevice(CAudioEnv *pcsVAudioEnv) = 0;

      // MIDI functions
      //!
      //! Play the loaded MIDI file.
      virtual void Play(bool bLooping = false) = 0;
      //!
      //! Check if the MIDI player is currently playing.
      virtual bool IsPlaying() const = 0;
      //!
      //! Stop the MIDI player playback.
      virtual void Stop() = 0;
      //!
      //! Set the playback volume of the loaded MIDI file.
      virtual void SetVolume(uint8_t un8VolProcent) = 0;

      // File I/O functions
      //!
      //! Load from MIDI (Musical Instrument Digital Interface) format.
      //! \param[in] source The name of the file (.mid) to be loaded or an instance of istream derived class.
      virtual void LoadMID(const CDataSource &source) = 0;

      // Friend classes
      friend class COssAudioEnv;
      friend class CAlsaAudioEnv;

    protected:
      // Make sure user can't construct this class directly
      inline CMidiPlayer() {}

      // Mixing functions
      virtual void _MixWithBuffer(unsigned char *pucBuffer, uint32_t un32BuffSize, volatile bool *pbForceExit) = 0;
  };

} // namespace vge
#endif
