// ############################################################################################
//  NonDummySoundBufferBaseClass.cc
//  Implementation of CSoundBufferCImp class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "../Base/MessageDisplayer.h"
#include "../System/SystemUtilities.h"
#include "NonDummySoundBufferBaseClass.h"
#include "WaveStructs.h"

// C++ headers
#include <fstream>


namespace vge {

// Compression type
#define _VSW_NONE_  0
#define _VSW_DUMMY_ 1

// VSW header
#define _VSW_ID1_ 'V'
#define _VSW_ID2_ 'S'
#define _VSW_ID3_ 'W'
#define _VSW_VER_ 100

struct _TVSWHeader {
  char          acID[3];       // File ID, must be "VSW"
  uint8_t       un8Version;    // File version
  _TAudioFormat tFormat;       // The audio format
  uint32_t      un32Size;      // Size of the uncompressed audio
  uint8_t       un8CompL;      // Compression type (left/mono channel)
  uint8_t       un8CompR;      // Compression type (right channel)
  uint32_t      un32SizeCompL; // Compressed audio size in bytes (left/mono channel)
  uint32_t      un32SizeCompR; // Compressed audio size in bytes (right channel)
  uint16_t      un16Reserved1; // Reserved, must be zero
  uint32_t      un32Reserved2; // Reserved, must be zero
} _attr_packed;


//
// Sound helper functions
//
static uint32_t _CompressSoundBuffer_(u_char *pucDstBuff, const u_char *pucSrcBuff, uint32_t un32SrcBuffSize);
static uint32_t _UncompressSoundBuffer_(u_char *pucDstBuff, const u_char *pucSrcBuff);


//
// The base of all non-dummy sound buffer classes
//
CSoundBufferCImp::CSoundBufferCImp()
: _pfnGD_(NULL)
{
  memset(&_tD, 0, sizeof(_tD));
}

CSoundBufferCImp::~CSoundBufferCImp() throw()
{
  _FreeBuffer_();
}

void CSoundBufferCImp::Play(bool bLooping)
{
  for(size_t cnt = 0; cnt < _MAX_SB_; cnt++) {
    if(CAudioEnvCImp::_instAudioEnv->_atPSB[cnt].pcsSB == NULL) {
      CAudioEnvCImp::_instAudioEnv->_atPSB[cnt].tData.pucLeft     = _tD.pucLeft;
      CAudioEnvCImp::_instAudioEnv->_atPSB[cnt].tData.pucRight    = _tD.pucRight;
      CAudioEnvCImp::_instAudioEnv->_atPSB[cnt].tData.pucLeftMax  = _tD.pucLeftMax;
      CAudioEnvCImp::_instAudioEnv->_atPSB[cnt].tData.pucRightMax = _tD.pucRightMax;
      CAudioEnvCImp::_instAudioEnv->_atPSB[cnt].tData.bPlaying    = true;
      CAudioEnvCImp::_instAudioEnv->_atPSB[cnt].tData.bLooping    = bLooping;
      CAudioEnvCImp::_instAudioEnv->_atPSB[cnt].tData.pucLeftCur  = _tD.pucLeft;
      CAudioEnvCImp::_instAudioEnv->_atPSB[cnt].tData.pucRightCur = _tD.pucRight;
      CAudioEnvCImp::_instAudioEnv->_atPSB[cnt].tData.un32PosMax  = _tD.un32PosMax;
      CAudioEnvCImp::_instAudioEnv->_atPSB[cnt].tData.un32PosStep = _tD.un32PosStep;
      CAudioEnvCImp::_instAudioEnv->_atPSB[cnt].tData.un32PosCur  = 0;
      CAudioEnvCImp::_instAudioEnv->_atPSB[cnt].pcsSB = this;
      return;
    }
  }
}

bool CSoundBufferCImp::IsPlaying() const
{
  for(size_t cnt = 0; cnt < _MAX_SB_; cnt++) {
    if(CAudioEnvCImp::_instAudioEnv->_atPSB[cnt].pcsSB == this)
      return(true);
  }
  return(false);
}

void CSoundBufferCImp::Stop()
{
  for(size_t cnt = 0; cnt < _MAX_SB_; cnt++) {
    if(CAudioEnvCImp::_instAudioEnv->_atPSB[cnt].pcsSB == this) {
      CAudioEnvCImp::_instAudioEnv->_atPSB[cnt].tData.bPlaying = false;
      return;
    }
  }
}

void CSoundBufferCImp::AdjustAmplitude(float fLeftGain, float fRightGain)
{
  // Check
  if(_tD.pucLeft == NULL)
    VError(VERR_NotInitialized, _NFL_, "No sound data to be amplified (this CSoundBufferCImp is uninitialized)");

  // Adjust amplitude
  if(_tF.un8Bits == 8) {
    const size_t size = size_t(_tD.pucLeftMax - _tD.pucLeft);
    int8_t *n8L = reinterpret_cast<int8_t*>(_tD.pucLeft);
    int8_t *n8R = reinterpret_cast<int8_t*>(_tD.pucRight);
    int32_t n32Tmp;
    for(size_t i = 0; i < size; ++i) {
      n32Tmp = int32_t(float(*n8L) * fLeftGain);
      if(n32Tmp > 127) n32Tmp = 127;
      if(n32Tmp < -128) n32Tmp = -128;
      *n8L++ = int8_t(n32Tmp);
      if(_tF.un8Channels == 2) {
        n32Tmp = int32_t(float(*n8R) * fRightGain);
        if(n32Tmp > 127) n32Tmp = 127;
        if(n32Tmp < -128) n32Tmp = -18;
        *n8R++ = int8_t(n32Tmp);
      }
    }
  }
  else {
    const size_t size = size_t(_tD.pucLeftMax - _tD.pucLeft) / 2;
    int16_t *n16L = reinterpret_cast<int16_t*>(_tD.pucLeft);
    int16_t *n16R = reinterpret_cast<int16_t*>(_tD.pucRight);
    int32_t  n32Tmp;
    for(size_t i = 0; i < size; ++i) {
      n32Tmp = int32_t(float(*n16L) * fLeftGain);
      if(n32Tmp > 32767) n32Tmp = 32767;
      if(n32Tmp < -32768) n32Tmp = -32768;
      *n16L++ = int16_t(n32Tmp);
      if(_tF.un8Channels == 2) {
        n32Tmp = int32_t(float(*n16R) * fRightGain);
        if(n32Tmp > 32767) n32Tmp = 32767;
        if(n32Tmp < -32768) n32Tmp = -32768;
        *n16R++ = int16_t(n32Tmp);
      }
    }
  }

}

void CSoundBufferCImp::ConvertMonoToStereo(bool bSilentRightChannel)
{
  // Check
  if(_tD.pucLeft == NULL)
    VError(VERR_NotInitialized, _NFL_, "No sound data to be converted (this CSoundBufferCImp is uninitialized)");
  if(_tF.un8Channels == 2) return;

  // Do conversion
  _Mono2Stereo_(bSilentRightChannel);
  _AssignCnvFunc_(_tD.un32PosStep != 0);
}

void CSoundBufferCImp::ConvertStereoToMono()
{
  // Check
  if(_tD.pucLeft == NULL)
    VError(VERR_NotInitialized, _NFL_, "No sound data to be converted (this CSoundBufferCImp is uninitialized)");
  if(_tF.un8Channels == 1) return;

  // Do conversion
  _Stereo2Mono_();
  _AssignCnvFunc_(_tD.un32PosStep != 0);
}

void CSoundBufferCImp::Convert16BitTo8Bit()
{
  // Check
  if(_tD.pucLeft == NULL)
    VError(VERR_NotInitialized, _NFL_, "No sound data to be converted (this CSoundBufferCImp is uninitialized)");
  if(_tF.un8Bits == 8) return;

  // Do conversion
  _16BitTo8Bit_();
  _AssignCnvFunc_(_tD.un32PosStep != 0);
}

void CSoundBufferCImp::SwapLRChannels()
{
  // Check
  if(_tD.pucLeft == NULL)
    VError(VERR_NotInitialized, _NFL_, "No sound data to be converted (this CSoundBufferCImp is uninitialized)");
  if(_tD.pucRight == NULL)
    VError(VERR_InvalidCall, _NFL_, "This CSoundBufferCImp is a mono sound bufer");
  if(IsPlaying())
    VError(VERR_InUse, _NFL_, "Can't operate on playing sound buffer");

  // Do swap
  u_char *pucLeft    = _tD.pucLeft;
  u_char *pucLeftMax = _tD.pucLeftMax;
  _tD.pucLeft      = _tD.pucRight;
  _tD.pucLeftMax   = _tD.pucRightMax;
  _tD.pucRight     = pucLeft;
  _tD.pucRightMax  = pucLeftMax;
}

void CSoundBufferCImp::LoadRAW(const CDataSource &source, ERawFormat eFormat, int32_t n32Frequency)
{
  // Get the stream
  const char *strmName = source.Name().c_str();
  pistream    &is      = source.PStream();

  // Check the parameter
  if(eFormat==ERF_Invalid || !n32Frequency) 
    VError(VERR_InvalidParameter, _NFL_, "The \"eFormat\" and/or \"n32Frequency\" parameter(s) contain(s) (an) invalid value(s)");

  // Determine the loading size
  size_t loadSize = source.Size();
  if(!loadSize) VError(VERR_InvalidParameter, _NFL_, "Couldn't determine the loading size\nIf the data source is an istream, the loading size should be explicitly set");
  size_t bsize = (loadSize + 3) & (~3);

  // Check the loading size
  if(bsize > _MAX_WBUFF_SIZE_)
    VError(VERR_NotSupported, _NFL_, "Can't load raw audio data of which channel data size > %u bytes", _MAX_WBUFF_SIZE_);

  // Calculate some informations
  _TAudioFormat tF;
  _ExtractFormat_(&tF, eFormat);
  tF.un32Frequency = n32Frequency;

  if(tF.un8Channels == 2) bsize >>= 1;

  // Allocate memory for buffer
  u_char *pucL = reinterpret_cast<u_char*>(malloc(bsize));
  if(pucL == NULL)
    VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for sound buffer");

  u_char *pucR = NULL;
  if(tF.un8Channels == 2) {
    pucR = reinterpret_cast<u_char*>(malloc(bsize));
    if(pucR == NULL) {
      free(pucL);
      VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for sound buffer");
    }
  }

  // Silent the buffer
  if(tF.un8Sign) {
    memset(pucL, 0, bsize);
    if(pucR) memset(pucL, 0, bsize);
  }
  else if (tF.un8Bits == 8) {
    memset(pucL, 0x80, bsize);
    if(pucR) memset(pucL, 0x80, bsize);
  }
  else {
    for(size_t i = 1; i < bsize; i += 2) pucL[i] = 0x80;
    if(pucR) for(size_t i = 1; i < bsize; i+=2) pucR[i] = 0x80;
  }

  // Read the file
  try {
    u_char  *pucTmpL = pucL;
    u_char  *pucTmpR = pucR;
    uint32_t un32Cnt = 0;
    while(un32Cnt < uint32_t(loadSize)) {
      if(tF.un8Bits == 8) {
        is.read(reinterpret_cast<char*>(pucTmpL), 1);
        pucTmpL++;
        un32Cnt++;
      }
      else {
        is.read(reinterpret_cast<char*>(pucTmpL), 2);
        pucTmpL += 2;
        un32Cnt += 2;
      }
      if(tF.un8Channels == 2) {
        if(tF.un8Bits == 8) {
          is.read(reinterpret_cast<char*>(pucTmpR), 1);
          pucTmpR++;
          un32Cnt++;
        }
        else {
          is.read(reinterpret_cast<char*>(pucTmpR), 2);
          pucTmpR += 2;
          un32Cnt += 2;
        }
      }
    }
  } 
  catch(ios_base::failure&) {
    free(pucL);
    if(_tF.un8Channels == 2) free(pucR);
    VError(VERR_ReadStream, _NFL_, "Couldn't read stream \"%s\"", strmName);
  }
  catch(...) {
    free(pucL);
    if(_tF.un8Channels == 2) free(pucR);
    throw;
  }

  // Free previous sound (if any)
  _FreeBuffer_();
  memset(&_tD, 0, sizeof(_tD));
  _pfnGD_ = NULL;

  // Copy data
  memcpy(&_tF, &tF, sizeof(_tF));

  // Calculate some parameters
  if(_tF.un8Channels == 2) {
    _tD.pucLeft     = pucL;
    _tD.pucLeftMax  = _tD.pucLeft + (loadSize >> 1) - 1;
    _tD.pucRight    = pucR;
    _tD.pucRightMax = _tD.pucRight + (loadSize >> 1) - 1;
  }
  else {
    _tD.pucLeft     = pucL;
    _tD.pucLeftMax  = _tD.pucLeft + loadSize - 1;
    _tD.pucRight    = NULL;
    _tD.pucRightMax = NULL;
  }
  if(_tF.un32Frequency != _primaryBufferFrequency_()) {
    if(_tF.un8Channels == 2)
      _tD.un32PosMax  = (loadSize >> 1) << _WSTEP_PREC_;
    else
      _tD.un32PosMax  = loadSize << _WSTEP_PREC_;
    _tD.un32PosStep = uint32_t(float(_tF.un32Frequency) / float(_primaryBufferFrequency_()) * float(1 << _WSTEP_PREC_));
    if(_tF.un8Bits == 16) _tD.un32PosStep <<= 1;
  }
  else {
    _tD.un32PosMax  = 0;
    _tD.un32PosStep = 0;
  }

  // Finalize the loading process
  _FinalizeLoad_();
 }

void CSoundBufferCImp::LoadWAV(const CDataSource &source)
{
  // Note that this function has some limitations :
  //   1. Only little-endian .wav files are supported
  //   2. Only 8-bit unsigned and 16-bit signed formats are supported

  // Get the stream
  const char *strmName = source.Name().c_str();
  pistream    &is      = source.PStream();

  // Read the header
  _TCWFPCMHEADER tHeader;
  is.read(reinterpret_cast<char*>(&tHeader), sizeof(tHeader));
  if(is.fail())
    VError(VERR_ReadStream, _NFL_, "Couldn't read WAV header from stream \"%s\"", strmName);

  // Check header
  if(memcmp(tHeader.tChunk.aucChunkID, "RIFF", 4))
    VError(VERR_InvalidStream, _NFL_, "\"%s\" is an invalid RIFF stream", strmName);
  if(memcmp(tHeader.tChunk.aucFormat, "WAVE", 4))
    VError(VERR_InvalidStream, _NFL_, "\"%s\" is an invalid WAVE stream", strmName);
  if(memcmp(tHeader.tFormat.aucSubChunk1ID, "fmt ", 4))
    VError(VERR_InvalidStream, _NFL_, "\"%s\" has an invalid \"fmt \" sub-chunk ID", strmName);
  if(memcmp(tHeader.tData.aucSubChunk2ID, "data", 4))
    VError(VERR_InvalidStream, _NFL_, "\"%s\" has an invalid \"data\" sub-chunk ID", strmName);
  if(tHeader.tChunk.un32ChunkSize != (20 + tHeader.tFormat.un32SubChunk1Size + tHeader.tData.un32SubChunk2Size))
    VError(VERR_InvalidStream, _NFL_, "\"%s\" has an invalid chunk / sub-chunk size", strmName);
  if(tHeader.tFormat.un32SubChunk1Size != 16)
    VError(VERR_InvalidStream, _NFL_, "\"%s\" has an invalid \"fmt \" sub-chunk size, only PCM format supported", strmName);
  if(tHeader.tFormat.un16AudioFormat != 1)
    VError(VERR_NotSupported, _NFL_, "\"%s\" has an invalid audio format, only PCM format supported", strmName);
  if(tHeader.tFormat.un16Channels > 2)
    VError(VERR_NotSupported, _NFL_, "\"%s\" has more than 2 audio channels; only 1 and 2 channels format are supported", strmName);
  if((tHeader.tFormat.un16BitsPerSample != 8) && (tHeader.tFormat.un16BitsPerSample != 16))
    VError(VERR_NotSupported, _NFL_, "\"%s\" has %d bps, only 8 and 16 bps supported", strmName, tHeader.tFormat.un16BitsPerSample);
  if(tHeader.tFormat.un32ByteRate != (tHeader.tFormat.un32SampleRate * tHeader.tFormat.un16Channels * tHeader.tFormat.un16BitsPerSample / 8))
    VError(VERR_InvalidStream, _NFL_, "\"%s\" has an invalid data", strmName);
  if(tHeader.tFormat.un16BlockAlign != (tHeader.tFormat.un16Channels * tHeader.tFormat.un16BitsPerSample / 8))
    VError(VERR_InvalidStream, _NFL_, "\"%s\" has an invalid data", strmName);

  // Calculate some information
  _TAudioFormat tF;
  size_t size = 0;
  size_t sget = 0;
  tF.un8Bits       = tHeader.tFormat.un16BitsPerSample;
  tF.un8Sign       = (tF.un8Bits == 16) ? 1 : 0;
  tF.un8Endian     = 0;
  tF.un8Channels   = tHeader.tFormat.un16Channels;
  tF.un32Frequency = tHeader.tFormat.un32SampleRate;
  if(tF.un8Channels == 1)
    size = tHeader.tData.un32SubChunk2Size;
  else
    size = tHeader.tData.un32SubChunk2Size / 2;
  if(tF.un8Bits == 8)
    sget = size;
  else
    sget = size / 2;

  // Allocate memory for buffer
  u_char *pucL = reinterpret_cast<u_char*>(malloc(size));
  if(pucL == NULL)
    VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for sound buffer");

  u_char *pucR = NULL;
  if(tF.un8Channels == 2) {
    pucR = reinterpret_cast<u_char*>(malloc(size));
    if(pucR == NULL) {
      free(pucL);
      VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for sound buffer");
    }
  }

  // Read sound data
  is.exceptions(ios::badbit | ios::failbit | ios::eofbit);
  try {
    u_char *pucLt = pucL;
    u_char *pucRt = pucR;
    for(size_t i = 0; i < sget; ++i) {
      // Left / mono channel
      if(tF.un8Bits == 8) {
        is.read(reinterpret_cast<char*>(pucLt), 1);
        pucLt++;
      }
      else {
        is.read(reinterpret_cast<char*>(pucLt), 2);
        pucLt += 2;
      }
  
      // Right channel
      if(tF.un8Channels == 2) {
        if(tF.un8Bits == 8) {
          is.read(reinterpret_cast<char*>(pucRt), 1);
          pucRt++;
        }
        else {
          is.read(reinterpret_cast<char*>(pucRt), 2);
          pucRt += 2;
        }
      }
    }
  }
  catch(ios_base::failure&) {
    free(pucL);
    if(_tF.un8Channels == 2) free(pucR);
    VError(VERR_ReadStream, _NFL_, "Couldn't read stream \"%s\"", strmName);
  }
  catch(...) {
    free(pucL);
    if(_tF.un8Channels == 2) free(pucR);
    throw;
  }

  // Free previous sound (if any)
  _FreeBuffer_();
  memset(&_tD, 0, sizeof(_tD));
  _pfnGD_ = NULL;

  // Copy data
  memcpy(&_tF, &tF, sizeof(_tF));

  // Calculate some parameters
  if(_tF.un8Channels == 2) {
    _tD.pucLeft     = pucL;
    _tD.pucLeftMax  = _tD.pucLeft + size - 1;
    _tD.pucRight    = pucR;
    _tD.pucRightMax = _tD.pucRight + size - 1;
  }
  else {
    _tD.pucLeft     = pucL;
    _tD.pucLeftMax  = _tD.pucLeft + size - 1;
    _tD.pucRight    = NULL;
    _tD.pucRightMax = NULL;
  }
  if(_tF.un32Frequency != _primaryBufferFrequency_()) {
    _tD.un32PosMax  = (size - 1) << _WSTEP_PREC_;
    _tD.un32PosStep = uint32_t(float(_tF.un32Frequency) / float(_primaryBufferFrequency_()) * float(1 << _WSTEP_PREC_));
    if(_tF.un8Bits == 16) _tD.un32PosStep <<= 1;
  }
  else {
    _tD.un32PosMax  = 0;
    _tD.un32PosStep = 0;
  }

  // Finalize the loading process
  _FinalizeLoad_();
}

void CSoundBufferCImp::LoadVSW(const CDataSource &source)
{
  // Get the stream
  const char *strmName = source.Name().c_str();
  pistream    &is      = source.PStream();

  // Read the header
  _TVSWHeader tHeader;
  is.read(reinterpret_cast<char*>(&tHeader), sizeof(tHeader));
  if(is.fail())
    VError(VERR_ReadStream, _NFL_, "Couldn't read VSW header from stream \"%s\"", strmName);

  // Check header
  if((tHeader.acID[0] != _VSW_ID1_) ||
     (tHeader.acID[1] != _VSW_ID2_) || (tHeader.acID[2] != _VSW_ID3_))
    VError(VERR_InvalidStream, _NFL_, "\"%s\" is an invalid VSW stream", strmName);
  if(tHeader.un8Version != _VSW_VER_)
    VError(VERR_IncorrectVersion, _NFL_, "\"%s\" has an incorrect version", strmName);
  if((tHeader.un8CompL != _VSW_NONE_) && (tHeader.un8CompL != _VSW_DUMMY_))
    VError(VERR_InvalidHeaderData, _NFL_, "Invalid compression type found in stream \"%s\"", strmName);
  if(tHeader.un16Reserved1 || tHeader.un32Reserved2)
    VError(VERR_InvalidHeaderData, _NFL_, "Invalid header data found in stream \"%s\"", strmName);

  // Allocate memory for temporary buffer
  vector<u_char> pucLoad(max(tHeader.un32SizeCompR, tHeader.un32SizeCompR));

  // Allocate memory for buffer
  u_char *pucL = reinterpret_cast<u_char*>(malloc(tHeader.un32Size));
  if(pucL == NULL)
    VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for sound buffer");

  u_char *pucR = NULL;
  if(tHeader.tFormat.un8Channels == 2) {
    pucR = reinterpret_cast<u_char*>(malloc(tHeader.un32Size));
    if(pucR == NULL) {
      free(pucL);
      VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for sound buffer");
    }
  }

  // Read & uncompress left / mono channel
  is.read(reinterpret_cast<char*>(&pucLoad[0]), tHeader.un32SizeCompL);
  if(is.fail()) {
    free(pucL);
    if(tHeader.tFormat.un8Channels == 2) free(pucR);
    VError(VERR_ReadStream, _NFL_, "Couldn't read left channel data from stream \"%s\"", strmName);
  }
  if(tHeader.un8CompL == _VSW_DUMMY_) {
    if(!_UncompressSoundBuffer_(pucL,  &pucLoad[0])) {
      free(pucL);
      if(tHeader.tFormat.un8Channels == 2) free(pucR);
      VError(VERR_UncompressData, _NFL_, "Couldn't uncompress left channel data");
    }
  }
  else
    copy_b(pucLoad, tHeader.un32SizeCompL, pucL);

  // Read & uncompress right channel
  is.read(reinterpret_cast<char*>(&pucLoad[0]), tHeader.un32SizeCompR);
  if(is.fail()) {
    free(pucL);
    if(tHeader.tFormat.un8Channels == 2) free(pucR);
    VError(VERR_ReadStream, _NFL_, "Couldn't read right channel data from stream \"%s\"", strmName);
  }
  if(tHeader.un8CompR == _VSW_DUMMY_) {
    if(!_UncompressSoundBuffer_(pucR,  &pucLoad[0])) {
      free(pucL);
      if(tHeader.tFormat.un8Channels == 2) free(pucR);
      VError(VERR_UncompressData, _NFL_, "Couldn't uncompress right channel data");
    }
  }
  else
    copy_b(pucLoad, tHeader.un32SizeCompR, pucR);

  // Free previous sound (if any)
  _FreeBuffer_();
  memset(&_tD, 0, sizeof(_tD));
  _pfnGD_ = NULL;

  // Copy data
  memcpy(&_tF, &tHeader.tFormat, sizeof(_tF));

  // Calculate some parameters
  if(_tF.un8Channels == 2) {
    _tD.pucLeft     = pucL;
    _tD.pucLeftMax  = _tD.pucLeft + tHeader.un32Size - 1;
    _tD.pucRight    = pucR;
    _tD.pucRightMax = _tD.pucRight + tHeader.un32Size - 1;
  }
  else {
    _tD.pucLeft     = pucL;
    _tD.pucLeftMax  = _tD.pucLeft + tHeader.un32Size - 1;
    _tD.pucRight    = NULL;
    _tD.pucRightMax = NULL;
  }
  if(_tF.un32Frequency != _primaryBufferFrequency_()) {
    _tD.un32PosMax  = tHeader.un32Size << _WSTEP_PREC_;
    _tD.un32PosStep = uint32_t(float(_tF.un32Frequency) / float(_primaryBufferFrequency_()) * float(1 << _WSTEP_PREC_));
    if(_tF.un8Bits == 16) _tD.un32PosStep <<= 1;
  }
  else {
    _tD.un32PosMax  = 0;
    _tD.un32PosStep = 0;
  }

  // Finalize the loading process
  _FinalizeLoad_();
}

void CSoundBufferCImp::SaveVSW(const CDataSink &sink) const
{
  // Check
  if(_tD.pucLeft == NULL)
    VError(VERR_NotInitialized, _NFL_, "No data to be saved (this CSoundBufferCImp is not initialized)");

  // Allocate memory for temporary buffers
  uint32_t mSize = _tD.pucLeftMax - _tD.pucLeft + 1;
  vector<u_char> pucCompL(mSize), pucCompR;
  if(_tF.un8Channels == 2) pucCompR.resize(mSize);

  // Compress sound buffer
  uint32_t lSize = _CompressSoundBuffer_(&pucCompL[0], _tD.pucLeft, mSize);
  uint32_t rSize = 0;
  if(_tF.un8Channels == 2)
    rSize = _CompressSoundBuffer_(&pucCompR[0], _tD.pucRight, mSize);

  // Make header
  _TVSWHeader tHeader;
  memset(&tHeader, 0, sizeof(tHeader));
  tHeader.acID[0]    = _VSW_ID1_;
  tHeader.acID[1]    = _VSW_ID2_;
  tHeader.acID[2]    = _VSW_ID3_;
  tHeader.un8Version = _VSW_VER_;
  memcpy(&tHeader.tFormat, &_tF, sizeof(_tF));
  tHeader.un32Size   = mSize;
  if(lSize) {
    tHeader.un8CompL      = _VSW_DUMMY_;
    tHeader.un32SizeCompL = lSize;
  }
  else {
    tHeader.un8CompL      = _VSW_NONE_;
    tHeader.un32SizeCompL = mSize;
  }
  if(_tF.un8Channels == 2) {
    if(rSize) {
      tHeader.un8CompR      = _VSW_DUMMY_;
      tHeader.un32SizeCompR = rSize;
    }
    else {
      tHeader.un8CompR      = _VSW_NONE_;
      tHeader.un32SizeCompR = mSize;
    }
  }
  else {
    tHeader.un8CompR      = _VSW_NONE_;
    tHeader.un32SizeCompR = 0;
  }

  // Get the stream
  const char *strmName = sink.Name().c_str();
  ostream    &os       = sink.Stream();
  os.exceptions(ios::badbit | ios::failbit | ios::eofbit);

  // Write the data
  try {
    // Write the header
    os.write(reinterpret_cast<const char*>(&tHeader), sizeof(tHeader));
    // Write left channel data
    if(lSize)
      os.write(reinterpret_cast<const char*>(&pucCompL[0]), lSize);
    else
      os.write(reinterpret_cast<const char*>(_tD.pucLeft), mSize);
    // Write right channel data
    if(_tF.un8Channels == 2) {
      if(rSize)
        os.write(reinterpret_cast<const char*>(&pucCompR[0]), rSize);
      else
        os.write(reinterpret_cast<const char*>(_tD.pucRight), mSize);
    }
  }
  catch(ios_base::failure&) {
    VError(VERR_WriteStream, _NFL_, "Couldn't write stream \"%s\"", strmName);
  }
  catch(...) {
    throw;
  }
}

void CSoundBufferCImp::_ExtractFormat_(_TAudioFormat *ptFormat, ERawFormat eFormat)
{
  if(eFormat == ERF_U8Mono) {
    ptFormat->un8Bits     = 8;
    ptFormat->un8Sign     = 0;
    ptFormat->un8Endian   = 0;
    ptFormat->un8Channels = 1;
  }
  else if(eFormat == ERF_U8Stereo) {
    ptFormat->un8Bits     = 8;
    ptFormat->un8Sign     = 0;
    ptFormat->un8Endian   = 0;
    ptFormat->un8Channels = 2;
  }
  else if(eFormat == ERF_S8Mono) {
    ptFormat->un8Bits     = 8;
    ptFormat->un8Sign     = 1;
    ptFormat->un8Endian   = 0;
    ptFormat->un8Channels = 1;
  }
  else if(eFormat == ERF_S8Stereo) {
    ptFormat->un8Bits     = 8;
    ptFormat->un8Sign     = 1;
    ptFormat->un8Endian   = 0;
    ptFormat->un8Channels = 2;
  }
  else if(eFormat == ERF_U16LEMono) {
    ptFormat->un8Bits     = 16;
    ptFormat->un8Sign     = 0;
    ptFormat->un8Endian   = 0;
    ptFormat->un8Channels = 1;
  }
  else if(eFormat == ERF_U16LEStereo) {
    ptFormat->un8Bits     = 16;
    ptFormat->un8Sign     = 0;
    ptFormat->un8Endian   = 0;
    ptFormat->un8Channels = 2;
  }
  else if(eFormat == ERF_S16LEMono) {
    ptFormat->un8Bits     = 16;
    ptFormat->un8Sign     = 1;
    ptFormat->un8Endian   = 0;
    ptFormat->un8Channels = 1;
  }
  else if(eFormat == ERF_S16LEStereo) {
    ptFormat->un8Bits     = 16;
    ptFormat->un8Sign     = 1;
    ptFormat->un8Endian   = 0;
    ptFormat->un8Channels = 2;
  }
  else if(eFormat == ERF_U16BEMono) {
    ptFormat->un8Bits     = 16;
    ptFormat->un8Sign     = 0;
    ptFormat->un8Endian   = 1;
    ptFormat->un8Channels = 1;
  }
  else if(eFormat == ERF_U16BEStereo) {
    ptFormat->un8Bits     = 16;
    ptFormat->un8Sign     = 0;
    ptFormat->un8Endian   = 1;
    ptFormat->un8Channels = 2;
  }
  else if(eFormat == ERF_S16BEMono) {
    ptFormat->un8Bits     = 16;
    ptFormat->un8Sign     = 1;
    ptFormat->un8Endian   = 1;
    ptFormat->un8Channels = 1;
  }
  else if(eFormat == ERF_S16BEStereo) {
    ptFormat->un8Bits     = 16;
    ptFormat->un8Sign     = 1;
    ptFormat->un8Endian   = 1;
    ptFormat->un8Channels = 2;
  }
}

void CSoundBufferCImp::_SwapEndian_()
{
  // Swap the endians
  size_t size = size_t(_tD.pucLeftMax - _tD.pucLeft);
  if(_tF.un8Channels == 2) {
    for(size_t i = 0; i < size; i+=2) {
      u_char ucTmp = _tD.pucLeft[i];
      _tD.pucLeft[i  ] = _tD.pucLeft[i+1];
      _tD.pucLeft[i+1] = ucTmp;
      ucTmp = _tD.pucRight[i];
      _tD.pucRight[i  ] = _tD.pucRight[i+1];
      _tD.pucRight[i+1] = ucTmp;
    }
  }
  else {
    for(size_t i = 0; i < size; i+=2) {
      u_char ucTmp = _tD.pucLeft[i];
      _tD.pucLeft[i  ] = _tD.pucLeft[i+1];
      _tD.pucLeft[i+1] = ucTmp;
    }
  }

  // Adjust the format
  static const uint8_t sw[2] = {1, 0};
  _tF.un8Endian = sw[_tF.un8Endian];
}

void CSoundBufferCImp::_InvertSign_()
{
  // Preparation
  const uint32_t msbOffset = (_tF.un8Bits == 16) && (_tF.un8Endian == 0);
  const uint32_t incOffset = _tF.un8Bits / 8;

  // Invert the L channel
  for(u_char *i = (_tD.pucLeft + msbOffset); i <= _tD.pucLeftMax; i += incOffset)
    *i ^= 0x80;

  // Invert the R channel
  if (_tF.un8Channels == 2) {
    for(u_char *i = (_tD.pucRight + msbOffset); i <= _tD.pucRightMax; i += incOffset)
      *i ^= 0x80;
  }

  // Adjust the format
  _tF.un8Sign = !_tF.un8Sign;
}

void CSoundBufferCImp::_Mono2Stereo_(bool bSilentRightChannel)
{
  // Allocate memory for the right-channel buffer
  uint32_t un32Size = uint32_t(_tD.pucLeftMax - _tD.pucLeft + 1);
  _tD.pucRight = reinterpret_cast<u_char*>(malloc(un32Size));
  if(_tD.pucRight == NULL)
    VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for new sound buffer");

  // Copy data
  if(bSilentRightChannel)
    memset(_tD.pucRight, 0, un32Size);
  else
    memcpy(_tD.pucRight, _tD.pucLeft, un32Size);

  // Adjust the format
  _tD.pucRightMax = _tD.pucRight + un32Size - 1;
  _tF.un8Channels = 2;
}

void CSoundBufferCImp::_Stereo2Mono_()
{
  bool   bAllMuted = true;
  size_t size;

  // Convert stereo to mono (for signed sound buffer only)
  if(_tF.un8Bits == 8) {
    size = size_t(_tD.pucLeftMax - _tD.pucLeft + 1);
    int8_t *n8L = reinterpret_cast<int8_t*>(_tD.pucLeft);
    int8_t *n8R = reinterpret_cast<int8_t*>(_tD.pucRight);
    int32_t n32Tmp;
    for(size_t i = 0; i < size; ++i) {
      n32Tmp = int32_t(*n8L) + int32_t(*n8R++);
      *n8L   = int8_t(n32Tmp / 2);
      if(*n8L != 0) bAllMuted = false;
      n8L++;
    }
  }
  else {
    size = size_t(_tD.pucLeftMax - _tD.pucLeft + 1) / 2;
    int16_t *n16L = reinterpret_cast<int16_t*>(_tD.pucLeft);
    int16_t *n16R = reinterpret_cast<int16_t*>(_tD.pucRight);
    int32_t  n32Tmp;
    for(size_t i = 0; i < size; ++i) {
      n32Tmp = int32_t(*n16L) + int32_t(*n16R++);
      *n16L  = int16_t(n32Tmp / 2);
      if(*n16L != 0) bAllMuted = false;
      n16L++;
    }
  }
  if(bAllMuted) memcpy(_tD.pucLeft, _tD.pucRight, size);

  // Adjust the format
  free(_tD.pucRight);
  _tD.pucRight    = NULL;
  _tD.pucRightMax = NULL;
  _tF.un8Channels = 1;
}

void CSoundBufferCImp::_16BitTo8Bit_()
{
  // Allocate memory
  size_t  size = size_t(_tD.pucLeftMax - _tD.pucLeft) / 2;
  u_char *pucL = reinterpret_cast<u_char*>(malloc(size));
  u_char *pucR = NULL;
  if(pucL == NULL)
    VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for temporary sound buffer");
  if(_tF.un8Channels == 2) {
    pucR = reinterpret_cast<u_char*>(malloc(size));
    if(pucR == NULL) {
      free(pucL);
      VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for temporary sound buffer");
    }
  }

  // Convert 16-bit to 8-bit (can be used for signed and unsigned sound buffer)
  if(_tF.un8Sign == 0) {
    uint8_t  *un8L  = reinterpret_cast<uint8_t*>(pucL);
    uint8_t  *un8R  = reinterpret_cast<uint8_t*>(pucR);
    uint16_t *un16L = reinterpret_cast<uint16_t*>(_tD.pucLeft);
    uint16_t *un16R = reinterpret_cast<uint16_t*>(_tD.pucRight);
    for(size_t i = 0; i < size; ++i) {
      *un8L++ = uint8_t(*un16L++ / 256);
      if(_tF.un8Channels == 2) *un8R++ = uint8_t(*un16R++ / 256);
    }
  }
  else {
    int8_t  *n8L  = reinterpret_cast<int8_t*>(pucL);
    int8_t  *n8R  = reinterpret_cast<int8_t*>(pucR);
    int16_t *n16L = reinterpret_cast<int16_t*>(_tD.pucLeft);
    int16_t *n16R = reinterpret_cast<int16_t*>(_tD.pucRight);
    for(size_t i = 0; i < size; ++i) {
      *n8L++ = uint8_t(*n16L++ / 256);
      if(_tF.un8Channels == 2) *n8R++ = uint8_t(*n16R++ / 256);
    }
  }

  // Free old buffer
  free(_tD.pucLeft);
  if(_tD.pucRight) free(_tD.pucRight);

  // Adjust the buffer
  _tD.pucLeft     = pucL;
  _tD.pucLeftMax  = _tD.pucLeft + size - 1;
  if(_tF.un8Channels == 2) {
    _tD.pucRight    = pucR;
    _tD.pucRightMax = _tD.pucRight + size - 1;
  }
  if(_tD.un32PosMax) _tD.un32PosMax  = _tD.un32PosMax >> 1;

  // Adjust the format
  _tF.un8Bits = 8;
}

void CSoundBufferCImp::_AssignCnvFunc_(bool bDifFreq)
{
  /*
     Table layout :

     I                 B C  B C
     0 | 0 0 0 0 <---  8 1  8 1 | __intexp_8Mn8Mn_
     1 | 0 0 0 1 <---  8 1  8 2 | __intexp_8Mn8St_
     2 | 0 0 1 0 <---  8 1 16 1 | __intexp_8Mn16Mn_
     3 | 0 0 1 1 <---  8 1 16 2 | __intexp_8Mn16St_
     4                            NULL
     5 | 0 1 0 1 <---  8 2  8 2 | __intexp_8St8St_
     6                            NULL
     7 | 0 1 1 1 <---  8 2 16 2 | __intexp_8St16St_
     8                            NULL
     9                            NULL
    10 | 1 0 1 0 <--- 16 1 16 1 | __intexp_16Mn16Mn_
    11 | 1 0 1 1 <--- 16 1 16 2 | __intexp_16Mn16St_
    12                            NULL
    13                            NULL
    14                            NULL
    15 | 1 1 1 1 <--- 16 2 16 2 | __intexp_16St16St_
  */

  // Table if the sampling frequency is the same
  static const _XCnvFunc_ sf[16] = { __intexp_8Mn8Mn_,
                                     __intexp_8Mn8St_,
                                     __intexp_8Mn16Mn_,
                                     __intexp_8Mn16St_,   NULL,
                                     __intexp_8St8St_,    NULL,
                                     __intexp_8St16St_,   NULL, NULL,
                                     __intexp_16Mn16Mn_,
                                     __intexp_16Mn16St_,  NULL, NULL, NULL,
                                     __intexp_16St16St_ };

  // Table if the sampling frequency is different
  static const _XCnvFunc_ df[16] = { __intexp_X8Mn8Mn_,
                                     __intexp_X8Mn8St_,
                                     __intexp_X8Mn16Mn_,
                                     __intexp_X8Mn16St_,   NULL,
                                     __intexp_X8St8St_,    NULL,
                                     __intexp_X8St16St_,   NULL, NULL,
                                     __intexp_X16Mn16Mn_,
                                     __intexp_X16Mn16St_,  NULL, NULL, NULL,
                                     __intexp_X16St16St_ };

  // Assign the appropriate conversion function
  uint8_t a = _tF.un8Bits/8 - 1;
  uint8_t b = _tF.un8Channels - 1;
  uint8_t c = _primaryBufferBits_()/8 - 1;
  uint8_t d = _primaryBufferChannels_() - 1;
  uint8_t e = (a << 3) | (b << 2) | (c << 1) | d;

  if(!bDifFreq) _pfnGD_ = sf[e];
  else          _pfnGD_ = df[e];

  // Should never get here
  if(!_pfnGD_)
    VError(VERR_NotSupported, _NFL_, "No audio format conversion routine can be applied; playback of this sound buffer is not possible");
}

void CSoundBufferCImp::_FinalizeLoad_()
{
  // Assuming that this system is a little-endian (x86),
  // swap the bytes if the sound buffer's is 16-bit and not little-endian
  if((_tF.un8Bits == 16) && (_tF.un8Endian != 0)) _SwapEndian_();

  // The VortexGE sound mixer always use signed numbers,
  // make sure that the sound buffer is also signed
  if(_tF.un8Sign == 0) _InvertSign_();

  // If the primary sound buffer is mono, there is no need for us to have
  // a stereo sound buffer, so convert it to mono
  if((_tF.un8Channels == 1) && (_primaryBufferChannels_() == 0)) _Stereo2Mono_();

  // If the primary sound buffer is 8-bit, there is no need for us to have
  // an 16-bit sound buffer, so convert it to 8-bit
  if((_tF.un8Bits == 16) && (_primaryBufferBits_() == 8)) {
    try {
      _16BitTo8Bit_();
    }
    catch (...) {
      free(_tD.pucLeft);
      if(_tD.pucRight) free(_tD.pucRight);
      memset(&_tD, 0, sizeof(_tD));
      memset(&_tF, 0, sizeof(_tF));
      throw;
    }
  }

  // Assign real-time conversion function
  _AssignCnvFunc_(_tD.un32PosStep != 0);
}

void CSoundBufferCImp::_FreeBuffer_()
{
  // If this sound buffer is playing, stop it first
  if(CAudioEnvCImp::_instAudioEnv) {
    for(size_t cnt = 0; cnt < _MAX_SB_; cnt++) {
      if(CAudioEnvCImp::_instAudioEnv->_atPSB[cnt].pcsSB == this) {
        VInfo(_NFL_, "Removing playing sound buffer %p from list index #%zd", this, cnt);
        CAudioEnvCImp::_instAudioEnv->_atPSB[cnt].tData.bPlaying = false;
        CAudioEnvCImp::_instAudioEnv->_atPSB[cnt].tData.bLooping = false;
        if(CAudioEnvCImp::_instAudioEnv->_csTWriter.IsRunning()) usleep(100000);
        CAudioEnvCImp::_instAudioEnv->_atPSB[cnt].pcsSB = NULL;
        break;
      }
    }
    if(CAudioEnvCImp::_instAudioEnv->_csTWriter.IsRunning()) usleep(100000);
  }

  // Free memory
  if(_tD.pucLeft)  free(_tD.pucLeft);
  if(_tD.pucRight) free(_tD.pucRight);
}


//
// Local functions implementation
//
static uint32_t _CompressSoundBuffer_(u_char *pucDstBuff, const u_char *pucSrcBuff, uint32_t un32SrcBuffSize)
{
  // Not implemented yet
  pucDstBuff      = pucDstBuff;
  pucSrcBuff      = pucSrcBuff;
  un32SrcBuffSize = un32SrcBuffSize;
  return(0);
}

static uint32_t _UncompressSoundBuffer_(u_char *pucDstBuff, const u_char *pucSrcBuff)
{
  // Not implemented yet
  pucDstBuff = pucDstBuff;
  pucSrcBuff = pucSrcBuff;
  return(0);
}

} // namespace vge
