// ############################################################################################
//  Exception.cc
//  Implementation of VortexGE exception classes
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "Exception.h"


namespace vge {

//
// Source information class
//
CSourceInfo::CSourceInfo(const char *pcFunction, const char *pcFile, int32_t n32Line)
: _csFunction(pcFunction ? pcFunction : ""), _csFile(pcFile ? pcFile : ""), _n32Line(n32Line)
{}

CSourceInfo::CSourceInfo(const CSourceInfo &csSInfo)
: _csFunction(csSInfo._csFunction), _csFile(csSInfo._csFile), _n32Line(csSInfo._n32Line)
{}

CSourceInfo::~CSourceInfo() throw()
{
}

CSourceInfo &CSourceInfo::operator=(const CSourceInfo &csSInfo)
{
  _csFunction = csSInfo._csFunction;
  _csFile     = csSInfo._csFile;
  _n32Line    = csSInfo._n32Line;
  return(*this);
}


//
// Generic exception class
//
CException::CException(EERRCode eCode, const char *pcWhat, const CSourceInfo &csLocation)
: _csWhat(pcWhat ? pcWhat : ""), _csLocation(csLocation), _eCode(eCode)
{}

CException::CException(const CException &csException)
: exception(*this), _csWhat(csException._csWhat), _csLocation(csException._csLocation), _eCode(csException._eCode)
{}

CException::~CException() throw()
{
}

CException &CException::operator=(const CException &csException)
{
  _eCode      = csException._eCode;
  _csWhat     = csException._csWhat;
  _csLocation = csException._csLocation;
  return(*this);
}


//
// Generic exception class for fatal errors
//
CFatalError::CFatalError(EERRCode eCode, const char *pcWhat, const CSourceInfo &csLocation)
: CException(eCode, pcWhat, csLocation)
{}

CFatalError::CFatalError(const CFatalError &csException)
: CException(csException)
{}

CFatalError::~CFatalError() throw()
{
}


//
// Exception class for logic errors
//
CLogicError::CLogicError(EERRCode eCode, const char *pcWhat, const CSourceInfo &csLocation)
: CException(eCode, pcWhat, csLocation)
{}

CLogicError::CLogicError(const CLogicError &csException)
: CException(csException)
{}

CLogicError::~CLogicError() throw()
{
}


//
// Exception class for runtime errors
//
CRuntimeError::CRuntimeError(EERRCode eCode, const char *pcWhat, const CSourceInfo &csLocation)
: CException(eCode, pcWhat, csLocation)
{}

CRuntimeError::CRuntimeError(const CRuntimeError &csException)
: CException(csException)
{}

CRuntimeError::~CRuntimeError() throw()
{
}


//
// Exception class for XWindow errors
//
CXWindowError::CXWindowError(EERRCode eCode, const char *pcWhat, const CSourceInfo &csLocation)
: CRuntimeError(eCode, pcWhat, csLocation)
{}

CXWindowError::CXWindowError(const CXWindowError &csException)
: CRuntimeError(csException)
{}

CXWindowError::~CXWindowError() throw()
{
}


//
// Exception class for system errors
//
CSystemError::CSystemError(EERRCode eCode, const char *pcWhat, const CSourceInfo &csLocation)
: CRuntimeError(eCode, pcWhat, csLocation)
{}

CSystemError::CSystemError(const CSystemError &csException)
: CRuntimeError(csException)
{}

CSystemError::~CSystemError() throw()
{
}


//
// Exception class for IO errors
//
CIOError::CIOError(EERRCode eCode, const char *pcWhat, const CSourceInfo &csLocation)
: CRuntimeError(eCode, pcWhat, csLocation)
{}

CIOError::CIOError(const CIOError &csException)
: CRuntimeError(csException)
{}

CIOError::~CIOError() throw()
{
}


//
// Exception class for scripting (VortexLua) errors
//
CScriptingError::CScriptingError(EERRCode eCode, const char *pcWhat, const CSourceInfo &csLocation)
: CRuntimeError(eCode, pcWhat, csLocation)
{}

CScriptingError::CScriptingError(const CScriptingError &csException)
: CRuntimeError(csException)
{}

CScriptingError::~CScriptingError() throw()
{
}


//
// Exported functions
//
void VGE_EXPORT VThrowExceptionBasedOnCode(EERRCode code, const char *msg, const CSourceInfo &loc)
{
  if((code >= _S_FATAL_ERRORS_) && (code <= _E_FATAL_ERRORS_)) {
    throw CFatalError(code, msg, loc);
  }
  else if((code >= _S_LOGIC_ERRORS_) && (code <= _E_LOGIC_ERRORS_)) {
    throw CLogicError(code, msg, loc);
  }
  else if((code >= _S_RTIME_ERRORS_) && (code <= _E_RTIME_ERRORS_)) {
    if((code >= _S_XWIND_ERRORS_) && (code <= _E_XWIND_ERRORS_))
      throw CXWindowError(code, msg, loc);
    else if((code >= _S_SYSTS_ERRORS_) && (code <= _E_SYSTS_ERRORS_))
      throw CSystemError(code, msg, loc);
    else if((code >= _S_INOUT_ERRORS_) && (code <= _E_INOUT_ERRORS_))
      throw CIOError(code, msg, loc);
    else if((code >= _S_VXLUA_ERRORS_) && (code <= _E_VXLUA_ERRORS_))
      throw CScriptingError(code, msg, loc);
    else
      throw CRuntimeError(code, msg, loc);
  }
  else {
    throw CException(code, msg, loc);
  }
}

} // namespace vge
