// ############################################################################################
//  LookUpTables.cc
//  Implementation of various LUTs
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  The source codes in this file is partialy adopted from the book :
//    Cutting-Edge 3D Game Programming with C++
//    John De Goes
//    Coriolis Group, Inc.
//    USA, 1996
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "LookUpTables.h"


namespace vge {

// Trigonometry LUTs
float xprv::_afSinTable_[_DEGREECOUNT_] = {0};
float xprv::_afCosTable_[_DEGREECOUNT_] = {0};

// Lighting LUTs
uint32_t xprv::_aun32RLightTable_[32 * 32] = {0};
uint32_t xprv::_aun32GLightTable_[32 * 32] = {0};
uint32_t xprv::_aun32BLightTable_[32 * 32] = {0};

// Fogged-lighting LUTs
uint32_t *xprv::_pun32RFogLightTable_ = NULL;
uint32_t *xprv::_pun32GFogLightTable_ = NULL;
uint32_t *xprv::_pun32BFogLightTable_ = NULL;

// Fogged-color LUTs
uint32_t *xprv::_pun32RFogColorTable_ = NULL;
uint32_t *xprv::_pun32GFogColorTable_ = NULL;
uint32_t *xprv::_pun32BFogColorTable_ = NULL;

// Addition LUT
uint8_t xprv::_aun8CAddTable_[32 * 32] = {0};

// Transparency LUTs
uint8_t xprv::_aun8CMeanTable_[32 * 32]       = {0};
uint8_t xprv::_aun8TLightTable_[32 * 32 * 32] = {0};
uint8_t *xprv::_pun8TRFogColorTable_          = NULL;
uint8_t *xprv::_pun8TGFogColorTable_          = NULL;
uint8_t *xprv::_pun8TBFogColorTable_          = NULL;
uint8_t *xprv::_pun8TRFogLightTable_          = NULL;
uint8_t *xprv::_pun8TGFogLightTable_          = NULL;
uint8_t *xprv::_pun8TBFogLightTable_          = NULL;


//
// Exported private functions implementation
//
namespace xprv {

  void _VInitSinCosTable_()
  {
    // Static variables
    static bool bInitialized = false;

    // Check
    if(bInitialized) return;
  
    // Generate the table
    float fUnit = float(3.141592654) * 2.0f / float(xprv::_DEGREECOUNT_);
    for(size_t i = 0; i < xprv::_DEGREECOUNT_; ++i) {
      xprv::_afSinTable_[i] = sinf(fUnit * float(int(i)));
      xprv::_afCosTable_[i] = cosf(fUnit * float(int(i)));
    }
    bInitialized = true;
  }

  void _VInitLightingTable_()
  {
    // Static variables
    static bool bInitialized = false;
  
    // Check
    if(bInitialized) return;
  
    // Generate the lighting LUTs
    for(uint32_t un32C1 = 0; un32C1 < 32; un32C1++)
      for(uint32_t un32C2 = 0; un32C2 < 32; un32C2++) {
        uint32_t un32T = (un32C1 * un32C2) / 31;
        xprv::_aun32RLightTable_[(un32C1 << 5) | un32C2] = un32T << 19;
        xprv::_aun32GLightTable_[(un32C1 << 5) | un32C2] = un32T << 11;
        xprv::_aun32BLightTable_[(un32C1 << 5) | un32C2] = un32T << 3;
      }

    // Generate the mean and addition LUTs
    for(uint32_t un32C1 = 0; un32C1 < 32; un32C1++)
      for(uint32_t un32C2 = 0; un32C2 < 32; un32C2++) {
        // Mean LUT
        xprv::_aun8CMeanTable_[(un32C1 << 5) | un32C2] = uint8_t((un32C1 + un32C2) / 2) << 3;
        // Addition LUT
        xprv::_aun8CAddTable_[(un32C1 << 5) | un32C2] = min<uint8_t>(31, uint8_t(un32C1 + un32C2));
        xprv::_aun8CAddTable_[(un32C1 << 5) | un32C2] <<= 3;
      }
  
    // Generate the lighting LUTs with transparency
    for(uint32_t un32C1 = 0; un32C1 < 32; un32C1++)
      for(uint32_t un32C2 = 0; un32C2 < 32; un32C2++) {
        uint32_t un32T = (un32C1 * un32C2) / 31;
        for(uint32_t un32OC = 0; un32OC < 32; un32OC++)
          xprv::_aun8TLightTable_[(un32OC << 10) | (un32C1 << 5) | un32C2] = uint8_t((un32T + un32OC) / 2) << 3;
      }

    // OK !
    bInitialized = true;
  }

} // namespace xprv

} // namespace vge
