// ############################################################################################
//  MessageDisplayer.cc
//  Implementation of console-message-displayer functions
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "MessageDisplayer.h"


namespace vge {

// Print buffer size
#define _PBUF_SIZE_ 8192

// ANSI escape commands
#define _ESC_CLEAR_SCREEN   "\033[2J\033[0;0H"
#define _ESC_COLOR_RESTORE  "\033[0m"
#define _ESC_COLOR_FATAL    "\033[1;37;41m"
#define _ESC_COLOR_ERROR    "\033[1;31;40m"
#define _ESC_COLOR_WARNING  "\033[1;33;40m"
#define _ESC_COLOR_INFO     "\033[1;34;40m"
#define _ESC_COLOR_LOCATION "\033[1;32;40m"
#define _ESC_COLOR_LINE     "\033[1;35;40m"
#define _ESC_COLOR_DETAIL   "\033[1;37;40m"


namespace xprv {

  // Reference to external function from "lib/Base/StartupFunctions.cc"
  void VGE_EXPORT _vgeStartupOTIF_() __attribute__((constructor));

  // since the object file created from this MessageDisplayer.cc will be always linked,
  // we declare a dummy function to force so that the object file which contains the
  // _vgeStartupOTIF_() function will be always linked (
  void _dummy_() { xprv::_vgeStartupOTIF_(); }

} // namespace xprv


// Message displayer target
static FILE *_ptInfo_    = stdout;
static FILE *_ptWarning_ = stdout;
static FILE *_ptError_   = stdout;
static FILE *_ptDebug_   = stdout;
static bool  _enDebug_   = false;
static bool  _feAbort_   = false;


//
// Message displayer functions
//
void VClrScr()
{
#ifdef __USE_COLORED_MESSAGES__
  fprintf(stdout, "%s", _ESC_CLEAR_SCREEN); fflush(stdout);
  fprintf(stderr, "%s", _ESC_CLEAR_SCREEN); fflush(stderr);
#else
  system("clear");
#endif
}

void VRedirectMsgs(FILE *ptInfo, FILE *ptWarning, FILE *ptError, FILE *ptDebug)
{
  _ptInfo_    = ptInfo;
  _ptWarning_ = ptWarning;
  _ptError_   = ptError;
  _ptDebug_   = ptDebug;
}

void VEnableDebugPrint(bool enable)
{
  _enDebug_ = enable;
}

void VAbortOnFatalError(bool abortAfterPrint)
{
  _feAbort_ = abortAfterPrint;
}

void VFatalError(EERRCode eCode, const char *pcMsg, ...)
{
  // Extract the pcMsg
  va_list vlTmp;
  char    _acPBuff_[_PBUF_SIZE_];
  va_start(vlTmp, pcMsg);
    vsnprintf(_acPBuff_, _PBUF_SIZE_, pcMsg, vlTmp);
  va_end(vlTmp);

  // Print the message
#ifdef __USE_COLORED_MESSAGES__
  fprintf(_ptError_, "%s\n%sFATAL ERROR:%s\n------------------------------------------------------------\n%s%s%s\n",
          _ESC_COLOR_RESTORE, _ESC_COLOR_FATAL, _ESC_COLOR_LINE, _ESC_COLOR_DETAIL,
          _acPBuff_, _ESC_COLOR_RESTORE);
#else
  fprintf(_ptError_, "\nFATAL ERROR:\n------------------------------------------------------------\n%s\n",
          _acPBuff_);
#endif

  // Abort or throw the exception
  if(_feAbort_) abort();
  throw CFatalError(eCode, _acPBuff_, CSourceInfo("", "", -1));
}

void VFatalError(EERRCode eCode, const char *pcFunctionName, const char *pcFile, int32_t n32Line, const char *pcMsg, ...)
{
  // Extract the pcMsg
  va_list vlTmp;
  char    _acPBuff_[_PBUF_SIZE_];
  va_start(vlTmp, pcMsg);
    vsnprintf(_acPBuff_, _PBUF_SIZE_, pcMsg, vlTmp);
  va_end(vlTmp);

  // Print the message
#ifdef __USE_COLORED_MESSAGES__
  fprintf(_ptError_, "%s\n%sFATAL ERROR:%s %s, line %d (%s)%s\n------------------------------------------------------------\n%s%s%s\n",
          _ESC_COLOR_RESTORE, _ESC_COLOR_FATAL, _ESC_COLOR_LOCATION,
          pcFile, n32Line, pcFunctionName,
          _ESC_COLOR_LINE, _ESC_COLOR_DETAIL, _acPBuff_, _ESC_COLOR_RESTORE);
#else
  fprintf(_ptError_, "\nFATAL ERROR: %s, line %d (%s)\n------------------------------------------------------------\n%s\n",
          pcFile, n32Line, pcFunctionName, _acPBuff_);
#endif

  // Abort or throw the exception
  if(_feAbort_) abort();
  throw CFatalError(eCode, _acPBuff_, CSourceInfo(pcFunctionName, pcFile, n32Line));
}

// Display error message
void VError(EERRCode eCode, const char *pcMsg, ...)
{
  // Extract the pcMsg
  va_list vlTmp;
  char    _acPBuff_[_PBUF_SIZE_];
  va_start(vlTmp, pcMsg);
    vsnprintf(_acPBuff_, _PBUF_SIZE_, pcMsg, vlTmp);
  va_end(vlTmp);

  // Print the message
#ifdef __USE_COLORED_MESSAGES__
  fprintf(_ptError_, "%s\n%sERROR:%s\n------------------------------------------------------------\n%s%s%s\n",
          _ESC_COLOR_RESTORE, _ESC_COLOR_ERROR, _ESC_COLOR_LINE, _ESC_COLOR_DETAIL,
          _acPBuff_, _ESC_COLOR_RESTORE);
#else
  fprintf(_ptError_, "\nERROR:\n------------------------------------------------------------\n%s\n",
          _acPBuff_);
#endif

  // Throw the exception
  VThrowExceptionBasedOnCode(eCode, _acPBuff_, CSourceInfo("", "", -1));
}

void VError(EERRCode eCode, const char *pcFunctionName, const char *pcFile, int32_t n32Line, const char *pcMsg, ...)
{
  // Extract the pcMsg
  va_list vlTmp;
  char    _acPBuff_[_PBUF_SIZE_];
  va_start(vlTmp, pcMsg);
    vsnprintf(_acPBuff_, _PBUF_SIZE_, pcMsg, vlTmp);
  va_end(vlTmp);

  // Print the message
#ifdef __USE_COLORED_MESSAGES__
  fprintf(_ptError_, "%s\n%sERROR:%s %s, line %d (%s)%s\n------------------------------------------------------------\n%s%s%s\n",
          _ESC_COLOR_RESTORE, _ESC_COLOR_ERROR, _ESC_COLOR_LOCATION,
          pcFile, n32Line, pcFunctionName,
          _ESC_COLOR_LINE, _ESC_COLOR_DETAIL, _acPBuff_, _ESC_COLOR_RESTORE);
#else
  fprintf(_ptError_, "\nERROR: %s, line %d (%s)\n------------------------------------------------------------\n%s\n",
          pcFile, n32Line, pcFunctionName, _acPBuff_);
#endif

  // Throw the exception
  VThrowExceptionBasedOnCode(eCode, _acPBuff_, CSourceInfo(pcFunctionName, pcFile, n32Line));
}

void VWarning(const char *pcMsg, ...)
{
  // Extract the pcMsg
  va_list vlTmp;
  char    _acPBuff_[_PBUF_SIZE_];
  va_start(vlTmp, pcMsg);
    vsnprintf(_acPBuff_, _PBUF_SIZE_, pcMsg, vlTmp);
  va_end(vlTmp);

  // Print the message
#ifdef __USE_COLORED_MESSAGES__
  fprintf(_ptWarning_, "%s\n%sWARNING:%s\n------------------------------------------------------------\n%s%s%s\n",
          _ESC_COLOR_RESTORE, _ESC_COLOR_WARNING, _ESC_COLOR_LINE, _ESC_COLOR_DETAIL,
          _acPBuff_, _ESC_COLOR_RESTORE);
#else
  fprintf(_ptWarning_, "\nWARNING:\n------------------------------------------------------------\n%s\n",
          _acPBuff_);
#endif
}

void VWarning(const char *pcFunctionName, const char *pcFile, int32_t n32Line, const char *pcMsg, ...)
{
  // Extract the pcMsg
  va_list vlTmp;
  char    _acPBuff_[_PBUF_SIZE_];
  va_start(vlTmp, pcMsg);
    vsnprintf(_acPBuff_, _PBUF_SIZE_, pcMsg, vlTmp);
  va_end(vlTmp);

  // Print the message
#ifdef __USE_COLORED_MESSAGES__
  fprintf(_ptWarning_, "%s\n%sWARNING:%s %s, line %d (%s)%s\n------------------------------------------------------------\n%s%s%s\n",
          _ESC_COLOR_RESTORE, _ESC_COLOR_WARNING, _ESC_COLOR_LOCATION,
          pcFile, n32Line, pcFunctionName,
          _ESC_COLOR_LINE, _ESC_COLOR_DETAIL, _acPBuff_, _ESC_COLOR_RESTORE);
#else
  fprintf(_ptWarning_, "\nWARNING: %s, line %d (%s)\n------------------------------------------------------------\n%s\n",
          pcFile, n32Line, pcFunctionName, _acPBuff_);
#endif
}

void VInfo(const char *pcMsg, ...)
{
  // Extract the pcMsg
  va_list vlTmp;
  char    _acPBuff_[_PBUF_SIZE_];
  va_start(vlTmp, pcMsg);
    vsnprintf(_acPBuff_, _PBUF_SIZE_, pcMsg, vlTmp);
  va_end(vlTmp);

  // Print the message
#ifdef __USE_COLORED_MESSAGES__
  fprintf(_ptInfo_, "%s\n%sINFO:%s\n------------------------------------------------------------\n%s%s%s\n",
          _ESC_COLOR_RESTORE, _ESC_COLOR_INFO, _ESC_COLOR_LINE, _ESC_COLOR_DETAIL,
          _acPBuff_, _ESC_COLOR_RESTORE);
#else
  fprintf(_ptInfo_, "\nINFO:\n------------------------------------------------------------\n%s\n",
          _acPBuff_);
#endif
}

void VInfo(const char *pcFunctionName, const char *pcFile, int32_t n32Line, const char *pcMsg, ...)
{
  // Extract the pcMsg
  va_list vlTmp;
  char    _acPBuff_[_PBUF_SIZE_];
  va_start(vlTmp, pcMsg);
    vsnprintf(_acPBuff_, _PBUF_SIZE_, pcMsg, vlTmp);
  va_end(vlTmp);

  // Print the message
#ifdef __USE_COLORED_MESSAGES__
  fprintf(_ptInfo_, "%s\n%sINFO:%s %s, line %d (%s)%s\n------------------------------------------------------------\n%s%s%s\n",
          _ESC_COLOR_RESTORE, _ESC_COLOR_INFO, _ESC_COLOR_LOCATION,
          pcFile, n32Line, pcFunctionName,
          _ESC_COLOR_LINE, _ESC_COLOR_DETAIL, _acPBuff_, _ESC_COLOR_RESTORE);
#else
  fprintf(_ptInfo_, "\nINFO: %s, line %d (%s)\n------------------------------------------------------------\n%s\n",
          pcFile, n32Line, pcFunctionName, _acPBuff_);
#endif
}

void VPrint(const char *pcMsg, ...)
{
  // Extract the pcMsg
  va_list vlTmp;
  char    _acPBuff_[_PBUF_SIZE_];
  va_start(vlTmp, pcMsg);
    vsnprintf(_acPBuff_, _PBUF_SIZE_, pcMsg, vlTmp);
  va_end(vlTmp);

  // Print the message
#ifdef __USE_COLORED_MESSAGES__
  fprintf(_ptInfo_, "%s%s%s", _ESC_COLOR_DETAIL, _acPBuff_, _ESC_COLOR_RESTORE);
#else
  fprintf(_ptInfo_, "%s", _acPBuff_);
#endif
}

void VDebugPrint(const char *pcMsg, ...)
{
  if(!_enDebug_) return;

  // Extract the pcMsg
  va_list vlTmp;
  char    _acPBuff_[_PBUF_SIZE_];
  va_start(vlTmp, pcMsg);
    vsnprintf(_acPBuff_, _PBUF_SIZE_, pcMsg, vlTmp);
  va_end(vlTmp);

  // Print the message
#ifdef __USE_COLORED_MESSAGES__
  fprintf(_ptDebug_, "%s%s%s", _ESC_COLOR_DETAIL, _acPBuff_, _ESC_COLOR_RESTORE);
#else
  fprintf(_ptDebug_, "%s", _acPBuff_);
#endif
}

void VAssertPrint(const char *pcFunctionName, const char *pcFile, int32_t n32Line, const char *pcAssertionText, const char *pcMsg, ...)
{
  // Extract the pcMsg
  char _acPBuff_[_PBUF_SIZE_];
  if(pcMsg) {
    va_list vlTmp;
    va_start(vlTmp, pcMsg);
      vsnprintf(_acPBuff_, _PBUF_SIZE_, pcMsg, vlTmp);
    va_end(vlTmp);
  }

  // Print the message
#ifdef __USE_COLORED_MESSAGES__
  if(pcMsg)
    fprintf(_ptDebug_, "%s\n%s%s, line %d (%s): %sassertion '%s%s%s' failed.\n%s%s%s\n",
           _ESC_COLOR_RESTORE, _ESC_COLOR_LOCATION, pcFile, n32Line, pcFunctionName,
           _ESC_COLOR_ERROR, _ESC_COLOR_INFO, pcAssertionText, _ESC_COLOR_ERROR,
           _ESC_COLOR_DETAIL, _acPBuff_, _ESC_COLOR_RESTORE);
  else
    fprintf(_ptDebug_, "%s\n%s%s, line %d (%s): %sassertion '%s%s%s' failed.%s\n",
           _ESC_COLOR_RESTORE, _ESC_COLOR_LOCATION, pcFile, n32Line, pcFunctionName,
           _ESC_COLOR_ERROR, _ESC_COLOR_INFO, pcAssertionText, _ESC_COLOR_ERROR,
           _ESC_COLOR_RESTORE);
#else
  if(pcMsg)
    fprintf(_ptDebug_, "\n%s, line %d (%s): assertion '%s' failed.\n%s\n",
           pcFile, n32Line, pcFunctionName, pcAssertionText, _acPBuff_);
  else
    fprintf(_ptDebug_, "\n%s, line %d (%s): assertion '%s' failed.\n",
           pcFile, n32Line, pcFunctionName, pcAssertionText);
#endif

  // Abort the application
  abort();
}

} // namespace vge
