/**
 * \file     SIMDTypes.h
 * <hr>
 * \internal This file contains the data types and functions needed for working with
 *           GCC's SIMD and auto-vectoring support.
 * <hr>
 */
// ############################################################################################
//  SIMDTypes.h
//  Definition of SIMD data types & their related functions
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _BASE___SIMD_TYPES_H_
#define _BASE___SIMD_TYPES_H_

// VortexGE headers
#include "VBasics.h"


// For experiment, just leave it as it
#ifdef _ASM_SSE_
#  define __USE_AVEC__
#endif


namespace vge {

  //
  // Vector of 4 packed int32_t values
  //
  typedef int32_t v4si __attr_vector_size(16);

  //
  // Union for easier v4si vector's elements access
  //
  union v4si_u {
    v4si    v;
    int32_t i[4];
  
    inline _attr_force_inline v4si_u()
    : v(__extension__(v4si){0, 0, 0, 0}) {}

    inline _attr_force_inline v4si_u(const v4si &v_)
    : v(v_) {}
  };

  //
  // v4si related functions
  //

  //
  // Generate a v4si vector of which all elements are zeroed.
  static inline v4si _attr_force_inline ZERO_V4SI()
  { return __extension__(v4si){0, 0, 0, 0}; }
  //
  // Generate a v4si vector of which all elements are set with the given value.
  static inline v4si _attr_force_inline MAKE_V4SI(int32_t ii)
  { return __extension__(v4si){ii, ii, ii, ii}; }
  //
  // Generate a v4si vector of which the elements are set with the given values.
  static inline v4si _attr_force_inline MAKE_V4SI(int32_t i3, int32_t i2, int32_t i1, int32_t i0)
  { return __extension__(v4si){i0, i1, i2, i3}; }
  //
  // Perform raw (bit by bit) conversion from 4 consecutive int32_t values
  // to v4si. Use with caution because the order of the generated vector's
  // elements are determined by the declaration order of the variables given
  // using the "startRef" parameter.
  static inline v4si& _attr_force_inline I32_TO_V4SI(int32_t &startRef)
  { return reinterpret_cast<v4si&>(startRef); }
  //
  // Perform raw (bit by bit) conversion from 4 consecutive int32_t values
  // to v4si. Use with caution because the order of the generated vector's
  // elements are determined by the declaration order of the variables given
  // using the "startRef" parameter.
  static inline const v4si& _attr_force_inline I32_TO_V4SI(const int32_t &startRef)
  { return reinterpret_cast<const v4si&>(startRef); }
  //
  // Perform raw (bit by bit) conversion from a v4si vector to 4 consecutive
  // int32_t values. Should be used in conjugation with I32_TO_V4SI() only.
  static inline int32_t& _attr_force_inline V4SI_TO_I32(v4si &ref)
  { return reinterpret_cast<int32_t&>(ref); }
  //
  // Perform raw (bit by bit) conversion from a v4si vector to 4 consecutive
  // int32_t values. Should be used in conjugation with I32_TO_V4SI() only.
  static inline const int32_t& _attr_force_inline V4SI_TO_I32(const v4si &ref)
  { return reinterpret_cast<const int32_t&>(ref); }
  //
  // Perform raw (bit by bit) conversion from 4 consecutive int32_t values
  // to v4si. Use with caution because the order of the generated vector's
  // elements are determined by the declaration order of the variables given
  // using the "startPtr" parameter.
  static inline v4si* _attr_force_inline I32_TO_V4SI(int32_t *startPtr)
  { return reinterpret_cast<v4si*>(startPtr); }
  //
  // Perform raw (bit by bit) conversion from 4 consecutive int32_t values
  // to v4si. Use with caution because the order of the generated vector's
  // elements are determined by the declaration order of the variables given
  // using the "startPtr" parameter.
  static inline const v4si* _attr_force_inline I32_TO_V4SI(const int32_t *startPtr)
  { return reinterpret_cast<const v4si*>(startPtr); }
  //
  // Perform raw (bit by bit) conversion from a v4si vector to 4 consecutive
  // int32_t values. Should be used in conjugation with I32_TO_V4SI() only.
  static inline int32_t* _attr_force_inline V4SI_TO_I32(v4si *ptr)
  { return reinterpret_cast<int32_t*>(ptr); }
  //
  // Perform raw (bit by bit) conversion from a v4si vector to 4 consecutive
  // int32_t values. Should be used in conjugation with I32_TO_V4SI() only.
  static inline const int32_t* _attr_force_inline V4SI_TO_I32(const v4si *ptr)
  { return reinterpret_cast<const int32_t*>(ptr); }


  //
  // Vector of 4 packed float values
  //
  typedef float v4sf __attr_vector_size(16);

  //
  // Union for easier v4sf vector's elements access
  //
  union v4sf_u {
    v4sf  v;
    float f[4];
  
    inline _attr_force_inline v4sf_u()
    : v(__extension__(v4sf){0.0f, 0.0f, 0.0f, 0.0f}) {}

    inline _attr_force_inline v4sf_u(const v4sf &v_)
    : v(v_) {}
  };

  //
  // v4sf related functions
  //

  //
  // Generate a v4sf vector of which all elements are zeroed.
  static inline v4sf _attr_force_inline ZERO_V4SF()
  { return __extension__(v4sf){0.0f, 0.0f, 0.0f, 0.0f}; }
  //
  // Generate a v4sf vector of which all elements are set with the given value.
  static inline v4sf _attr_force_inline MAKE_V4SF(float ff)
  { return __extension__(v4sf){ff, ff, ff, ff}; }
  //
  // Generate a v4sf vector of which the elements are set with the given values.
  static inline v4sf _attr_force_inline MAKE_V4SF(float f3, float f2, float f1, float f0)
  { return __extension__(v4sf){f0, f1, f2, f3}; }
  //
  // Perform raw (bit by bit) conversion from 4 consecutive float values
  // to v4sf. Use with caution because the order of the generated vector's
  // elements are determined by the declaration order of the variables given
  // using the "startRef" parameter.
  static inline v4sf& _attr_force_inline FLT_TO_V4SF(float &startRef)
  { return reinterpret_cast<v4sf&>(startRef); }
  //
  // Perform raw (bit by bit) conversion from 4 consecutive float values
  // to v4sf. Use with caution because the order of the generated vector's
  // elements are determined by the declaration order of the variables given
  // using the "startRef" parameter.
  static inline const v4sf& _attr_force_inline FLT_TO_V4SF(const float &startRef)
  { return reinterpret_cast<const v4sf&>(startRef); }
  //
  // Perform raw (bit by bit) conversion from a v4sf vector to 4 consecutive
  // float values. Should be used in conjugation with FLT_TO_V4SF() only.
  static inline float& _attr_force_inline V4SF_TO_FLT(v4sf &ref)
  { return reinterpret_cast<float&>(ref); }
  //
  // Perform raw (bit by bit) conversion from a v4sf vector to 4 consecutive
  // float values. Should be used in conjugation with FLT_TO_V4SF() only.
  static inline const float& _attr_force_inline V4SF_TO_FLT(const v4sf &ref)
  { return reinterpret_cast<const float&>(ref); }
  //
  // Perform raw (bit by bit) conversion from 4 consecutive float values
  // to v4sf. Use with caution because the order of the generated vector's
  // elements are determined by the declaration order of the variables given
  // using the "startPtr" parameter.
  static inline v4sf* _attr_force_inline FLT_TO_V4SF(float *startPtr)
  { return reinterpret_cast<v4sf*>(startPtr); }
  //
  // Perform raw (bit by bit) conversion from 4 consecutive float values
  // to v4sf. Use with caution because the order of the generated vector's
  // elements are determined by the declaration order of the variables given
  // using the "startPtr" parameter.
  static inline const v4sf* _attr_force_inline FLT_TO_V4SF(const float *startPtr)
  { return reinterpret_cast<const v4sf*>(startPtr); }
  //
  // Perform raw (bit by bit) conversion from a v4sf vector to 4 consecutive
  // float values. Should be used in conjugation with FLT_TO_V4SF() only.
  static inline float* _attr_force_inline V4SF_TO_FLT(v4sf *ptr)
  { return reinterpret_cast<float*>(ptr); }
  //
  // Perform raw (bit by bit) conversion from a v4sf vector to 4 consecutive
  // float values. Should be used in conjugation with FLT_TO_V4SF() only.
  static inline const float* _attr_force_inline V4SF_TO_FLT(const v4sf *ptr)
  { return reinterpret_cast<const float*>(ptr); }

} // namespace vge

#endif
