/** \file InterProcessCommunication.h */
// ############################################################################################
//  InterProcessCommunication.h
//  Definition and implementation of IPC classes
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _SYSTEM___INTER_PROCESS_COMMUNICATION_H_
#define _SYSTEM___INTER_PROCESS_COMMUNICATION_H_

// VortexGE headers
#include "../Base/SpecialClasses.h"


namespace vge {

  // Maximum values allowed
  //
  // If a server and it's client applications are compiled using different
  // values of these constants, one will most likely to get segmentation
  // fault sooner or later
  const uint32_t _MAX_MQ_CLIENT_ = uint32_t(32);
  const uint32_t _MAX_MQ_STRBUF_ = uint32_t(32);

  // Special constants for enumeration request
  const char _ENMBYID_ALL_[] = "[ ***** ALL ***** ]";

  // Helper functions
  extern VGE_EXPORT bool VEnableGetChar(bool bState);
  extern VGE_EXPORT char VGetChar();

  //!
  //! \internal
  //! \brief Structure for storing client list
  //!
  //! \api experiment
  struct _TClientData {
    uint32_t un32ClientID;           // The effective client ID
    char     acUID[_MAX_MQ_STRBUF_]; // The user ID set by client
  };

  //!
  //! \internal
  //! \brief Structure sent by client for connection request
  //!
  //! \api experiment
  struct _TCReqConnData {
    msgbuf   tMsgBuf;                // The msgbuf structure
    char     acSID[_MAX_MQ_STRBUF_]; // Transaction string ID, must be _VIPC_CSID_CONNECT_
    char     acUID[_MAX_MQ_STRBUF_]; // Any user ID set by client
    uint32_t un32RandomID;           // Any random value set by client
  };

  //!
  //! \internal
  //! \brief Structure sent by server in response of connection request
  //!
  //! \api experiment
  struct _TSRspConnData {
    msgbuf   tMsgBuf;                // The msgbuf structure
    char     acSID[_MAX_MQ_STRBUF_]; // Transaction string ID, must be _VIPC_SSID_CONNECT_
    char     acUID[_MAX_MQ_STRBUF_]; // Must match the random value previously sent by client
    uint32_t un32RandomID;           // Must match the random value previously sent by client
    uint32_t un32ClientID;           // Effective client ID returned, zero means connection refused
  };

  //!
  //! \internal
  //! \brief Structure sent by client for disconnection request
  //!
  //! \api experiment
  struct _TCReqDscnData {
    msgbuf   tMsgBuf;                // The msgbuf structure
    char     acSID[_MAX_MQ_STRBUF_]; // Transaction string ID, must be _VIPC_CSID_DISCONN_
    char     acUID[_MAX_MQ_STRBUF_]; // The user client ID
    uint32_t un32ClientID;           // The effective client ID
  };

  //!
  //! \internal
  //! \brief Structure sent by server for shut down notification
  //!
  //! \api experiment
  struct _TSNtcSDwnData {
    msgbuf   tMsgBuf;                // The msgbuf structure
    char     acSID[_MAX_MQ_STRBUF_]; // Transaction string ID, must be _VIPC_SSID_SHUTDWN_
    uint32_t un32ClientID;           // Effective client ID
  };

  //!
  //! \internal
  //! \brief Structure sent by client for enumeration request
  //!
  //! \api experiment
  struct _TCReqEByUData {
    msgbuf   tMsgBuf;                     // The msgbuf structure
    char     acSID[_MAX_MQ_STRBUF_];      // Transaction string ID, must be _VIPC_CSID_ENMBYID_
    char     acUID[_MAX_MQ_STRBUF_];      // The user client ID
    uint32_t un32ClientID;                // The effective client ID
    char     acUID2Enum[_MAX_MQ_STRBUF_]; // UID to be enumerated
  };

  //!
  //! \internal
  //! \brief Structure sent by server in response of enumeration request
  //!
  //! \api experiment
  struct _TSRspEByUData {
    msgbuf   tMsgBuf;                        // The msgbuf structure
    char     acSID[_MAX_MQ_STRBUF_];         // Transaction string ID, must be _VIPC_SSID_ENMBYID_
    uint32_t un32ClientID;                   // The effective client ID which ask for enumeration
    uint32_t un32EnumCount;                  // Total enumerated client
    uint32_t aun32ClientID[_MAX_MQ_CLIENT_]; // List of client ID of which UID is match
  };

  //!
  //! \internal
  //! \brief Structure sent by client for custom client to client transaction
  //!
  //! \api experiment
  struct _TCTrnCC2CData {
    msgbuf   tMsgBuf;                // The msgbuf structure
    char     acSID[_MAX_MQ_STRBUF_]; // Transaction string ID, must be _VIPC_CSID_CC2CTRN_
    uint32_t un32ClientID;           // The effective client ID of the sender
    uint32_t un32CustomDataSize;     // Data size following this structure (in bytes)
  };


  //!
  //! \brief Message queue server (manager) class
  //!
  //! \api experiment
  class VGE_EXPORT CMQServer : VCLASS_UNCOPYABLE {
    public:
      //!
      //! Construct a new MQ server (manager) instance and 
      //! initialize it if the "pcConnName" parameter is not NULL.
      CMQServer(const char *pcConnName = NULL);
      //!
      //! Standard dtor.
      ~CMQServer() throw();

      // Initialization functions
      //!
      //! Initialize the MQ server (manager) instance.
      void Initialize(const char *pcConnName);
      //!
      //! Shutdown the MQ server (manager) instance, unles
      //! bForce is set to true, shutdown will be failed if
      //! there is still active client.
      void ShutDown(bool bForce = false);

      // Execution functions
      //!
      //! Must be called from the server (manager) application
      //! to make the MQ server (manager) handle client requests.
      void Pool();
      //!
      //! Enumerate all registered clients.
      //! \param[in] aun32ClientID   Destination array that will be filled with the enumerated client IDs.
      //! \param[in] pcUID2Enumerate Optional UID, the default is to enumerate client will any UID.
      //! \param[in] un32CIDExcept   Optional CID, the default is to not ignore any CID.
      uint32_t EnumerateClientByUID(uint32_t aun32ClientID[_MAX_MQ_CLIENT_], const char *pcUID2Enumerate = _ENMBYID_ALL_, uint32_t un32CIDExcept = 0) const;

    protected:
      // Execution functions
      uint32_t _EnumerateClientByUID_(uint32_t aun32ClientID[_MAX_MQ_CLIENT_], const char *pcUID2Enumerate, uint32_t un32CIDExcept) const;

      // Data
      char         _acFileName[PATH_MAX];     // The temporary file used by ftok()
      key_t        _tKey;                     // A unique key to be used by message queue
      int32_t      _n32QID;                   // The message queue ID
      _TClientData _atCList[_MAX_MQ_CLIENT_]; // The client list
  };


  //!
  //! \brief Message queue client class
  //!
  //! \api experiment
  class VGE_EXPORT CMQClient : VCLASS_UNCOPYABLE {
    public:
      CMQClient();
      ~CMQClient() throw();

      // Initialization functions
      //!
      //! Connect to the given connection name (MQ server name).
      //! \param[in] pcConnName       Connection (MQ server) name.
      //! \param[in] pcUserID         Optional UID to be used.
      //! \param[in] un8TimeOutSecond Optional timeout specification (in seconds) for operation.
      void Connect(const char *pcConnName, const char *pcUserID = NULL, uint8_t un8TimeOutInSeconds = 3);
      //!
      //! Disconnect from server.
      void Disconnect();

      // Information functions
      //!
      //! Returns the CID of this client.
      inline uint32_t GetCID() const { return(_un32CID); }
      //!
      //! Check if this client is connected to server.
      inline bool Connected() const { return(_un32CID > 0); }
      //!
      //! Check if this client is not connected to server.
      inline bool Disconnected() const { return(_un32CID == 0); }

      // Execution functions
      //!
      //! Check if the server has order the client to shutdown.
      bool GotShutDownNotification();
      //!
      //! Enumerate all registered clients in the server.
      //! \param[in] aun32ClientID    Destination array that will be filled with the enumerated client IDs.
      //! \param[in] pcUID2Enumerate  Optional UID, the default is to enumerate client will any UID.
      //! \param[in] un8TimeOutSecond Optional timeout specification (in seconds) for operation.
      uint32_t EnumerateByUID(uint32_t aun32ClientID[_MAX_MQ_CLIENT_], const char *pcUID2Enumerate = _ENMBYID_ALL_, uint8_t un8TimeOutInSeconds = 3) const;
      //!
      //! Send direct client to client message.
      //! \param[in] un32TargetClientID  Target CID.
      //! \param[in] un8CustomIdentifier Any uint8_t to be used as custom identifier.
      //! \param[in] pvCustomData        Pointer to the buffer containing the data to be sent.
      //! \param[in] un32CustomDataSize  Size of the data to be sent(in bytes).
      void CC2CSend(uint32_t un32TargetClientID, uint8_t un8CustomIdentifier, const void *pvCustomData, uint32_t un32CustomDataSize) const;
      //!
      //! Receive direct client to client message.
      //! \param[in] pun32SourceClientID Source CID.
      //! \param[in] un8CustomIdentifier Any uint8_t to be used as custom identifier.
      //! \param[in] pvCustomData        Pointer to the buffer to be filled with the received data.
      //! \param[in] un32CustomDataSize  Size of the data to be received (in bytes).
      //! \return True if there is client data from the source client (CID) and the
      //!         the data un8CustomIdentifier and un32CustomDataSize match exactly.
      bool CC2CReceive(uint32_t *pun32SourceClientID, uint8_t un8CustomIdentifier, void *pvCustomData, uint32_t un32CustomDataSize) const;

    protected:
      key_t    _tKey;                      // A unique key to be used by message queue
      int32_t  _n32QID;                    // The message queue ID
      uint32_t _un32CID;                   // The effective client ID
      char     _acUserID[_MAX_MQ_STRBUF_]; // The user client ID
  };

} // namespace vge
#endif
