/** \file MemoryOperations.h */
// ############################################################################################
//  MemoryOperations.h
//  Implementation of memory manipulation functions
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _SYSTEM___MEMORY_OPERATIONS_H_
#define _SYSTEM___MEMORY_OPERATIONS_H_

// VortexGE headers
#include "../Base/VBasics.h"


//!
//! \ingroup Arch_Dep_Macros
//! \hideinitializer
//! If VortexGE is built with SSE support, using this macro in a class body
//! (definition), will cause dynamic instantiation of the class (using operator
//! new) to align its allocated memory address so that it is compatible with
//! SSE.
//! \n\n
//! If VortexGE is built without SSE support, dynamic instantiation will use the
//! default address alignment of the system.
//! \n\n
//! <b>Note that this macro will be undefined using \#undef (thus become unusable
//!  by one's applications), unless before including \ref VMain or VMain.h one has
//! already define the macro _KEEP_INTERNAL_MACROS_DEFINED.</b>
#define DECLARE_SSE_ENABLED_NEW_AND_DELETE_OPERATORS()          \
  inline void *operator new(size_t sz) {                        \
    void *m = ssemalloc(sz);                                    \
    if(!m) VError(VERR_NotEnoughMemory, _NFL_,                  \
                  "Couldn't allocate %zd bytes of memory", sz); \
    return(m);                                                  \
  }                                                             \
  inline void *operator new[](size_t sz) {                      \
    void *m = ssemalloc(sz);                                    \
    if(!m) VError(VERR_NotEnoughMemory, _NFL_,                  \
                  "Couldn't allocate %zd bytes of memory", sz); \
    return(m);                                                  \
  }                                                             \
  inline void operator delete(void *pvBuffer)                   \
    { ssefree(pvBuffer); }                                      \
  inline void operator delete[](void *pvBuffer)                 \
    { ssefree(pvBuffer); }                                      \
  inline void *operator new(size_t sz, const nothrow_t&) throw()          \
    { return(ssemalloc(sz)); }                                            \
  inline void *operator new[](size_t sz, const nothrow_t&) throw()        \
    { return(ssemalloc(sz)); }                                            \
  inline void operator delete(void *pvBuffer, const nothrow_t&) throw()   \
    { ssefree(pvBuffer); }                                                \
  inline void operator delete[](void *pvBuffer, const nothrow_t&) throw() \
    { ssefree(pvBuffer); }


namespace vge {

  // Aligned memory allocation functions
  //!
  //! Use this to allocate a chunk of memory if its start address needs
  //! to be specially aligned.
  //! \param[in] szSize      Size of the chunk to be allocated.
  //! \param[in] szAlignment Alignment factor.
  //!
  //! \return The start address of newly allocated chunk or NULL if failed.
  extern VGE_EXPORT void *amalloc(size_t szSize, size_t szAlignment = 16);
  //!
  //! In case of amalloc() is implemented using hack, this function can be used
  //! to obtain the "real" start address from the "fake" one returned by amalloc().
  //! \param[in] pvBuffer Start address of the chunk returned by amalloc().
  //! \return The real start address.
  //!
  //! \sntc Use amalloc() and afree() instead.
  extern VGE_EXPORT void *agetrealaddress(void *pvBuffer);
  //!
  //! Use this to free the memory chunks which were allocatted using amalloc().
  //! \param[in] pvBuffer Start address of the chunk returned by amalloc().
  extern VGE_EXPORT void afree(void *pvBuffer);


  // Aligned memory allocation for SSE
  //!
  //! Use this to allocate a chunk of memory if its start address needs
  //! to be specially aligned to support SSE inline assembly optimization.
  //! \return The start address of newly allocated chunk or NULL if failed.
  //!
  //! \note If SSE inline assembly optimization is not enabled then the normal malloc() will be actually used.
  inline void *ssemalloc(size_t sz)
  {
#ifdef _ASM_SSE_
    return(amalloc(sz, 16));
#else
    return(malloc(sz));
#endif
  }
  //!
  //! This function will actually call agetrealaddress().
  //! \param[in] pvBuffer Start address of the chunk returned by ssemalloc().
  //! \return The real start address.
  //!
  //! \sntc Use ssemalloc() and ssefree() instead.
  inline void *ssegetrealaddress(void *pvBuffer)
  {
#ifdef _ASM_SSE_
    return(agetrealaddress(pvBuffer));
#else
    return(pvBuffer);
#endif
  }
  //!
  //! Use this to free the memory chunks which were allocatted using ssemalloc().
  //! \param[in] pvBuffer Start address of the chunk returned by ssemalloc().
  //!
  //! \note If SSE inline assembly optimization is not enabled then the normal free() will be actually used.
  inline void ssefree(void *pvBuffer)
  {
#ifdef _ASM_SSE_
    afree(pvBuffer);
#else
    free(pvBuffer);
#endif
  }


  // Wrapper functions to select between ssemalloc/free() and normal
  // malloc/free(), depends on the existence of the _ALIGN_BUFF_MEM_
  // macro.
  //!
  //! Use this to allocate memory for special buffers of which start
  //! addresses need to be specially aligned in some configuration.
  inline void *abmmalloc(size_t sz)
  {
#ifdef _ALIGN_BUFF_MEM_
    return(ssemalloc(sz));
#else
    return(malloc(sz));
#endif
  }
  //!
  //! Use this to free the memory which were allocatted using abmmalloc().
  inline void abmfree(void *pvBuffer)
  {
#ifdef _ALIGN_BUFF_MEM_
    ssefree(pvBuffer);
#else
    free(pvBuffer);
#endif
  }


  // Helper function for memory operations
  //!
  //! \internal Empty MMX state, shouldn't need to call this function directly.
  inline void ResetMMX()
  {
    asm
    (
#ifdef _ASM_3DNOW_
      "femms"
#else
      "emms"
#endif
      ::: "mm0", "mm1", "mm2", "mm3", "mm4", "mm5", "mm6", "mm7",
          "st", "st(1)", "st(2)", "st(3)", "st(4)", "st(5)", "st(6)", "st(7)"
    );
  }

  // Memory operations using inline assembly with SSE (and 3DNow)
#ifdef _ASM_SSE_
  inline void ZeroMem4X8(uint32_t *buff, size_t bufferSizeDividedBy32)
  {
#ifdef _ALIGN_BUFF_MEM_
    asm ( "xorps %%xmm0, %%xmm0" ::: "xmm0" );
    while(bufferSizeDividedBy32--) {
      asm
      (
        "movntps %%xmm0, %0 \n\t"
        "movntps %%xmm0, %1     "
        : "=m"(buff[0]), "=m"(buff[4])
      );
      buff += 8;
    }
#else
    asm ( "pxor %%mm0, %%mm0" ::: "mm0", "st" );
    while(bufferSizeDividedBy32--) {
      asm
      (
        "movntq %%mm0, %0 \n\t"
        "movntq %%mm0, %1 \n\t"
        "movntq %%mm0, %2 \n\t"
        "movntq %%mm0, %3     "
        : "=m"(buff[0]), "=m"(buff[2]), "=m"(buff[4]), "=m"(buff[6])
      );
      buff += 8;
    }
    ResetMMX();
#endif
  }

  inline void FillMem4X8(uint32_t *buff, uint32_t value, size_t bufferSizeDividedBy32)
  {
#ifdef _ALIGN_BUFF_MEM_
    asm
    (
      "movd           %0, %%xmm0 \n\t"
      "shufps $0, %%xmm0, %%xmm0     "
       :: "m"(value)
        : "xmm0"
    );
    while(bufferSizeDividedBy32--) {
      asm
      (
        "movntps %%xmm0, %0\n\t"
        "movntps %%xmm0, %1    "
        : "=m"(buff[0]), "=m"(buff[4])
      );
      buff += 8;
    }
#else
    asm
    (
      "movd         %0, %%mm0 \n\t"
      "punpckldq %%mm0, %%mm0     "
       :: "m"(value)
        : "mm0", "st"
    );
    while(bufferSizeDividedBy32--) {
      asm
      (
        "movntq %%mm0, %0\n\t"
        "movntq %%mm0, %1\n\t"
        "movntq %%mm0, %2\n\t"
        "movntq %%mm0, %3    "
        : "=m"(buff[0]), "=m"(buff[2]), "=m"(buff[4]), "=m"(buff[6])
      );
      buff += 8;
    }
    ResetMMX();
#endif
  }

  inline void CopyMem4X8(uint32_t *dstBuff, const uint32_t *srcBuff, size_t bufferSizeDividedBy32)
  {
    asm
    (
      "prefetchnta %0 \n\t"
      "prefetchnta %1 \n\t"
      "prefetchnta %2     "
      :: "m"(srcBuff[0]), "m"(srcBuff[8]), "m"(srcBuff[16])
    );
#ifdef _ALIGN_BUFF_MEM_
    while(bufferSizeDividedBy32--) {
      asm
      (
        "prefetchnta %4         \n\t"
        "movaps      %2, %%xmm0 \n\t"
        "movaps      %3, %%xmm1 \n\t"
        "movntps %%xmm0, %0     \n\t"
        "movntps %%xmm1, %1         "
        : "=m"(dstBuff[0]), "=m"(dstBuff[4])
        :  "m"(srcBuff[0]),  "m"(srcBuff[4]), "m"(srcBuff[24])
        : "xmm0", "xmm1"
      );
      dstBuff += 8;
      srcBuff += 8;
    }
#else
    while(bufferSizeDividedBy32--) {
      asm
      (
        "prefetchnta %8        \n\t"
        "movq        %4, %%mm0 \n\t"
        "movq        %5, %%mm1 \n\t"
        "movq        %6, %%mm2 \n\t"
        "movq        %7, %%mm3 \n\t"
        "movntq   %%mm0, %0    \n\t"
        "movntq   %%mm1, %1    \n\t"
        "movntq   %%mm2, %2    \n\t"
        "movntq   %%mm3, %3        "
        : "=m"(dstBuff[0]), "=m"(dstBuff[2]), "=m"(dstBuff[4]), "=m"(dstBuff[6])
        :  "m"(srcBuff[0]),  "m"(srcBuff[2]),  "m"(srcBuff[4]),  "m"(srcBuff[6]), "m"(srcBuff[24])
        : "mm0", "mm1", "mm2", "mm3", "st", "st(1)", "st(2)", "st(3)"
      );
      dstBuff += 8;
      srcBuff += 8;
    }
    ResetMMX();
#endif
  }

  // Memory operations using inline assembly with MMX (and 3DNow)
  // On CPU which also support SSE, this may even slower than just using
  // plain x86 assembly
#elif defined(_ASM_MMX_)
  inline void ZeroMem4X8(uint32_t *buff, size_t bufferSizeDividedBy32)
  {
    asm ( "pxor %%mm0, %%mm0" ::: "mm0", "st" );
    while(bufferSizeDividedBy32--) {
      asm
      (
        "movq %%mm0, %0 \n\t"
        "movq %%mm0, %1 \n\t"
        "movq %%mm0, %2 \n\t"
        "movq %%mm0, %3     "
        : "=m"(buff[0]), "=m"(buff[2]), "=m"(buff[4]), "=m"(buff[6])
      );
      buff += 8;
    }
    ResetMMX();
  }

  inline void FillMem4X8(uint32_t *buff, uint32_t value, size_t bufferSizeDividedBy32)
  {
    asm
    (
      "movd         %0, %%mm0 \n\t"
      "punpckldq %%mm0, %%mm0     "
       :: "m"(value)
        : "mm0", "st"
    );
    while(bufferSizeDividedBy32--) {
      asm
      (
        "movq %%mm0, %0\n\t"
        "movq %%mm0, %1\n\t"
        "movq %%mm0, %2\n\t"
        "movq %%mm0, %3    "
        : "=m"(buff[0]), "=m"(buff[2]), "=m"(buff[4]), "=m"(buff[6])
      );
      buff += 8;
    }
    ResetMMX();
  }

  inline void CopyMem4X8(uint32_t *dstBuff, const uint32_t *srcBuff, size_t bufferSizeDividedBy32)
  {
    while(bufferSizeDividedBy32--) {
      asm
      (
        "movq    %4, %%mm0 \n\t"
        "movq    %5, %%mm1 \n\t"
        "movq    %6, %%mm2 \n\t"
        "movq    %7, %%mm3 \n\t"
        "movq %%mm0, %0    \n\t"
        "movq %%mm1, %1    \n\t"
        "movq %%mm2, %2    \n\t"
        "movq %%mm3, %3        "
        : "=m"(dstBuff[0]), "=m"(dstBuff[2]), "=m"(dstBuff[4]), "=m"(dstBuff[6])
        :  "m"(srcBuff[0]),  "m"(srcBuff[2]),  "m"(srcBuff[4]),  "m"(srcBuff[6])
        : "mm0", "mm1", "mm2", "mm3", "st", "st(1)", "st(2)", "st(3)"
      );
      dstBuff += 8;
      srcBuff += 8;
    }
    ResetMMX();
  }

  // Memory operations using plain x86_64 or x86 inline assembly
#elif defined(_ASM_ASM_)
  inline void ZeroMem4X8(uint32_t *buff, size_t bufferSizeDividedBy32)
  {
#ifdef _ASM_A64_
    bufferSizeDividedBy32 *= 4;
#else
    bufferSizeDividedBy32 *= 8;
#endif
    asm volatile
    (
      "cld       \n\t"
#ifdef _ASM_A64_
      "rep stosq     "
#else
      "rep stosl     "
#endif
      : "=&D"(buff), "=c"(bufferSizeDividedBy32)
      :   "0"(buff),  "1"(bufferSizeDividedBy32), "a"(0)
      : "memory"
    );
  }

  inline void FillMem4X8(uint32_t *buff, uint32_t value, size_t bufferSizeDividedBy32)
  {
#ifdef _ASM_A64_
    const uint64_t value64 = (uint64_t(value) << 32) | uint64_t(value);
    bufferSizeDividedBy32 *= 4;
#else
    bufferSizeDividedBy32 *= 8;
#endif
    asm volatile
    (
      "cld       \n\t"
#ifdef _ASM_A64_
      "rep stosq     "
#else
      "rep stosl     "
#endif
      : "=&D"(buff), "=c"(bufferSizeDividedBy32)
#ifdef _ASM_A64_
      :   "0"(buff),  "1"(bufferSizeDividedBy32), "a"(value64)
#else
      :   "0"(buff),  "1"(bufferSizeDividedBy32), "a"(value)
#endif
      : "memory"
    );
  }

  inline void CopyMem4X8(uint32_t *dstBuff, const uint32_t *srcBuff, size_t bufferSizeDividedBy32)
  {
#ifdef _ASM_A64_
    bufferSizeDividedBy32 *= 4;
#else
    bufferSizeDividedBy32 *= 8;
#endif
    asm volatile
    (
      "cld        \n\t"
#ifdef _ASM_A64_
      "rep movsq      "
#else
      "rep movsl      "
#endif
      : "=&D"(dstBuff), "=&S"(srcBuff), "=c"(bufferSizeDividedBy32) 
      :   "0"(dstBuff),   "1"(srcBuff),  "2"(bufferSizeDividedBy32)
      : "memory"
    );
  }

  // Memory operations using unrolled QWORD variables assignment
#elif defined(__USE_QW_MEMOPS__)
  inline void ZeroMem4X8(uint32_t *buff, size_t bufferSizeDividedBy32)
  {
    uint64_t *tmp = reinterpret_cast<uint64_t*>(buff);
    for(size_t i = 0; i < bufferSizeDividedBy32; ++i) {
      *tmp++ = 0; *tmp++ = 0; *tmp++ = 0; *tmp++ = 0;
    }
  }

  inline void FillMem4X8(uint32_t *buff, uint32_t value, size_t bufferSizeDividedBy32)
  {
    const uint64_t  val = (uint64_t(value) << 32) | uint64_t(value);
          uint64_t *tmp = reinterpret_cast<uint64_t*>(buff);
    for(size_t i = 0; i < bufferSizeDividedBy32; ++i) {
      *tmp++ = val; *tmp++ = val; *tmp++ = val; *tmp++ = val;
    }
  }

  inline void CopyMem4X8(uint32_t *dstBuff, const uint32_t *srcBuff, size_t bufferSizeDividedBy32)
  {
          uint64_t *d = reinterpret_cast<      uint64_t*>(dstBuff);
    const uint64_t *s = reinterpret_cast<const uint64_t*>(srcBuff);
    for(size_t i = 0; i < bufferSizeDividedBy32; ++i) {
      *d++ = *s++; *d++ = *s++; *d++ = *s++; *d++ = *s++;
    }
  }

  // Memory operations using unrolled DWORD variables assignment
#else
  inline void ZeroMem4X8(uint32_t *buff, size_t bufferSizeDividedBy32)
  {
    uint32_t *tmp = buff;
    for(size_t i = 0; i < bufferSizeDividedBy32; ++i) {
      *tmp++ = 0; *tmp++ = 0; *tmp++ = 0; *tmp++ = 0;
      *tmp++ = 0; *tmp++ = 0; *tmp++ = 0; *tmp++ = 0;
    }
  }

  inline void FillMem4X8(uint32_t *buff, uint32_t value, size_t bufferSizeDividedBy32)
  {
    uint32_t *tmp = buff;
    for(size_t i = 0; i < bufferSizeDividedBy32; ++i) {
      *tmp++ = value; *tmp++ = value; *tmp++ = value; *tmp++ = value;
      *tmp++ = value; *tmp++ = value; *tmp++ = value; *tmp++ = value;
    }
  }

  inline void CopyMem4X8(uint32_t *dstBuff, const uint32_t *srcBuff, size_t bufferSizeDividedBy32)
  {
          uint32_t *d = dstBuff;
    const uint32_t *s = srcBuff;
    for(size_t i = 0; i < bufferSizeDividedBy32; ++i) {
      *d++ = *s++; *d++ = *s++; *d++ = *s++; *d++ = *s++;
      *d++ = *s++; *d++ = *s++; *d++ = *s++; *d++ = *s++;
    }
  }
#endif

} // namespace vge
#endif


/**
 * \fn void vge::ZeroMem4X8(uint32_t *buff, size_t bufferSizeDividedBy32)
 * Use this function to zero the given memory chunk.
 * \param[in] buff                  Start address of the chunk (in 32-bit mode).
 * \param[in] bufferSizeDividedBy32 Size of the chunk divided by 32.
 *
 *  <b>Example:</b>
 *  \code
 *  uint32_t *buff = reinterpret_cast<uint32_t*>(ssemalloc(1024));
 *  ZeroMem4X8(buff, 1024/32);
 *  \endcode
 *  \n
*/

/**
 * \fn void vge::FillMem4X8(uint32_t *buff, uint32_t value, size_t bufferSizeDividedBy32)
 * Use this function to fill the given memory chunk with the given value.
 * \param[in] buff                  Start address of the chunk (in 32-bit mode).
 * \param[in] value                 Value to be filled in (in 32-bit integer).
 * \param[in] bufferSizeDividedBy32 Size of the chunk divided by 32.
 *
 *  <b>Example:</b>
 *  \code
 *  uint32_t *buff = reinterpret_cast<uint32_t*>(ssemalloc(1024));
 *  FillMem4X8(buff, 1024/32);
 *  \endcode
 *  \n
*/

/**
 * \fn void vge::CopyMem4X8(uint32_t *dstBuff, const uint32_t *srcBuff, size_t bufferSizeDividedBy32)
 * Use this function to copy the given source memory chunk to the given destination memory chunk.
 * \param[out] dstBuff               Start address of the destination chunk (in 32-bit mode).
 * \param[in]  srcBuff               Start address of the source chunk (in 32-bit mode).
 * \param[in]  bufferSizeDividedBy32 Size of the data to be copied divided by 32.
 *
 *  <b>Example:</b>
 *  \code
 *  uint32_t *buff1 = reinterpret_cast<uint32_t*>(ssemalloc(1024));
 *  uint32_t *buff2 = reinterpret_cast<uint32_t*>(ssemalloc(1024));
 *  CopyMem4X8(buff1, buff2, 1024/32);
 *  \endcode
 *  \n
*/
