/** \file Thread.h */
// ############################################################################################
//  Thread.h
//  Definition & implementation of CThread class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _SYSTEM___THREAD_H_
#define _SYSTEM___THREAD_H_

// VortexGE headers
#include "../Base/SpecialClasses.h"


namespace vge {

  // Foward reference from "lib/System/Thread.h"
  class CThread;

  //!
  //! \brief Thread data structure
  //!
  //! Shouldn't need to access this structure directly. Use
  //! the VTM_*() helper functions instead.
  //!
  //! \api pretty stable
  typedef struct TThreadData {
    CThread       *pcsThread;  //!< Pointer to the CThread instance owning this thread data structure.
    volatile bool *pbExit;     //!< True if the thread loop is ordered to exit.
    volatile bool *pbSuspend;  //!< True if the thread loop is ordered to suspend.
    void          *pvUserData; //!< Pointer to the custom user data.
  } *PTThreadData;

  //!
  //! \brief User thread-callback funtion
  //!
  //! \api pretty stable
  //! 
  //! The user thread-callback funtion must be constructed like this :
  //! \code
  //! uint64_t MyCallback(TThreadData *ptThreadData)
  //! {
  //!   while(!(*ptThreadData->pbExit)) {
  //!     _if_unlikely(*ptThreadData->pbSuspend) {
  //!       usleep(100000); // Change this value as needed
  //!       continue;
  //!     }
  //!
  //!     <perform user specific actions>
  //!
  //!   }
  //!   return(<return value>);
  //! }
  //! \endcode
  typedef uint64_t (*PFNVortexThreadCallback)(TThreadData *ptThreadData);


  //!
  //! \brief The thread class
  //!
  //! \api pretty stable
  class VGE_EXPORT CThread : VCLASS_UNCOPYABLE {
    public:
      //!
      //! Create and initialize the thread class.
      //! \n\n
      //! If pfnUserCallback is NULL, the ThreadCallback() virtual member function
      //! will be called, otherwise the given callback function will be called.
      CThread(PFNVortexThreadCallback pfnUserCallback = NULL, void *pvUserData = NULL);
      //!
      //! Standard dtor
      virtual ~CThread() throw();

      // Initialization function
      //!
      //! (Re-)initialize the thread class.
      //! \n\n
      //! If pfnUserCallback is NULL, the ThreadCallback() virtual member function
      //! will be called, otherwise the given callback function will be called.
      void Initialize(PFNVortexThreadCallback pfnUserCallback, void *pvUserData);

      // Thread functions
      //!
      //! Start the thread.
      void Start();
      //!
      //! Stop the thread.
      void Stop(uint32_t un32TimeOut_mS = 2500);
      //!
      //! Kill the thread.
      void Kill();

      // Information functions
      //!
      //! Check if the thread is initialized (it can be running or suspeneded).
      inline bool IsRunning() const { return(_tThreadHandle != 0); }
      //!
      //! Check if the thread is suspended.
      inline bool IsSuspended() const { return(_bSuspend); }
      //!
      //! Suspend the thread.
      inline void Suspend() { _bSuspend = true; }
      //!
      //! Resume the thread.
      inline void Resume() { _bSuspend = false; }
      //!
      //! Get the exit code returned by the thread callback.
      inline uint64_t GetExitStatus() const { return(_un64ExitStatus); }
      //!
      //! Get the raw thread handle.
      //! \n\n
      //! \sntc
      inline pthread_t GetThreadHandle() const { return(_tThreadHandle); }

      // Thread callback function
      //!
      //! This virtual member function will be called if on initialization,
      //! either via ctor or via Initialize(), the pfnUserCallback parameter
      //! is set to NULL, otherwise the given callback function will be called.
      virtual uint64_t ThreadCallback(TThreadData &threadData);

    protected:
      // Thread functions
      static void *_ThreadProc(void *pvData) _attr_noreturn;

      // Data
      TThreadData             _tTD;             // Thread control data
      pthread_t               _tThreadHandle;   // Thread handle
      PFNVortexThreadCallback _pfnUserCallback; // User callback
      bool                    _bExit;           // Exit flag
      bool                    _bSuspend;        // Suspend flag
      bool                    _bTEN;            // Exit acknowledge flag
      int64_t                 _un64ExitStatus;  // Exit status
  };

  // Helper functions for thread
  //!
  //! Check if the thread callback loop is ordered to exit or not.
  inline bool VTM_NOT_EXIT(TThreadData &threadData)
  { return(!(*threadData.pbExit)); }
  //!
  //! Check if the thread callback loop is ordered to suspend or not.
  inline bool VTM_SUSPENDED(TThreadData &threadData, u_long sleepAmount_mS = 100000)
  {
    if(*threadData.pbSuspend) {
      usleep(sleepAmount_mS);
      return(true);
    }
    return(false);
  }
  //!
  //! Returns the CThread instance of the corresponding thread data structure.
  inline CThread &VTM_INSTANCE(TThreadData &threadData)
  { return(*threadData.pcsThread); }
  //!
  //! Returns the pointer to the user data of the corresponding thread data structure.
  template <typename UserDataT> inline
  UserDataT *VTM_USER_DATA(TThreadData &threadData)
  { return(reinterpret_cast<UserDataT*>(threadData.pvUserData)); }


  //!
  //! \brief The mutex class
  //!
  //! \api pretty stable
  class VGE_EXPORT CMutex : VCLASS_UNCOPYABLE {
    public:
      // Implement a standard exception-safe locker class for this mutex class
      VCLASS_LOCK_IMPLEMENT(StdLocker, CMutex, Lock(), Unlock())

    public:
      CMutex();
      ~CMutex() throw();

      // Mutex functions
      //!
      //! Lock the mutex, will throw on error.
      void Lock();
      //!
      //! Unlock the mutex, will throw on error or if the lock is
      //! owned by another thread.
      void Unlock();
      //!
      //! Try to lock the mutex, will throw on error.
      //! \n\n
      //! If the given timeout is elapsed but it still can't lock the mutex,
      //! this function will throw an exception of which error code is set
      //! to VERR_Timeout. If the given timeout is zero then this function
      //! will only try to lock once.
      void TryLock(uint32_t un32TimeOut_uS); // Will return false on error or

    protected:
      pthread_mutex_t _tMutex; // Mutex handle
      pthread_t       _tOwner; // Owning thread
  };

} // namespace vge
#endif
