/** \file DialogTypeDefs.h */
// ############################################################################################
//  DialogTypeDefs.h
//  Definition of various data types used by VortexGE dialogs & controls
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _VIDEO_ENVIRONMENT___TYPEDEFS_H_
#define _VIDEO_ENVIRONMENT___TYPEDEFS_H_

// VortexGE headers
#include "../Base/UString.h"
#include "KeyButtonSymbol.h"


// For experiment, just leave it as it
#define __USE_UTF8_DIALOG_STRING__


namespace vge {

  // Foward reference from within the implementation file
  // (eg.: for X11 -> "lib/VideoEnvironment/VID_X11/VX11Environment.h")
  class CVideoEnv;

  // Foward reference from within the implementation file
  // (eg.: for X11 -> "lib/VideoEnvironment/VID_X11/VX11Dialog.h")
  class CSLTextBox;
  class CMLTextBox;


  //! 
  //! \brief IDs of simple dialog controls
  //! 
  //! \api beta
  enum ECtrlID {
    ECID_Control, ECID_StringControl, ECID_Panel, ECID_PanelBar,
    ECID_HLine, ECID_VLine, ECID_Frame, ECID_StatusBar,
    ECID_Label, ECID_Button, ECID_CheckBox, ECID_RadioButton,
    ECID_SpinButton, ECID_SpinButtonEx, ECID_SliderBar,
    ECID_ListBox, ECID_DropDownList, ECID_SLTextBox,
    ECID_MLTextBox, ECID_PopUpMenu, ECID_MenuBar,
    ECID_HLayout, ECID_VLayout
  };

  //! 
  //! \brief Used for specifying the control's border style
  //!
  //! \api beta
  enum EBorderStyle { EBS_Raised, EBS_Sunken };

  //! 
  //! \brief Used for specifying horizontal text aligment mode
  //!
  //! \api beta
  enum EHAlignment { EHTA_Left, EHTA_Center, EHTA_Right };
  //! 
  //! \brief Used for specifying vertical text aligment mode
  //!
  //! \api beta
  enum EVAlignment { EVTA_Top, EVTA_Center, EVTA_Bottom };

  //! 
  //! \brief Used for specifying style of spin button control
  //!
  //! \api beta
  enum ESBStyle {
    ESBS_VLeft, ESBS_VRight,
    ESBS_HTop, ESBS_HBottom
  };
  //! 
  //! \brief Used for specifying value mode of spin button control
  //!
  //! \api beta
  enum ESBValue {
    ESBV_Integer, ESBV_Float1, ESBV_Float2,
    ESBV_Float3, ESBV_Float4, ESBV_Float5
  };

  //! 
  //! \brief Used for specifying value mode for single line text box control
  //!
  //! \api beta
  enum ETBType {
    ETBT_TextNormal, ETBT_TextLowerCase, ETBT_TextUpperCase,
    ETBT_AZ09SNormal, ETBT_AZ09SLowerCase, ETBT_AZ09SUpperCase,
    ETBT_AZ09Normal, ETBT_AZ09LowerCase, ETBT_AZ09UpperCase,
    ETBT_Hexadecimal, ETBT_Integer, ETBT_PositiveInteger,
    ETBT_Float, ETBT_PositiveFloat
  };
  //! 
  //! \brief Used for indicating events in text box control
  //!
  //! \api beta
  enum ETBEvent {
    ETBE_ConfirmOK, ETBE_ConfirmCancel, 
    ETBE_AddChar, ETBE_DelChar, ETBE_ReplaceChar,
    ETBE_Paste
  };


  //!
  //! \brief Enumeration used for specifying the size of a message box
  //!
  //! \api beta
  enum EMsgBoxSize {
    EMBS_Tiny, EMBS_Small, EMBS_Medium,
    EMBS_Large, EMBS_Huge, EMBS_Full
  };
  //!
  //! \brief Enumeration used for specifying what icon to be used in a message box
  //!
  //! \api pretty stable
  enum EMsgBoxIcon {
    EMBI_Information, EMBI_Question,
    EMBI_Warning, EMBI_Error
  };
  //!
  //! \brief Enumeration used for specifying what buttons to be included in a message box
  //!
  //! \api pretty stable
  enum EMsgBoxButtons {
    EMBB_OK, EMBB_OKCancel, EMBB_YesNo, EMBB_YesNoCancel,
    EMBB_AbortRetry, EMBB_AbortRetryIgnore
  };


  //!
  //! \brief Enumeration used for determining what button is pressed in a message box
  //!
  //! \api pretty stable
  enum EMsgBoxRetVal {
    EMBRV_OK, EMBRV_Cancel, EMBRV_Yes, EMBRV_No,
    EMBRV_Abort, EMBRV_Retry, EMBRV_Ignore
  };


  //! 
  //! VortexGE will invoke the user supplied callback function prototyped like
  //! this when events occured in a single line text box control.
  //! \n\n
  //! Returning true will cause the specific action related to the event to be
  //! carried (for example: returning true on ETBE_KeyPress will cause the new
  //! character in "charCode" to be appended to the editBuffer).
  //! \n\n
  //! By using this callback, users can confirm and eject changes made in the
  //! text box and even implement custom value modes which are not defined in
  //! ETBType (by selecting ETBT_TextNormal and implementing the character
  //! filtering from inside the callback function).
  //! \n\n
  //! If the callback function alters whatever displayed in the window, it must
  //! restore the original contents before returning (since, depends on the event,
  //! the text box control may or may not redraw some of the window's area).
  //!
  //! \param[out]    eventCode   Event code as defined in the ETBEvent enumeration.
  //! \param[out]    iVideoEnv   Video environment used by the control instance.
  //! \param[out]    slTextBox   CSLTextBox instance which is calling the callback.
  //! \param[in,out] editBuffer  Edit buffer of the control.
  //! \param[in,out] pasteBuffer Characters to be pasted (valid only on ETBE_Paste event).
  //! \param[in,out] charCode    Character which is currently processed.
  //! \param[out]    keySym      Key symbol as defined in the EKeySymbol enumeration.
  //!
  //! On "ETBE_AddChar" event, "charCode"  will contain the new character to be put,
  //! on "ETBE_DelChar" event it will contain the old character, on other events its
  //! content is not defined.
  //! \n\n
  //! Only on "ETBE_AddChar" event that replacing the content of "charCode" will be
  //! make sense.
  //!
  //! \api beta
#ifndef __USE_UTF8_DIALOG_STRING__
  typedef bool(*PFNSLTextBoxCallback)(ETBEvent eventCode, CVideoEnv &iVideoEnv,
                                      CSLTextBox &slTextBox, string &editBuffer,
                                      string *pasteBuffer, char &charCode,
                                      EKeySymbol keySym);
#else
  typedef bool(*PFNSLTextBoxCallback)(ETBEvent eventCode, CVideoEnv &iVideoEnv,
                                      CSLTextBox &slTextBox, CUString &editBuffer,
                                      CUString *pasteBuffer, CUString::Utf32 &charCode,
                                      EKeySymbol keySym);
#endif

  //! 
  //! VortexGE will invoke the user supplied callback function prototyped like
  //! this when events occured in a multi line text box control.
  //! \n\n
  //! Returning true will cause the specific action related to the event to be
  //! carried (for example: returning true on ETBE_KeyPress will cause the new
  //! character in "charCode" to be appended to the editBuffer).
  //! \n\n
  //! By using this callback, users can confirm and eject changes made in the
  //! text box and implement custom character filtering.
  //! \n\n
  //! If the callback function alters whatever displayed in the window, it must
  //! restore the original contents before returning (since, depends on the event,
  //! the text box control may or may not redraw some of the window's area).
  //!
  //! \param[out]    eventCode   Event code as defined in the ETBEvent enumeration.
  //! \param[out]    iVideoEnv   Video environment used by the control instance.
  //! \param[out]    mlTextBox   CMLTextBox instance which is calling the callback.
  //! \param[in,out] editBuffer  Edit buffer of the control.
  //! \param[out]    actRow      Row position at which the "charCode" comes from or to be sent to.
  //! \param[out]    actCol      Column position at which the "charCode" comes from or to be sent to.
  //! \param[in,out] pasteBuffer Characters to be pasted (valid only on ETBE_Paste event).
  //! \param[in,out] charCode    Character which is currently processed.
  //! \param[out]    keySym      Key symbol as defined in the EKeySymbol enumeration.
  //!
  //! On "ETBE_AddChar" and "ETBE_ReplaceChar" events, "charCode"  will contain the new
  //! character to be put, on "ETBE_DelChar" event it will contain the old character,
  //! on other events its content is not defined.
  //! \n\n
  //! Only on "ETBE_AddChar" and "ETBE_ReplaceChar" events that replacing the content
  //! of "charCode" will be make sense.
  //!
  //! \api beta
#ifndef __USE_UTF8_DIALOG_STRING__
  typedef bool(*PFNMLTextBoxCallback)(ETBEvent eventCode, CVideoEnv &iVideoEnv,
                                      CMLTextBox &mlTextBox, vector<string> &editBuffer,
                                      size_t actRow, size_t actCol,
                                      vector<string> *pasteBuffer, char &charCode,
                                      EKeySymbol keySym);
#else
  typedef bool(*PFNMLTextBoxCallback)(ETBEvent eventCode, CVideoEnv &iVideoEnv,
                                      CMLTextBox &mlTextBox, vector<CUString> &editBuffer,
                                      size_t actRow, size_t actCol,
                                      vector<CUString> *pasteBuffer, CUString::Utf32 &charCode,
                                      EKeySymbol keySym);
#endif


  //! \brief Doublet int32_t XY structure
  //!
  //! \api pretty stable
  struct TInt32XY {
    int32_t x, y;
  
    inline TInt32XY(int32_t x_ = 0, int32_t y_ = 0)
    : x(x_), y(y_)
    {}

    inline TInt32XY &operator+=(const TInt32XY &val)
    { x += val.x; y += val.y; return(*this); }
  
    inline TInt32XY &operator-=(const TInt32XY &val)
    { x -= val.x; y -= val.y; return(*this); }

    inline TInt32XY operator+(const TInt32XY &val)
    { TInt32XY tmp; tmp.x = x + val.x; tmp.y = y + val.y; return(tmp); }

    inline TInt32XY operator-(const TInt32XY &val)
    { TInt32XY tmp; tmp.x = x - val.x; tmp.y = y - val.y; return(tmp); }
  };


} // namespace vge
#endif
