// ############################################################################################
//  ScriptManager.cc
//  Implementation of CLuaManager class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "../Base/MessageDisplayer.h"
#include "../System/Archiver.h"
#include "../System/SystemUtilities.h"
#include "ExecutableByteCodes.h"
#include "ExecutionSpace.h"
#include "ScriptManager.h"

// C++ headers
#include <fstream>

// 3rd party headers
extern "C" {
#include "../../3rdParty/lua/lauxlib.h"
#include "../../3rdParty/lua/lmem.h"
#include "../../3rdParty/lua/lstring.h"
#include "../../3rdParty/lua/lundump.h"
}


namespace vge {

// Extension of xspace saved state files
#define XSPACE_STATE_FILE_EXT ".vl.xspace.state"

// VWA magic numbers
#define _VWA_MGC_ "VWA"


//
// The script manager class
//
CLuaManager::CLuaManager(const string& pathToWorkingDirectory)
: _workDir(pathToWorkingDirectory), _dasls(false)
{
  // Remove the trailing '/' if exist
  if(_workDir[_workDir.length() - 1] == '/') _workDir.erase(_workDir.length() - 1);

  // Validate the working directory
  _ValidateWorkDir_(_workDir);
}

CLuaManager::~CLuaManager() throw()
{
}

void CLuaManager::QueryNodesInGroup(vector<string> &queriedNodes, const string &groupName) const
{
  typedef list<CLuaExecutionSpace*>::const_iterator IterT;

  queriedNodes.clear();
  if((groupName == CLuaExecutionSpace::_sn_NONE) ||
     (groupName == CLuaExecutionSpace::_sn_SELF)) {
    // Do nothing
  }
  else if(groupName == CLuaExecutionSpace::_sn_ALL) {
    for(IterT i = _xlist.begin(); i != _xlist.end(); ++i)
      queriedNodes.push_back((*i)->_nameID);
  }
  else {
    for(IterT i = _xlist.begin(); i != _xlist.end(); ++i)
      if((*i)->_groupID == groupName) queriedNodes.push_back((*i)->_nameID);
  }
}

void CLuaManager::LERegister(CLuaExecutable &luaExecutable)
{
  // Check if the same executable has ever been registered before
  for(list<CLuaExecutable*>::const_iterator it = _elist.begin(); it != _elist.end(); ++it)
    if(*it == &luaExecutable) return;
  // Unregister if the executable is currently registered to other manager
  if(luaExecutable._mgr) luaExecutable._mgr->LEUnregister(luaExecutable);
  // Register it
  _elist.push_back(&luaExecutable);
  luaExecutable._mgr = this;
}

void CLuaManager::LEUnregister(CLuaExecutable &luaExecutable)
{
  luaExecutable._mgr = NULL;
  _elist.remove(&luaExecutable);
}

bool CLuaManager::LERegistered(const CLuaExecutable &luaExecutable) const
{
  for(list<CLuaExecutable*>::const_iterator it = _elist.begin(); it != _elist.end(); ++it)
    if(*it == &luaExecutable) return(true);
  return(false);
}

CLuaExecutable *CLuaManager::LEFindByName(const string &name) const
{
  for(list<CLuaExecutable*>::const_iterator it = _elist.begin(); it != _elist.end(); ++it)
    if((*it)->_nameID == name) return(*it);
  return(NULL);
}

void CLuaManager::LXRegister(CLuaExecutionSpace &luaExecutionSpace)
{
  // Check if the same executable has ever been registered before
  for(list<CLuaExecutionSpace*>::const_iterator it = _xlist.begin(); it != _xlist.end(); ++it)
    if(*it == &luaExecutionSpace) return;

  // Unregister if the executable is currently registered to other manager
  if(luaExecutionSpace._mgr) luaExecutionSpace._mgr->LXUnregister(luaExecutionSpace);

  // Register it
  _xlist.push_back(&luaExecutionSpace);
  luaExecutionSpace._mgr = this;

  // Load the state if needed
  if(!_dasls && !luaExecutionSpace._nameID.empty()) LoadState(luaExecutionSpace);
}

void CLuaManager::LXUnregister(CLuaExecutionSpace &luaExecutionSpace)
{
  // Unregister it
  luaExecutionSpace._mgr = NULL;
  _xlist.remove(&luaExecutionSpace);

  // Save the state if wanted
  if(!_dasls && !luaExecutionSpace._nameID.empty()) SaveState(luaExecutionSpace);
}

bool CLuaManager::LXRegistered(const CLuaExecutionSpace &luaExecutionSpace) const
{
  for(list<CLuaExecutionSpace*>::const_iterator it = _xlist.begin(); it != _xlist.end(); ++it)
    if(*it == &luaExecutionSpace) return(true);
  return(false);
}

CLuaExecutionSpace *CLuaManager::LXFindByName(const string &name) const
{
  for(list<CLuaExecutionSpace*>::const_iterator it = _xlist.begin(); it != _xlist.end(); ++it)
    if((*it)->_nameID == name) return(*it);
  return(NULL);
}

const string CLuaManager::GenerateXSpaceStateFileName(const CLuaExecutionSpace &luaExecutionSpace) const
{
  string bnm  = luaExecutionSpace._groupID.empty()
                ? "no_group"
                : luaExecutionSpace._groupID;
         bnm += ".";
         bnm += luaExecutionSpace._nameID.empty()
                ? "no_name"
                : luaExecutionSpace._nameID;
         bnm += XSPACE_STATE_FILE_EXT;
  return(bnm);
}

void CLuaManager::LoadState(CLuaExecutionSpace &luaExecutionSpace)
{
  // Generate file name
  const string &bnm = GenerateXSpaceStateFileName(luaExecutionSpace);
  const string  fnm = _workDir + "/" + bnm;

  // Open the file for loading state
  ifstream is;
  is.open(fnm.c_str());
  if(is.fail()) {
    // Check if the file is actually exist
    TFileInfo fi;
    try        { VGetFileInfo(fi, fnm); }
    catch(...) { return; }
    // If exist but can't be opened, then throw the appropriate error
    if(fi.Type.IsRegularFile) 
      VError(VERR_OpenFile, _NFL_, "Couldn't open file \"%s\" for reading", fnm.c_str());
    else
      VError(VERR_OpenFile, _NFL_, "\"%s\" is not a regular file and thus it can't be opened", fnm.c_str());
  }

  // Load the state
  const string sn = "file://" + fnm;
  CDataSource ds(is, sn.c_str(), 0);
  LoadState(ds, luaExecutionSpace);
}

void CLuaManager::LoadState(const CDataSource &ds, CLuaExecutionSpace &luaExecutionSpace)
{
  // Preparation
  pistream   &is = ds.PStream();
  const char *nm = ds.Name().c_str();

  // Load the state
  is.exceptions(ios::badbit | ios::failbit | ios::eofbit);
  try {
    luaExecutionSpace.LoadFromStream(is);
  }
  catch(const CException &e) { // Regenerate the exception if it is a VortexGE exception
    if(e.Code() == VERR_InvalidHeaderData) {
      string msg  = e.What();
             msg += " in stream \"";
             msg += nm;
             msg += "\"";
      VError(e.Code(), e.Location().Function(), e.Location().File(), e.Location().Line(), msg.c_str());
    }
    else
      throw;
  }
  catch(ios_base::failure&) {
    VError(VERR_ReadStream, _NFL_, "Couldn't read CLuaExecutionSpace state data from stream \"%s\"", nm);
  }
  catch(...) {
    throw;
  }
}

void CLuaManager::SaveState(const CLuaExecutionSpace &luaExecutionSpace) const
{
  // Generate file name
  const string &bnm = GenerateXSpaceStateFileName(luaExecutionSpace);
  const string  onm = _workDir + "/" + bnm;
  const string  fnm = _workDir + "/$" + bnm;

  // Open the file for saving state
  ofstream os;
  os.open(fnm.c_str());
  if(os.fail()) VError(VERR_OpenFile, _NFL_, "Couldn't open file \"%s\" for writing", fnm.c_str());

  // Save the state
  const string sn = "file://" + fnm;
  CDataSink ds(os, sn.c_str());
  SaveState(ds, luaExecutionSpace);

  // Close the file so that we can safely delete it soon
  os.close();

  // Remove the old file and rename the temporary file
  VRemoveFile(onm, true);
  VRenameFile(fnm, onm);
}

void CLuaManager::SaveState(const CDataSink &ds, const CLuaExecutionSpace &luaExecutionSpace) const
{
  // Preparation
  ostream    &os = ds.Stream();
  const char *nm = ds.Name().c_str();

  // Save the state
  os.exceptions(ios::badbit | ios::failbit | ios::eofbit);
  try {
    luaExecutionSpace.SaveToStream(os);
  }
  catch(ios_base::failure&) {
    VError(VERR_WriteStream, _NFL_, "Couldn't write CLuaExecutionSpace state data to stream \"%s\"", nm);
  }
  catch(...) {
    throw;
  }
}

void CLuaManager::ForceLoadAllStates()
{
  typedef list<CLuaExecutionSpace*>::iterator IterT;
  for(IterT i = _xlist.begin(); i != _xlist.end(); ++i) LoadState(**i);
}

void CLuaManager::ForceSaveAllStates() const
{
  typedef list<CLuaExecutionSpace*>::const_iterator IterT;
  for(IterT i = _xlist.begin(); i != _xlist.end(); ++i) SaveState(**i);
}

void CLuaManager::Compile(CLuaExecutable &exe, const string &sourceCode, bool stripDebugInfo)
{
  // Compilation helper structure
  struct _CH {
    CLuaExecutable *_exe;
    const char     *_sourceCode;
    bool            _stripDebugInfo;

    _CH(CLuaExecutable &dex, const char *scode, bool sdi)
    : _exe(&dex), _sourceCode(scode), _stripDebugInfo(sdi)
    {}

    static int32_t _writer(lua_State *L, const void *byteCodes, size_t size, void *userData)
    {
      UNUSED_PARAM(L);
      CLuaExecutable &e = *reinterpret_cast<CLuaExecutable*>(userData);
      return(!e.Write(reinterpret_cast<const char*>(byteCodes), size));
    }

    static int _pmain(lua_State* L)
    {
      // Get access to the user data
      _CH *ch = reinterpret_cast<_CH*>(lua_touserdata(L, 1));

      // Compile the script
      if(luaL_loadbuffer(L, ch->_sourceCode, strlen(ch->_sourceCode),
                         ch->_exe->GetName().c_str())) {
        luaL_error(L, lua_tostring(L, -1));
        return(1);
      }

      // Clear previously stored byte codes (if any)
      ch->_exe->Clear();

      // Dump the byte codes
      lua_lock(L);
      luaU_dump(L, clvalue(L->top-1)->l.p, _CH::_writer, ch->_exe, ch->_stripDebugInfo);
      lua_unlock(L);

      // Done!
      return(0);
    }
  };

  // Create a new Lua state
  lua_State *L = lua_open();
  if(!L) VError(VERR_NotEnoughMemory, _NFL_, "Not enough memory to create a new Lua state for compiling script");

  // Compile the script
  _CH ch(exe, sourceCode.c_str(), stripDebugInfo);
  if(lua_cpcall(L, _CH::_pmain, &ch)) {
    const string msg(lua_tostring(L, -1));
    lua_pop(L, 1); lua_close(L);
    VError(VERR_CompileScript, _NFL_, "%s", msg.c_str());
  }

  // Clean up the Lua state
  lua_pop(L, 1);
  lua_close(L);
}

void CLuaManager::SaveWorkingDirectory(const CDataSink &sink) const
{
  ForceSaveAllStates();
  ArchiveWorkDirData(sink, _workDir);
}

void CLuaManager::LoadWorkingDirectory(const CDataSource &source)
{
  ExtractWorkDirArchive(_workDir, source);
  ForceLoadAllStates();
}

void CLuaManager::ArchiveWorkDirData(const CDataSink &sink, const string& workDir)
{
  // List the files in the working directory
  vector<string> dl;
  VListDirectory(dl, workDir);

  // The archiver
  CArchiver arc(workDir);

  // Initialize the archiver's user data
  arc.SetUserDataStr(_VWA_MGC_);

  // Initialize the archiver's file list
  for(size_t i = 0; i < dl.size(); ++i) {
    if(dl[i].rfind(XSPACE_STATE_FILE_EXT) != (dl[i].length() - strlen(XSPACE_STATE_FILE_EXT))) continue;
    arc.AddInputFile(dl[i]);
  }

  // Create the archive
  arc.ProcessFiles(sink);
}

void CLuaManager::ExtractWorkDirArchive(const string& workDir, const CDataSource &source)
{
  // Delete all existing state files in the working directory
  vector<string> dl;
  VListDirectory(dl, workDir);
  for(size_t i = 0; i < dl.size(); ++i) {
    if(dl[i].rfind(XSPACE_STATE_FILE_EXT) != (dl[i].length() - strlen(XSPACE_STATE_FILE_EXT))) continue;
    unlink(dl[i].c_str());
  }

  // Validate the working directory
  _ValidateWorkDir_(workDir);

  // Our de-archiver class
  class _CDeArc : public CDeArchiver {
    public:
      inline _CDeArc(const string &outputFilesBasePath, const string &ud, const char *nm)
      : CDeArchiver(outputFilesBasePath), _ud(ud), _nm(nm)
      {}

      virtual ~_CDeArc() throw()
      {}

      virtual void ValidateUserData(const vector<char>& userData) const
      {
        if(userData.size() == _ud.length()) {
          if(memcmp(&userData[0], &_ud[0], _ud.length()) == 0) return;
        }
        VError(VERR_InvalidHeaderData, _NFL_, "\"%s\" is an invalid VWA (VortexLua's working-directory-stored-data) stream", _nm);
      }

    private:
      const string  _ud;
      const char   *_nm;
    };

  // Extract the archive
  _CDeArc dearc(workDir, _VWA_MGC_, source.Name().c_str());
  dearc.ProcessFiles(source);
}

void CLuaManager::_ValidateWorkDir_(const string &workDir)
{
  // Check the given path
  TFileInfo fi;
  try {
    VGetFileInfo(fi, workDir);
  }
  // If got exception then, it maybe that the directory
  // is not yet exist, try to create the directory
  catch(...) {
    VCreateDir(workDir, 0775);
    VInfo(_NFL_, "VortexLua working directory has been created");
    return;
  }

  // If we get here, check if it is really a directory
  if(!fi.Type.IsDirectory) 
      VError(VERR_InvalidParameter, _NFL_, "The given path \"%s\" is not a directory", workDir.c_str());
}

} // namespace vge
