// ############################################################################################
//  mcatedit-MCBuilder.cc
//  Implementation of mcatedit's CMCBuilder class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This program is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// mcatedit's headers
#include "mcatedit-MCBuilder.h"

// C++ headers
#include <fstream>


//
// Our message catalog builder class
//
void CMCBuilder::Clear()
{
  path.clear();
  keys.Clear();
  translations.clear();
  curKey     = 0;
  curTrans   = 0;
  curTransNI = 0;
}

inline const string _CaseFolding_(const string &str)
{
#ifndef __USE_UTF8_DIALOG_STRING__
  string _str;
  for(size_t i = 0; i < str.length(); ++i) _str += tolower(str[i]);
  return(_str);
#else
  //! \todo Add options for selecting :
  //!         1. To use full or simple case-folding
  //!         2. To use (or not to use) special case-folding
  return(CUString(str).FoldCase().ToUtf8());
#endif

}

uint32_t CMCBuilder::Search_Start(const TSearchOption &opt)
{
  _sresult.text.clear();
  _sresult.index.clear();
  _sresult.current = 0;

  if(opt.wordsToFind.empty()) return(0);

  // Tokenize the sentence (if needed)
  string         fullSentence = opt.wordsToFind;
  vector<string> words;
  if(opt.matchingRule != TSearchOption::MatchExactSequence) {
    VTokenize(words, fullSentence);
    if(words.empty()) return(0);
  }

  // Perform case-folding (if needed)
  if(!opt.caseSensitive) {
    if(opt.matchingRule == TSearchOption::MatchExactSequence)
      fullSentence = _CaseFolding_(fullSentence);
    else {
      for(size_t i = 0; i < words.size(); ++i) 
        words[i] = _CaseFolding_(words[i]);
    }
  }

  // Search in original messages
  vector<_TTextEntry> *searchInPtr;
  if(opt.findIn == TSearchOption::OriginalMessage) {
    if(keys.msg.empty()) return(0);
    searchInPtr = &keys.msg;
  }
  // Search in translated messages
  else {
    if(translations[curTrans].msg.empty()) return(0);
    searchInPtr = &translations[curTrans].msg;
  }

  // For convenience
  const vector<_TTextEntry> &searchIn = *searchInPtr;

  // Do the search: match one word
  if(opt.matchingRule == TSearchOption::MatchOneWord) {
    // Walk trough all the messages
    for(size_t i = 0; i < searchIn.size(); ++i) {
      // Perform case-folding (if needed)
      string sText;
      if(opt.caseSensitive) sText = searchIn[i].text;
      else                  sText = _CaseFolding_(searchIn[i].text);
      // Search for words
      for(size_t j = 0; j < words.size(); ++j) {
        // Search in the string
        if(sText.find(words[j]) == string::npos) continue;
        // Push the text
        _sresult.text.push_back(&searchIn[i].text);
        // Push the key index
        if(opt.findIn == TSearchOption::OriginalMessage)
          _sresult.index.push_back(i);
        // Find and push the key index
        else {
          for(size_t k = 0; k < keys.msg.size(); ++k) {
            if(keys.msg[k].id == searchIn[i].id) {
              _sresult.index.push_back(k);
              break;
            }
          }
        }
        // No need to search for other words
        break;
      }
    }
  }
  // Do the search: match all words
  else if(opt.matchingRule == TSearchOption::MatchAllWords) {
    // Walk trough all the messages
    for(size_t i = 0; i < searchIn.size(); ++i) {
      // Perform case-folding (if needed)
      string sText;
      if(opt.caseSensitive) sText = searchIn[i].text;
      else                  sText = _CaseFolding_(searchIn[i].text);
      // Search for all words in the string
      bool found = true;
      for(size_t j = 0; j < words.size(); ++j) {
        if(sText.find(words[j]) == string::npos) {
          found = false;
          break;
        }
      }
      // Found a match ?
      if(!found) continue;
      // Push the text
      _sresult.text.push_back(&searchIn[i].text);
      // Push the key index
      if(opt.findIn == TSearchOption::OriginalMessage)
        _sresult.index.push_back(i);
      // Find and push the key index
      else {
        for(size_t k = 0; k < keys.msg.size(); ++k) {
          if(keys.msg[k].id == searchIn[i].id) {
            _sresult.index.push_back(k);
            break;
          }
        }
      }
    }
  }
  // Do the search: match exact sequence
  else {
    // Walk trough all the messages
    for(size_t i = 0; i < searchIn.size(); ++i) {
      // Perform case-folding (if needed)
      string sText;
      if(opt.caseSensitive) sText = searchIn[i].text;
      else                  sText = _CaseFolding_(searchIn[i].text);
      // Search in the string
      if(sText.find(fullSentence) == string::npos) continue;
      // Push the text
      _sresult.text.push_back(&searchIn[i].text);
      // Push the key index
      if(opt.findIn == TSearchOption::OriginalMessage)
        _sresult.index.push_back(i);
      // Find and push the key index
      else {
        for(size_t k = 0; k < keys.msg.size(); ++k) {
          if(keys.msg[k].id == searchIn[i].id) {
            _sresult.index.push_back(k);
            break;
          }
        }
      }
    }
  }

  // Return the number of the total match(es)
  return(_sresult.text.size());
}

bool CMCBuilder::Search_SelectFirstMatch()
{
  if(_sresult.current) {
    _sresult.current = 0;
    return(true);
  }
  return(false);
}

bool CMCBuilder::Search_SelectPrevMatch()
{
  if(_sresult.current) {
    _sresult.current--;
    return(true);
  }
  return(false);
}

bool CMCBuilder::Search_SelectNextMatch()
{
  if(_sresult.current < _sresult.text.size()-1) {
    _sresult.current++;
    return(true);
  }
  return(false);
}

bool CMCBuilder::Search_SelectLastMatch()
{
  if(_sresult.current != _sresult.text.size()-1) {
    _sresult.current = _sresult.text.size()-1;
    return(true);
  }
  return(false);
}

bool CMCBuilder::Search_GetResult(TSearchResult &result)
{
  if(_sresult.text.empty()) return(false);

  result.resultCurIndex =  _sresult.current;
  result.resultTotIndex =  _sresult.text.size();
  result.fullText       = *_sresult.text[_sresult.current];

  curKey     = _sresult.index[_sresult.current];
  curTransNI = keys.msg[curKey].id;

  return(true);
}


void CMCBuilder::NewCatalog(const string &path_)
{
  Clear();
  path = path_;
}

void CMCBuilder::OpenCatalog(const string &path_)
{
  // Clear the old catalog
  Clear();

  // Save the catalog
  VLoadMessageCatalog(path_, keys, translations);
  // Copy the new path
  path = path_;

  // Select the first message
  for(size_t i = 0; i < translations.size(); ++i) {
    if(!translations[i].locale.empty()) {
      curTrans = i;
      break;
    }
  }
  if(keys.msg.empty())
    curTransNI = 0;
  else
    curTransNI = keys.msg[0].id;
}

void CMCBuilder::SaveCatalog(const string &path_)
{
  // Select the path to be used
  const string pth = !path_.empty() ? path_ : path;

  // Save the catalog
  VSaveMessageCatalog(pth, keys, translations);

  // Copy the path name
  if(pth != path) path = pth;
}

bool CMCBuilder::KeyHasTranslations(uint32_t id) const
{
  for(size_t i = 0; i < translations.size(); ++i) {
    for(size_t j = 0; j < translations[i].msg.size(); ++j) {
      if(!translations[i].msg[j].text.empty() && (translations[i].msg[j].id == id))
        return(true);
    }
  }
  return(false);
}

void CMCBuilder::CreateNewKey()
{
  keys.msg.push_back(_TTextEntry(keys.GenerateNewNumericID()));
  curKey     = keys.msg.size() - 1;
  curTransNI = keys.msg[curKey].id;
}

void CMCBuilder::DeleteCurrentKey()
{
  if(HasNoKey()) return;
  const uint32_t id = GetCurrentKeyID();
  keys.msg.erase(keys.msg.begin() + curKey);
  for(size_t i = 0; i < translations.size(); ++i) {
    for(size_t j = 0; j < translations[i].msg.size(); ++j) {
      // Just mark it as invalid for now, will be fully
      // deleted when saving catalog later
      if(translations[i].msg[j].id == id) translations[i].msg[j].valid = false;
    }
  }
  if(curKey >= keys.msg.size()) curKey = keys.msg.size() - 1;
  curTransNI = keys.msg[curKey].id;
}

bool CMCBuilder::SelectFirstKey()
{
  if(HasNoKey()) return(false);
  if(!GetCurrentKeyIndex()) return(false);
  curKey = 0;
  curTransNI = keys.msg[curKey].id;
  return(true);
}

bool CMCBuilder::SelectPreviousKey()
{
  if(HasNoKey()) return(false);
  if(!GetCurrentKeyIndex()) return(false);
  curKey--;
  curTransNI = keys.msg[curKey].id;
  return(true);
}

bool CMCBuilder::SelectNextKey()
{
  if(HasNoKey()) return(false);
  if(GetCurrentKeyIndex() >= GetLastKeyIndex()) return(false);
  curKey++;
  curTransNI = keys.msg[curKey].id;
  return(true);
}

bool CMCBuilder::SelectLastKey()
{
  if(HasNoKey()) return(false);
  if(GetCurrentKeyIndex() >= GetLastKeyIndex()) return(false);
  curKey = keys.msg.size() - 1;
  curTransNI = keys.msg[curKey].id;
  return(true);
}

bool CMCBuilder::HasTranslationTable(const string &localeName) const
{
  for(size_t f = 0; f < translations.size(); ++f)
    if(translations[f].locale == localeName) return(true);
  return(false);
}

bool CMCBuilder::IsEmptyTranslationTable(const string &localeName) const
{
  for(size_t i = 0; i < translations.size(); ++i) {
    if(translations[i].locale != localeName) continue;
    if(translations[i].msg.empty()) continue;
    for(size_t j = 0; j < translations[i].msg.size(); ++j) {
      if(!translations[i].msg[j].text.empty()) return(false);
    }
  }
  return(true);
}

void CMCBuilder::ListAllLocales(vector<string> &dst)
{
  for(size_t i = 0; i < translations.size(); ++i) {
    if(translations[i].locale.empty()) continue;
    dst.push_back(translations[i].locale);
  }
}

const string &CMCBuilder::GetCurrentTranslationTableText() const
{
  const static string emptyString = "";

  for(size_t i = 0; i < translations[curTrans].msg.size(); ++i) {
    if(translations[curTrans].msg[i].valid &&
       translations[curTrans].msg[i].id == curTransNI) {
      return(translations[curTrans].msg[i].text);
    }
  }
  return(emptyString);
}

bool CMCBuilder::SelectTranslationTable(const string &localeName)
{
  for(size_t i = 0; i < translations.size(); ++i) {
    if(translations[i].locale != localeName) continue;
    curKey   = 0;
    curTrans = i;
    if(!keys.msg.empty()) curTransNI = keys.msg[0].id;
    return(true);
  }
  return(false);
}

_TTranslations *CMCBuilder::GetTranslationTable(const string &localeName)
{
  for(size_t f = 0; f < translations.size(); ++f)
    if(translations[f].locale == localeName) return(&translations[f]);
  return(NULL);
}

void CMCBuilder::DeleteTranslationTable(const string &localeName)
{
  for(size_t i = 0; i < translations.size(); ++i) {
    if(translations[i].locale == localeName) {
      translations.erase(translations.begin() + i);
      break;
    }
  }
}

bool CMCBuilder::SelectFirstUntranslatedMessage()
{
  if(HasNoTranslation() || !keys.msg.size()) return(false);
  return(SearchAndSelectUntranslatedMessage(0, keys.msg.size() - 1));
}

bool CMCBuilder::SelectPreviousUntranslatedMessage()
{
  if(HasNoTranslation() || !curKey) return(false);
  return(SearchAndSelectUntranslatedMessage(curKey - 1, 0));
}

bool CMCBuilder::SelectNextUntranslatedMessage()
{
  if(HasNoTranslation() || curKey==GetLastKeyIndex() || !keys.msg.size()) return(false);
  return(SearchAndSelectUntranslatedMessage(curKey + 1, keys.msg.size() - 1));
}

bool CMCBuilder::SelectLastUntranslatedMessage()
{
  if(HasNoTranslation() || !keys.msg.size()) return(false);
  return(SearchAndSelectUntranslatedMessage(keys.msg.size() - 1, 0));
}

bool CMCBuilder::SearchAndSelectUntranslatedMessage(size_t startIndex, size_t endIndex)
{
  const ssize_t ad = (endIndex >= startIndex) ? 1 : -1;

  // Check in the range of [startIndex, endIndex)
  for(size_t i = startIndex; i != endIndex; i += ad) {
    const uint32_t id = keys.msg[i].id;
          bool     at = false;
    for(size_t j = 0; j < translations[curTrans].msg.size(); ++j) {
      if(translations[curTrans].msg[j].id == id) {
        if(!translations[curTrans].msg[j].text.empty()) {
          at = true;
          break;
        }
      }
    }
    if(!at) {
      curKey     = i;
      curTransNI = keys.msg[curKey].id;
      return(true);
    }
  }

  // Check the end index
  const uint32_t id = keys.msg[endIndex].id;
        bool     at = false;
  for(size_t j = 0; j < translations[curTrans].msg.size(); ++j) {
    if(translations[curTrans].msg[j].id == id) {
      if(!translations[curTrans].msg[j].text.empty()) {
        at = true;
        break;
      }
    }
  }
  if(!at) {
    curKey     = endIndex;
    curTransNI = keys.msg[curKey].id;
    return(true);
  }

  // Not found
  return(false);
}

void CMCBuilder::SetCurrentTranslationText(const string &text)
{
  for(size_t i = 0; i < translations[curTrans].msg.size(); ++i) {
    if((translations[curTrans].msg[i].valid) &&
       (translations[curTrans].msg[i].id == curTransNI)) {
      translations[curTrans].msg[i].text = text;
      return;
    }
  }
  translations[curTrans].msg.push_back(_TTextEntry(curTransNI, text));
}
